"""
Helper functions do identify issues in the mc production.

Felix Metzner 2021
"""

import os
import json
import multiprocessing

from functools import partial
from textops import cat, find, grep
from typing import Union, Optional, Tuple, List, Dict, Any, NamedTuple

from rdstar.utilities import PathType
from rdstar.mc_production import check_mc


__all__ = [
    "EventsInfo",
    "FoundFile",
    "MCProductionEvaluator",
]


class EventsInfo(NamedTuple):
    input_events: int
    output_events: int
    fraction: float

    @staticmethod
    def get_strange_logs_list() -> List[str]:
        __strange_logs = [
            "/group/belle2/TMP/metzner/mc_production_august_2020/Dlnu/gsim/log/DzeroTau/evtgen_exp_07_DzeroTau-0.log",
            "/group/belle2/TMP/metzner/mc_production_august_2020/DstlnuMVA/gsim/log/DstarzeroTau/evtgen_exp_07_DstarzeroTau-0.log",
            "/group/belle2/TMP/metzner/mc_production_august_2020/GapDEtaViaDstst/gsim/log/GapDstetaViaDprimeoneMixed/evtgen_exp_07_GapDstetaViaDprimeoneMixed-1.log",
        ]  # type: List[str]
        return __strange_logs

    @staticmethod
    def from_log(log_file_path: PathType) -> "EventsInfo":
        in_lines = log_file_path | cat() | grep("\#input events").tolist()  # noqa: W605
        assert len(in_lines) == 1, in_lines
        in_line = in_lines[0]
        out_lines = log_file_path | cat() | grep("\#output events").tolist()  # noqa: W605
        assert len(out_lines) == 1, out_lines
        out_line = out_lines[0]

        assert isinstance(in_line, str), (in_line, type(in_line))
        assert isinstance(out_line, str), (out_line, type(out_line))
        assert in_line.startswith("RESULT") or log_file_path in EventsInfo.get_strange_logs_list(), in_line
        assert out_line.startswith("RESULT") or log_file_path in EventsInfo.get_strange_logs_list(), out_line

        in_events = in_line.rsplit(sep=": ", maxsplit=1)[1].replace(" ", "")
        out_events, frac_str = out_line.rsplit(sep=": ", maxsplit=1)[1].split(sep=" (", maxsplit=1)
        out_events = out_events.replace(" ", "")
        fraction = float(frac_str.split(sep="%)", maxsplit=1)[0])

        assert in_events.isdigit(), in_events
        assert out_events.isdigit(), out_events

        return EventsInfo(
            input_events=int(in_events),
            output_events=int(out_events),
            fraction=fraction,
        )

    def as_dict(self) -> Dict[str, Union[int, float]]:
        return {
            "input_events": self.input_events,
            "output_events": self.output_events,
            "fraction": self.fraction,
        }


class FoundFile(NamedTuple):
    exp: str
    data_type: str
    file_index: int
    events: Optional[EventsInfo] = None
    issues: Optional[List[Tuple[str, int]]] = None

    @staticmethod
    def from_path(
        file_path: PathType,
        events_info: Optional[EventsInfo] = None,
        issues: Optional[List[Tuple[str, int]]] = None,
    ) -> "FoundFile":
        file_info = FoundFile.get_info_from_file_path(file_path=file_path)
        return FoundFile(
            exp=file_info.exp,
            data_type=file_info.data_type,
            file_index=file_info.file_index,
            events=events_info,
            issues=issues,
        )

    def to_path(self, category: str) -> PathType:
        base_dir_path = "/group/belle2/TMP/metzner/mc_production_august_2020/"
        file_name = f"evtgen_exp_{self.exp}_{self.data_type}-{self.file_index}.log"
        return os.path.join(base_dir_path, category, "gsim/log", self.data_type, file_name)

    @staticmethod
    def get_info_from_file_path(file_path: PathType) -> "FoundFile":
        file_name, _ = os.path.splitext(os.path.basename(file_path))
        assert "." not in file_name, file_name
        assert file_name[-1].isdigit(), file_name
        assert file_name.startswith("evtgen_exp_")
        assert isinstance(file_name, str)
        exp, rest = file_name.split(sep="_exp_", maxsplit=1)[1].split(sep="_", maxsplit=1)
        data_type, file_index = rest.split(sep="-", maxsplit=1)
        assert file_index.isdigit(), (rest, data_type, file_index)
        return FoundFile(
            exp=exp,
            data_type=data_type,
            file_index=int(file_index),
        )

    def as_dict(self) -> Dict[str, Any]:
        return {
            "exp": self.exp,
            "data_type": self.data_type,
            "file_index": self.file_index,
            "events": self.events.as_dict() if self.events is not None else {},
            "issues": [] if self.issues is None else self.issues,
        }

    @staticmethod
    def from_dict(file_info_as_dict) -> "FoundFile":
        return FoundFile(
            exp=file_info_as_dict["exp"],
            data_type=file_info_as_dict["data_type"],
            file_index=file_info_as_dict["file_index"],
            events=EventsInfo(**file_info_as_dict["events"]),
            issues=file_info_as_dict["issues"] if file_info_as_dict["issues"] else None,
        )


class MCProductionEvaluator:
    default_log_file_name = "mc_production_check.json"  # type: str

    base_dir_path = "/group/belle2/TMP/metzner/mc_production_august_2020/"  # type: PathType

    log_file_extension = ".log"  # type: str
    log_dir_signature = "gsim/log"  # type: PathType
    mdst_dir_signature = "gsim/mdst"  # type: PathType

    mc_categories = [
        "Dlnu",
        "DlnuMVA",
        "Dstlnu",
        "DstlnuMVA",
        "Dststlnu",
        "Dststtaunu",
        "GapD1",
        "GapDEta",
        "GapDEtaViaDstst",
        "GapDPiPi",
        "GapDPiPiViaDstst",
        "GapDstPiPi",
        "GapDstPiPiViaDstst",
        "SignalDststlnupim",
        "SignalDststlnupimMVA",
        "SignalDststlnupiz",
        "SignalDststlnupizMVA",
        "SignalDststtaunupim",
        "SignalDststtaunupiz",
    ]  # type: List[str]

    error_log_lines = [
        "could not fork new process for connection: Resource temporarily unavailable",
        "could not connect to server: Connection timed out",
        "ERR    : PntDB:could not connect to server: Connection timed out",
    ]  # type: List[str]

    def __init__(
        self,
        info_output_path: PathType,
        processes: int = 10,
    ) -> None:
        self.processes = processes  # type: int
        self.problematic_log_files = {}  # type: Dict[str, Dict[str, FoundFile]]
        self.checked_categories = []  # type: List[str]
        self.info_log_path = info_output_path  # type: PathType

        self.already_checked_categories = None  # type: Optional[List[str]]
        if os.path.isfile(info_output_path):
            old_info, checked_cats = self.read_info_from_file(input_path=info_output_path)
            self.problematic_log_files = old_info
            self.already_checked_categories = checked_cats

    def is_log_file(self, path_to_file: PathType) -> bool:
        name, ext = os.path.splitext(os.path.basename(path_to_file))
        if ext != self.log_file_extension:
            return False
        if name.endswith("log"):
            return False
        if not name.startswith("evtgen"):
            return False
        return True

    def update_problematic_files_dict(
        self,
        category: str,
        file: PathType,
        events_info: EventsInfo,
        issues: List[Tuple[str, int]],
    ) -> None:
        if category not in self.problematic_log_files.keys():
            self.problematic_log_files.update({category: {}})

        assert file not in self.problematic_log_files[category].keys(), file
        file_info = FoundFile.from_path(file_path=file, events_info=events_info, issues=issues)
        self.problematic_log_files[category].update({str(file): file_info})

    def check_mc_production(self, overwrite_output: bool = False) -> None:
        for mc_category in self.mc_categories:
            if self.already_checked_categories is not None and mc_category in self.already_checked_categories:
                continue
            mc_dir = os.path.join(self.base_dir_path, mc_category, self.log_dir_signature)
            print(f"Looking into {mc_dir}:")

            log_file_list_obj = mc_dir | find(f"*{self.log_file_extension}")
            log_file_list = [lf for lf in log_file_list_obj if self.is_log_file(path_to_file=lf)]

            problematic_files_info = []
            with multiprocessing.Pool(processes=self.processes) as process_pool:
                log_info_list = process_pool.map(partial(MCProductionEvaluator.check_log_file), log_file_list)
                for log_info in log_info_list:
                    if log_info is not None:
                        problematic_files_info.append(log_info)

            if problematic_files_info:
                for log_file_path, events_info, lines_with_errors in problematic_files_info:
                    self.update_problematic_files_dict(
                        category=mc_category,
                        file=log_file_path,
                        events_info=events_info,
                        issues=lines_with_errors,
                    )

            self.checked_categories.append(mc_category)

            self.write_info_to_file(output_path=self.info_log_path, overwrite=True, backup=True)

        self.write_info_to_file(output_path=self.info_log_path, overwrite=overwrite_output)

        print(self.problematic_log_files)

    @staticmethod
    def check_log_file(
        log_file_path: PathType,
    ) -> Optional[Tuple[str, EventsInfo, List[Tuple[str, int]]]]:
        lines_with_error = []  # type: List[Tuple[str, int]]
        for error_line in MCProductionEvaluator.error_log_lines:
            lines = log_file_path | cat() | grep(error_line).tolist()
            if lines:
                print(f"The following line occurred {len(lines)} times in {log_file_path}:\n\t{error_line}")
                lines_with_error.append((error_line, len(lines)))

        if lines_with_error:
            events_info = EventsInfo.from_log(log_file_path=log_file_path)
            return str(log_file_path), events_info, lines_with_error

        return None

    def write_info_to_file(
        self,
        output_path: PathType,
        overwrite: bool = False,
        backup: bool = False,
    ) -> None:
        _output_path = output_path
        if backup:
            dir_path, file = os.path.split(output_path)
            _output_path = os.path.join(dir_path, f"backup_{file}")

        if not overwrite and os.path.isfile(_output_path):
            raise RuntimeError(f"File exists, but overwrite is set to False:\n\t{_output_path}")
        log_file_dict = {
            category: {p: info.as_dict() for p, info in c_dict.items()}
            for category, c_dict in self.problematic_log_files.items()
        }  # type: Dict[str, Union[List[str], Dict[str, Any]]]
        log_file_dict.update({"checked_categories": self.checked_categories})
        with open(_output_path, "w") as out_file:
            json.dump(log_file_dict, out_file, indent=2)

    @staticmethod
    def read_info_from_file(
        input_path: PathType,
    ) -> Tuple[Dict[str, Dict[str, FoundFile]], List[str]]:
        assert os.path.isfile(input_path), input_path
        with open(input_path, "r") as in_file:
            log_file_dict = json.load(fp=in_file)

        checked_categories = log_file_dict.pop("checked_categories")
        assert isinstance(checked_categories, list), checked_categories
        assert all(isinstance(category, str) for category in checked_categories), checked_categories
        log_file_dict = {
            category: {p: FoundFile.from_dict(info) for p, info in c_dict.items()}
            for category, c_dict in log_file_dict.items()
        }
        return log_file_dict, checked_categories

    @staticmethod
    def get_files_frank_found() -> List[FoundFile]:
        files_frank_found = [
            FoundFile(exp="55", data_type="DprimeonezeroE", file_index=112),
            FoundFile(exp="55", data_type="DprimeonezeroE", file_index=119),
            FoundFile(exp="55", data_type="DprimeonezeroE", file_index=124),
            FoundFile(exp="55", data_type="DprimeonezeroE", file_index=131),
            FoundFile(exp="55", data_type="DprimeonezeroE", file_index=137),
            FoundFile(exp="55", data_type="DprimeonezeroE", file_index=142),
            FoundFile(exp="55", data_type="DprimeonezeroE", file_index=28),
            FoundFile(exp="55", data_type="DprimeonezeroE", file_index=39),
            FoundFile(exp="55", data_type="DprimeonezeroE", file_index=68),
            FoundFile(exp="55", data_type="DprimeonezeroE", file_index=81),
            FoundFile(exp="55", data_type="DprimeonezeroE", file_index=91),
            FoundFile(exp="55", data_type="DprimeonezeroE", file_index=92),
            FoundFile(exp="55", data_type="DprimeonezeroE", file_index=93),
            FoundFile(exp="55", data_type="DprimeonezeroE", file_index=94),
            FoundFile(exp="55", data_type="DprimeonezeroE", file_index=95),
            FoundFile(exp="55", data_type="DprimeonezeroE", file_index=96),
            FoundFile(exp="55", data_type="DprimeonezeroE", file_index=99),
            FoundFile(exp="55", data_type="DprimeonezeroE", file_index=9),
            FoundFile(exp="55", data_type="DprimeonezeroMu", file_index=126),
            FoundFile(exp="55", data_type="DprimeonezeroMu", file_index=145),
            FoundFile(exp="55", data_type="DprimeonezeroMu", file_index=146),
            FoundFile(exp="55", data_type="DprimeonezeroMu", file_index=150),
            FoundFile(exp="55", data_type="DprimeonezeroMu", file_index=151),
            FoundFile(exp="55", data_type="DprimeonezeroMu", file_index=23),
            FoundFile(exp="55", data_type="DprimeonezeroMu", file_index=30),
            FoundFile(exp="55", data_type="DprimeonezeroMu", file_index=42),
            FoundFile(exp="55", data_type="DprimeonezeroMu", file_index=49),
            FoundFile(exp="55", data_type="DprimeonezeroMu", file_index=50),
            FoundFile(exp="55", data_type="DprimeonezeroMu", file_index=58),
            FoundFile(exp="55", data_type="DprimeonezeroMu", file_index=67),
            FoundFile(exp="55", data_type="DprimeonezeroMu", file_index=79),
            FoundFile(exp="61", data_type="DprimeonezeroE", file_index=12),
            FoundFile(exp="61", data_type="DprimeonezeroE", file_index=14),
            FoundFile(exp="61", data_type="DprimeonezeroE", file_index=16),
            FoundFile(exp="61", data_type="DprimeonezeroE", file_index=17),
            FoundFile(exp="61", data_type="DprimeonezeroMu", file_index=14),
            FoundFile(exp="61", data_type="DprimeonezeroMu", file_index=15),
            FoundFile(exp="61", data_type="DprimeonezeroMu", file_index=1),
            FoundFile(exp="61", data_type="DprimeonezeroMu", file_index=20),
            FoundFile(exp="61", data_type="DprimeonezeroMu", file_index=23),
            FoundFile(exp="61", data_type="DprimeonezeroMu", file_index=26),
            FoundFile(exp="61", data_type="DprimeonezeroMu", file_index=27),
            FoundFile(exp="61", data_type="DprimeonezeroMu", file_index=28),
            FoundFile(exp="61", data_type="DprimeonezeroMu", file_index=29),
            FoundFile(exp="61", data_type="DprimeonezeroMu", file_index=32),
            FoundFile(exp="43", data_type="DprimeoneminusMu", file_index=80),
            FoundFile(exp="43", data_type="DprimeoneminusMu", file_index=82),
            FoundFile(exp="39", data_type="DoneminusE", file_index=46),
            FoundFile(exp="39", data_type="DoneminusE", file_index=47),
            FoundFile(exp="39", data_type="DoneminusE", file_index=48),
            FoundFile(exp="61", data_type="DoneminusMu", file_index=28),
            FoundFile(exp="37b", data_type="DtwostarminusMu", file_index=2),
            FoundFile(exp="55", data_type="DtwostarzeroMu", file_index=34),
            FoundFile(exp="55", data_type="DtwostarzeroMu", file_index=89),
            FoundFile(exp="37a", data_type="DzerostarminusMu", file_index=40),
        ]  # type: List[FoundFile]
        return files_frank_found

    @staticmethod
    def get_log_file_dir_path() -> PathType:
        mc_check_path = os.path.dirname(check_mc.__file__)
        assert os.path.isdir(mc_check_path), mc_check_path
        return mc_check_path

    @staticmethod
    def get_log_file_path() -> PathType:
        log_file_path = os.path.join(
            MCProductionEvaluator.get_log_file_dir_path(),
            MCProductionEvaluator.default_log_file_name,
        )
        assert os.path.isfile(log_file_path), log_file_path
        return log_file_path


if __name__ == "__main__":
    log_path = "/home/belle2/metzner/RDStar/rdstar/mc_production/check_mc/mc_production_check.log"  # type: str
    mc_evaluator = MCProductionEvaluator(info_output_path=log_path, processes=16)
    try:
        mc_evaluator.check_mc_production(overwrite_output=True)
    except KeyboardInterrupt as ki:
        mc_evaluator.write_info_to_file(output_path=log_path, overwrite=True)
        raise ki
