"""
Reweighting of Branching Fractions of events involving the decays
    * B -> D l nu
    * B -> D* l nu
    * B -> D** l nu

Felix Metzner and Patrick Ecker, 2021
"""

import copy
import numpy as np
import pandas as pd

from typing import Tuple, Dict, List

from rdstar.offline_analysis.rdstar_sample_weights import signal_mc_bfs
from rdstar.offline_analysis.rdstar_samples import get_b_flavour_from_data_type, non_b_data_types

from rdstar.offline_analysis.generator_level_mc_matching import gen_b1_id_col_name, gen_b2_id_col_name

from rdstar.branching_fractions import b_meson_bf_calculation as bf_calc
from rdstar.branching_fractions.old_new_bf_comparison_mixed_b import MixedBBFComparison
from rdstar.branching_fractions.old_new_bf_comparison_charged_b import ChargedBBFComparison
from rdstar.branching_fractions import belle_charged_b_meson_decay_info as belle_charged_b_bfs
from rdstar.branching_fractions import belle_neutral_b_meson_decay_info as belle_neutral_b_bfs


__all__ = [
    "BMesonBFReweighter",
]


def get_second_daughter_pdg_col(b_index: int) -> str:
    assert b_index in (0, 1), b_index
    pdg_string_start: str = f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo0__cm"
    return f"{pdg_string_start}mcDaughter__bo1__cmPDG__bc__bc__bc__bc"


def get_third_daughter_pdg_col(b_index: int) -> str:
    assert b_index in (0, 1), b_index
    pdg_string_start: str = f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo0__cm"
    return f"{pdg_string_start}mcDaughter__bo2__cmPDG__bc__bc__bc__bc"


class BMesonBFReweighter:
    b_bf_weight_col_name: str = "__weight_b_meson_bf__"

    _own_b1_gen_id_col: str = "__b1_gen_id_for_bf_reweighting__"
    _own_b2_gen_id_col: str = "__b2_gen_id_for_bf_reweighting__"

    b1_gen_id_col_name: str = gen_b1_id_col_name
    b2_gen_id_col_name: str = gen_b2_id_col_name

    _required_cols: List[str] = [
        b1_gen_id_col_name,
        b2_gen_id_col_name,
        get_second_daughter_pdg_col(b_index=0),
        get_second_daughter_pdg_col(b_index=1),
        get_third_daughter_pdg_col(b_index=0),
        get_third_daughter_pdg_col(b_index=1),
    ]

    # Dstst Gen IDs
    dstst_gen_ids: Tuple[int, ...] = (51, 52, 53, 54, 55, 56, 57, 58)

    # Old branching fractions which shall be reweighted:
    _gen_id_belle_bf_dict: Dict[str, Dict[int, float]] = {
        "b_charged": {
            1: belle_charged_b_bfs.belle_bf_bp_to_d_tau_nu,
            2: belle_charged_b_bfs.belle_bf_bp_to_d_st_tau_nu,
            3: belle_charged_b_bfs.belle_bf_bp_to_d_l_nu,
            4: belle_charged_b_bfs.belle_bf_bp_to_d_st_l_nu,
            # Dstst modes
            510: belle_charged_b_bfs.belle_bf_bp_to_done_to_dst_pi_l_nu,
            520: belle_charged_b_bfs.belle_bf_bp_to_dtwo_st_l_nu,
            530: belle_charged_b_bfs.belle_bf_bp_to_done_prime_l_nu,
            540: belle_charged_b_bfs.belle_bf_bp_to_dzero_st_l_nu,
            550: belle_charged_b_bfs.belle_bf_bp_to_done_to_dst_pi_tau_nu,
            560: belle_charged_b_bfs.belle_bf_bp_to_dtwo_st_tau_nu,
            570: belle_charged_b_bfs.belle_bf_bp_to_done_prime_tau_nu,
            580: belle_charged_b_bfs.belle_bf_bp_to_dzero_st_tau_nu,
            # Done Gap Case
            5100: belle_charged_b_bfs.belle_bf_bp_to_done_to_d_pi_pi_l_nu,
            5500: belle_charged_b_bfs.belle_bf_bp_to_done_to_d_pi_pi_tau_nu,  # Will not be reweighted!
        },
        "b_mixed": {
            1: belle_neutral_b_bfs.belle_bf_b0_to_d_tau_nu,
            2: belle_neutral_b_bfs.belle_bf_b0_to_d_st_tau_nu,
            3: belle_neutral_b_bfs.belle_bf_b0_to_d_l_nu,
            4: belle_neutral_b_bfs.belle_bf_b0_to_d_st_l_nu,
            # Dstst modes
            510: belle_neutral_b_bfs.belle_bf_b0_to_done_to_dst_pi_l_nu,
            520: belle_neutral_b_bfs.belle_bf_b0_to_dtwo_st_l_nu,
            530: belle_neutral_b_bfs.belle_bf_b0_to_done_prime_l_nu,
            540: belle_neutral_b_bfs.belle_bf_b0_to_dzero_st_l_nu,
            550: belle_neutral_b_bfs.belle_bf_b0_to_done_to_dst_pi_tau_nu,
            560: belle_neutral_b_bfs.belle_bf_b0_to_dtwo_st_tau_nu,
            570: belle_neutral_b_bfs.belle_bf_b0_to_done_prime_tau_nu,
            580: belle_neutral_b_bfs.belle_bf_b0_to_dzero_st_tau_nu,
            # Done Gap Case
            5100: belle_neutral_b_bfs.belle_bf_b0_to_done_to_d_pi_pi_l_nu,
            5500: belle_neutral_b_bfs.belle_bf_b0_to_done_to_d_pi_pi_tau_nu,  # Will not be reweighted!
        },
    }

    # New branching fractions to reweight to:
    _gen_id_updated_bf_dict: Dict[str, Dict[int, float]] = {
        "b_charged": {
            1: signal_mc_bfs["DZTau"].bf,
            2: signal_mc_bfs["DSZTau"].bf,
            3: signal_mc_bfs["DZE"].bf,
            4: signal_mc_bfs["DSZE"].bf,
            # Dstst modes
            510: signal_mc_bfs["BMD1E"].bf,
            520: signal_mc_bfs["BMD2sE"].bf,
            530: signal_mc_bfs["BMDp1E"].bf,
            540: signal_mc_bfs["BMD0sE"].bf,
            550: signal_mc_bfs["TBMD1Tau"].bf,
            560: signal_mc_bfs["TBMD2sTau"].bf,
            570: signal_mc_bfs["TBMDp1Tau"].bf,
            580: signal_mc_bfs["TBMD0sTau"].bf,
            # Done Gap Case
            5100: bf_calc.bf_bp_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap.nominal_value,
        },
        "b_mixed": {
            1: signal_mc_bfs["DMTau"].bf,
            2: signal_mc_bfs["DSMTau"].bf,
            3: signal_mc_bfs["DME"].bf,
            4: signal_mc_bfs["DSME"].bf,
            # Dstst modes
            510: signal_mc_bfs["BZD1E"].bf,
            520: signal_mc_bfs["BZD2sE"].bf,
            530: signal_mc_bfs["BZDp1E"].bf,
            540: signal_mc_bfs["BZD0sE"].bf,
            550: signal_mc_bfs["TBZD1Tau"].bf,
            560: signal_mc_bfs["TBZD2sTau"].bf,
            570: signal_mc_bfs["TBZDp1Tau"].bf,
            580: signal_mc_bfs["TBZD0sTau"].bf,
            # Done Gap Case
            5100: bf_calc.bf_b0_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap.nominal_value,
        },
    }

    own_do_nothing_id: int = -9999
    own_others_id: int = -9998

    _not_reweighted_keys: Tuple[int, ...] = (5200, 5300, 5400, 5500, 5600, 5700, 5800)

    double_counting_cases: Dict[int, Tuple[str, ...]] = {
        # Dstst regular Cases
        510: ("BZD1E", "BZD1Mu", "BMD1E", "BMD1Mu"),
        520: ("BZD2sE", "BZD2sMu", "BMD2sE", "BMD2sMu"),
        530: ("BZDp1E", "BZDp1Mu", "BMDp1E", "BMDp1Mu"),
        540: ("BZD0sE", "BZD0sMu", "BMD0sE", "BMD0sMu"),
        550: ("TBZD1Tau", "TBMD1Tau"),
        560: ("TBZD2sTau", "TBMD2sTau"),
        570: ("TBZDp1Tau", "TBMDp1Tau"),
        580: ("TBZD0sTau", "TBMD0sTau"),
        # Dstst Gap Cases
        5100: ("GapD1Charged", "GapD1Mixed"),
    }
    double_counting_cases_map: Dict[str, int] = {kk: k for k, v in double_counting_cases.items() for kk in v}

    double_counting_data_types: Tuple[str, ...] = tuple([t for ts in double_counting_cases.values() for t in ts])

    def __init__(
        self,
        data_type_column_name: str,
        rescale_others: bool,
    ) -> None:
        assert all([dt in signal_mc_bfs.keys() for dt in [v for vs in self.double_counting_cases.values() for v in vs]])

        flavour_keys: Tuple[str, ...] = tuple(["b_charged", "b_mixed"])
        assert all(k in flavour_keys for k in self._gen_id_belle_bf_dict.keys())
        assert all(k in flavour_keys for k in self._gen_id_updated_bf_dict.keys())
        assert all(
            set(self._gen_id_belle_bf_dict[flavour_keys[0]]) == set(v) for v in self._gen_id_belle_bf_dict.values()
        ), self._gen_id_belle_bf_dict
        assert all(
            set(self._gen_id_updated_bf_dict[flavour_keys[0]]) == set(v) for v in self._gen_id_updated_bf_dict.values()
        ), self._gen_id_updated_bf_dict

        new_ids: Tuple[int, ...] = tuple(set([v for vs in self._gen_id_updated_bf_dict.values() for v in vs]))
        belle_ids: Tuple[int, ...] = tuple(set([v for vs in self._gen_id_belle_bf_dict.values() for v in vs]))
        assert all([k in belle_ids for k in new_ids]), (new_ids, belle_ids)
        assert all([k in new_ids or k in self._not_reweighted_keys for k in belle_ids]), (new_ids, belle_ids)

        self.data_type_col: str = data_type_column_name
        self.rescale_others: bool = rescale_others
        self.all_reweight_ids: Tuple[int, ...] = new_ids

    @property
    def required_cols(self) -> Tuple[str, ...]:
        return tuple([*self._required_cols, self.data_type_col])

    def add_bf_weights(self, df: pd.DataFrame) -> None:
        assert self.b_bf_weight_col_name not in df.columns, self.b_bf_weight_col_name
        assert all(c in df.columns for c in self.required_cols), [c for c in self.required_cols if c not in df.columns]

        unique_dts: List[str] = list(df[self.data_type_col].unique())

        if df.empty:
            df[self.b_bf_weight_col_name] = 1.0
            return

        self._add_special_gen_ids(df=df)
        df.loc[:, self.b_bf_weight_col_name] = np.NaN

        for data_type in unique_dts:
            dt_cond: pd.Series = df[self.data_type_col] == data_type

            if data_type in non_b_data_types:
                bf_weight_values: np.ndarray = self._get_neutral_weights(df=df.loc[dt_cond])
            else:
                bf_weight_values = self._get_bf_weights(df=df.loc[dt_cond], data_type=data_type)

            assert len(bf_weight_values) == sum(dt_cond), (len(bf_weight_values), sum(dt_cond), len(df.loc[dt_cond]))
            df.loc[dt_cond, self.b_bf_weight_col_name] = bf_weight_values

        assert not np.any(np.isnan(df[self.b_bf_weight_col_name])), np.sum(np.isnan(df[self.b_bf_weight_col_name]))
        self._remove_special_gen_ids(df=df)

    def add_bf_weights_for_datatype(self, df: pd.DataFrame, data_type: str) -> None:
        assert self.b_bf_weight_col_name not in df.columns, self.b_bf_weight_col_name
        assert all(c in df.columns for c in self.required_cols), [c for c in self.required_cols if c not in df.columns]

        if df.empty:
            df[self.b_bf_weight_col_name] = 1.0
            return

        unique_dts: List[str] = list(df[self.data_type_col].unique())
        assert len(unique_dts) == 1, unique_dts
        assert unique_dts[0] == data_type, (unique_dts, data_type)

        if data_type in non_b_data_types:
            df.loc[:, self.b_bf_weight_col_name] = self._get_neutral_weights(df=df)
            return

        self._add_special_gen_ids(df=df)
        bf_weight_values: np.ndarray = self._get_bf_weights(df=df, data_type=data_type)
        assert len(bf_weight_values) == len(np.unique(df.index)), (
            len(bf_weight_values),
            len(np.unique(df.index)),
            len(df),
        )
        df.loc[:, self.b_bf_weight_col_name] = bf_weight_values
        self._remove_special_gen_ids(df=df)

    def _get_bf_weights(self, df: pd.DataFrame, data_type: str) -> np.ndarray:
        assert self._own_b1_gen_id_col in df.columns, self._own_b1_gen_id_col
        assert self._own_b2_gen_id_col in df.columns, self._own_b2_gen_id_col

        b_flavour: str = self.get_b_flavour_from_data_type(data_type=data_type)
        old_gen_id_to_bf_mapping: Dict[int, float] = copy.copy(self._gen_id_belle_bf_dict[b_flavour])
        new_gen_id_to_bf_mapping: Dict[int, float] = copy.copy(self._gen_id_updated_bf_dict[b_flavour])
        assert self.own_do_nothing_id not in old_gen_id_to_bf_mapping.keys(), self.own_do_nothing_id
        assert self.own_do_nothing_id not in new_gen_id_to_bf_mapping.keys(), self.own_do_nothing_id
        assert self.own_others_id not in old_gen_id_to_bf_mapping.keys(), self.own_others_id
        assert self.own_others_id not in new_gen_id_to_bf_mapping.keys(), self.own_others_id

        default_scale_weight: float = self._get_default_scale_weight(b_flavour=b_flavour)
        old_gen_id_to_bf_mapping.update({self.own_do_nothing_id: 1.0})
        new_gen_id_to_bf_mapping.update({self.own_do_nothing_id: 1.0})
        old_gen_id_to_bf_mapping.update({self.own_others_id: 1.0})
        new_gen_id_to_bf_mapping.update({self.own_others_id: default_scale_weight})

        b1_gen_ids: np.ndarray = df[self._own_b1_gen_id_col].values
        b2_gen_ids: np.ndarray = df[self._own_b2_gen_id_col].values

        b1_old_bf: np.ndarray = np.vectorize(old_gen_id_to_bf_mapping.__getitem__)(b1_gen_ids)
        b2_old_bf: np.ndarray = np.vectorize(old_gen_id_to_bf_mapping.__getitem__)(b2_gen_ids)
        b1_new_bf: np.ndarray = np.vectorize(new_gen_id_to_bf_mapping.__getitem__)(b1_gen_ids)
        b2_new_bf: np.ndarray = np.vectorize(new_gen_id_to_bf_mapping.__getitem__)(b2_gen_ids)

        b1_bf_weight: np.ndarray = b1_new_bf / b1_old_bf
        b2_bf_weight: np.ndarray = b2_new_bf / b2_old_bf

        if data_type in self.double_counting_data_types:
            # Reweight all gen_ids but the one matching the data_type;
            # If the gen_id of the data type appears for both B mesons, apply the weight once!
            dt_gen_id: int = self.double_counting_cases_map[data_type]

            b1_bf_weight[(b1_gen_ids == dt_gen_id) & (b2_gen_ids != dt_gen_id)] = 1.0
            b2_bf_weight[b2_gen_ids == dt_gen_id] = 1.0

        weight_values: np.ndarray = b1_bf_weight * b2_bf_weight

        assert not np.any(np.isnan(weight_values)), np.sum(np.isnan(weight_values))
        return weight_values

    @staticmethod
    def get_b_flavour_from_data_type(data_type: str) -> str:
        flavour: str = get_b_flavour_from_data_type(data_type=data_type)
        if flavour == "Charged":
            return "b_charged"
        elif flavour == "Mixed":
            return "b_mixed"
        else:
            raise RuntimeError(f"B meson flavour {flavour} is not known...")

    @staticmethod
    def add_neutral_weights(df: pd.DataFrame) -> None:
        assert BMesonBFReweighter.b_bf_weight_col_name not in df.columns, BMesonBFReweighter.b_bf_weight_col_name
        assert all(col in df.columns for col in BMesonBFReweighter._required_cols)

        df.loc[:, BMesonBFReweighter.b_bf_weight_col_name] = 1.0

    @staticmethod
    def _get_neutral_weights(df: pd.DataFrame) -> np.ndarray:
        assert BMesonBFReweighter.b1_gen_id_col_name in df.columns, BMesonBFReweighter.b1_gen_id_col_name
        neutral_weights: np.ndarray = np.ones_like(df[BMesonBFReweighter.b1_gen_id_col_name].values)
        assert len(neutral_weights) == len(np.unique(df.index)), (len(neutral_weights), len(np.unique(df.index)))
        return neutral_weights

    def _add_special_gen_ids(self, df: pd.DataFrame) -> None:
        assert self._own_b1_gen_id_col not in df.columns, self._own_b1_gen_id_col
        assert self._own_b2_gen_id_col not in df.columns, self._own_b2_gen_id_col

        df.loc[:, self._own_b1_gen_id_col] = df[self.b1_gen_id_col_name].values
        df.loc[:, self._own_b2_gen_id_col] = df[self.b2_gen_id_col_name].values

        _b1_is_dss_1pi: pd.Series = self.is_dstst_to_one_pion(df=df, b_index=0)
        _b2_is_dss_1pi: pd.Series = self.is_dstst_to_one_pion(df=df, b_index=1)

        _b1_is_done_regular: pd.Series = (df[self.b1_gen_id_col_name].isin(self.dstst_gen_ids)) & _b1_is_dss_1pi
        _b2_is_done_regular: pd.Series = (df[self.b2_gen_id_col_name].isin(self.dstst_gen_ids)) & _b2_is_dss_1pi
        df.loc[_b1_is_done_regular, self._own_b1_gen_id_col] *= 10
        df.loc[_b2_is_done_regular, self._own_b2_gen_id_col] *= 10

        _b1_is_dss_gap: pd.Series = (df[self.b1_gen_id_col_name].isin(self.dstst_gen_ids)) & (~_b1_is_dss_1pi)
        _b2_is_dss_gap: pd.Series = (df[self.b2_gen_id_col_name].isin(self.dstst_gen_ids)) & (~_b2_is_dss_1pi)
        df.loc[_b1_is_dss_gap, self._own_b1_gen_id_col] *= 100
        df.loc[_b2_is_dss_gap, self._own_b2_gen_id_col] *= 100

        _b1_reweighted: pd.Series = df[self._own_b1_gen_id_col].isin(self.all_reweight_ids)
        _b2_reweighted: pd.Series = df[self._own_b2_gen_id_col].isin(self.all_reweight_ids)

        _b1_do_not_reweighted: pd.Series = df[self._own_b1_gen_id_col].isin(self._not_reweighted_keys)
        _b2_do_not_reweighted: pd.Series = df[self._own_b2_gen_id_col].isin(self._not_reweighted_keys)

        df.loc[_b1_do_not_reweighted, self._own_b1_gen_id_col] = self.own_do_nothing_id
        df.loc[_b2_do_not_reweighted, self._own_b2_gen_id_col] = self.own_do_nothing_id

        df.loc[~(_b1_do_not_reweighted | _b1_reweighted), self._own_b1_gen_id_col] = self.own_others_id
        df.loc[~(_b2_do_not_reweighted | _b2_reweighted), self._own_b2_gen_id_col] = self.own_others_id

    def _remove_special_gen_ids(self, df: pd.DataFrame) -> None:
        assert self._own_b1_gen_id_col in df.columns, self._own_b1_gen_id_col
        assert self._own_b2_gen_id_col in df.columns, self._own_b2_gen_id_col

        del df[self._own_b1_gen_id_col]
        del df[self._own_b2_gen_id_col]

        assert self._own_b1_gen_id_col not in df.columns, self._own_b1_gen_id_col
        assert self._own_b2_gen_id_col not in df.columns, self._own_b2_gen_id_col

    def _get_default_scale_weight(self, b_flavour: str) -> float:
        if not self.rescale_others:
            return 1.0

        if b_flavour == "b_charged":
            return ChargedBBFComparison().get_charged_correction_factor()
        elif b_flavour == "b_mixed":
            return MixedBBFComparison().get_mixed_correction_factor()
        else:
            raise KeyError(f"The flavour {b_flavour} is not valid, must be one of [b_charged, b_mixed].")

    @staticmethod
    def is_dstst_to_one_pion(
        df: pd.DataFrame,
        b_index: int,
    ) -> pd.Series:
        assert b_index in (0, 1), b_index
        d4_pdgs: Tuple[int, ...] = (22, -999)
        pion_pdgs: Tuple[int, ...] = (111, 211)

        _is_dstst_to_pion: pd.Series = abs(df[get_second_daughter_pdg_col(b_index=b_index)]).isin(pion_pdgs)
        _has_two_dstst_daughters: pd.Series = df[get_third_daughter_pdg_col(b_index=b_index)].isin(d4_pdgs)
        return _is_dstst_to_pion & _has_two_dstst_daughters
