import os
import numpy as np
import pandas as pd
import b2luigi as luigi
import matplotlib.pyplot as plt

from luigi.util import inherits
from typing import Optional, Dict, Generator, Tuple, List

from templatefitter.plotter import plot_style
from templatefitter.minimizer import MinimizeResult
from templatefitter.plotter.plot_utilities import export

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.fitting.fit_case import TemplateSetup
from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.mc_matching.old_mc_matching_info import OldMCMatchingIDsQueryInfo

from rdstar.offline_analysis.dss_fitting.systematics import DSSFFSystematics

from rdstar.offline_analysis.dss_fitting.tasks.fitting_task import FittingTask
from rdstar.offline_analysis.dss_fitting.tasks.preparation_task import CopyToSSDTask
from rdstar.offline_analysis.dss_fitting.tasks.splitter_task import SampleSplitterTask
from rdstar.offline_analysis.dss_fitting.setup.fitting_setup import FittingSetupCollection


__all__ = [
    "FFRatioCalculationTask",
]


logger = log_to_default(name=__name__)


YieldDictType = Dict[int, Dict[str, float]]


@inherits(SampleSplitterTask)
class FFRatioCalculationTask(utilities.BaseOfflineTask):
    decay_mode_column_name: str = Cols.decay_mode_column_name

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_dir = os.path.join(self.working_dir, "ff_results", self.fitting_setup_name)

        self.fitting_setup = FittingSetupCollection.available_setups[self.fitting_setup_name]

        self.fit_case = self.fitting_setup.fit_case_dict.pivot_fit_case
        self.fit_output_tag = self.fit_case.output_tag

        self.yield_parameters = self.fitting_setup.ratio_yield_parameters
        components = self.fit_case.all_components
        assert components is not None
        self.yield_parameter_labels = (
            self.fit_case.plot_labels
            if self.fit_case.plot_labels
            else {comp: OldMCMatchingIDsQueryInfo.get_sig_id_name_to_latex_label()[comp] for comp in components}
        )

        self.variable = self.fitting_setup.fit_in_bins_of

        self._all_mode_ids: Optional[List[int]] = None
        self._input_dict: Optional[Dict[str, PathType]] = None

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    @property
    def all_mode_ids(self) -> List[int]:
        if self._all_mode_ids is None:
            self._all_mode_ids = [
                d_id for d_ids in [d_ids for _, d_ids in self.fit_case.reco_mode_infos] for d_id in d_ids
            ]
        return self._all_mode_ids

    def generate_output_dict(self) -> Dict[str, PathType]:
        output_dict: Dict[str, PathType] = dict()
        for param in self.yield_parameters:
            output_key = f"{self.fitting_setup_name}_{param}_ratio"
            output_path = os.path.join(self.base_dir, f"{output_key}.pdf")
            output_dict.update({output_key: output_path})
        return output_dict

    def load_fit_result(self, bin_number: int) -> MinimizeResult:
        fit_result_path = self.input_dict[f"{self.fit_output_tag}_FitResult_{self.fitting_setup_name}_bin_{bin_number}"]
        return MinimizeResult.load_from(path=fit_result_path)

    def generate_yield_dict(self) -> YieldDictType:
        yield_dict: YieldDictType = dict()
        for bin_number in range(0, self.variable.n_bins):
            fit_result = self.load_fit_result(bin_number=bin_number)

            bin_dict: Dict[str, float] = dict()
            for yield_parameter in self.yield_parameters:
                yield_in_bin = fit_result.params.get_param_value(f"yield_{yield_parameter}")
                error_in_bin = fit_result.params.get_param_error(f"yield_{yield_parameter}")

                bin_dict.update({f"{yield_parameter}_value": yield_in_bin, f"{yield_parameter}_error": error_in_bin})
            yield_dict.update({bin_number: bin_dict})
        return yield_dict

    @staticmethod
    def get_yield_in_bin(yield_dict: YieldDictType, yield_parameter: str, bin_number: int) -> float:
        yield_in_bin = yield_dict[bin_number][f"{yield_parameter}_value"]
        return yield_in_bin

    @staticmethod
    def get_error_in_bin(yield_dict: YieldDictType, yield_parameter: str, bin_number: int) -> float:
        error_in_bin = yield_dict[bin_number][f"{yield_parameter}_error"]
        return error_in_bin

    def get_yields(self, yield_dict: YieldDictType, yield_parameter) -> np.ndarray:
        bins = self.variable.n_bins
        yields = np.array(
            [
                self.get_yield_in_bin(yield_dict=yield_dict, yield_parameter=yield_parameter, bin_number=i)
                for i in range(0, bins)
            ]
        )
        return yields

    def get_errors(self, yield_dict: YieldDictType, yield_parameter) -> np.ndarray:
        bins = self.variable.n_bins
        errors = np.array(
            [
                self.get_error_in_bin(yield_dict=yield_dict, yield_parameter=yield_parameter, bin_number=i)
                for i in range(0, bins)
            ]
        )
        return errors

    @staticmethod
    def generate_cov_matrix(errors: np.ndarray) -> np.ndarray:
        dim = len(errors)
        cov = np.zeros((dim, dim))
        for i in range(0, dim):
            cov[i][i] = errors[i] ** 2

        return cov

    @staticmethod
    def generate_jacobi_matrix(yields: np.ndarray) -> np.ndarray:
        dim = len(yields)
        jacobi = np.zeros((dim, dim))
        for i in range(0, dim):
            for j in range(0, dim):
                if i == j:
                    jacobi[i][j] = 1 / np.sum(yields) - yields[i] / (np.sum(yields) ** 2)
                else:
                    jacobi[i][j] = -yields[i] / (np.sum(yields) ** 2)

        return jacobi

    def calculate_ratio_errors(
        self,
        yields: np.ndarray,
        errors: np.ndarray,
    ) -> np.ndarray:
        assert yields.shape == errors.shape, (yields.shape, errors.shape)
        dim = len(yields)
        cov = self.generate_cov_matrix(errors=errors)
        jacobi = self.generate_jacobi_matrix(yields=yields)
        dov = np.matmul(np.matmul(jacobi, cov), np.transpose(jacobi))

        ratio_errors = [np.sqrt(dov[i][i]) for i in range(0, dim)]
        return np.array(ratio_errors)

    def calculate_ratios(
        self,
        yields: np.ndarray,
        errors: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray]:
        assert len(yields) == len(errors), (len(yields), len(errors))
        dim = len(yields)

        ratios = np.array([(yields[i] / np.sum(yields)) for i in range(0, dim)])
        ratio_errors = self.calculate_ratio_errors(yields=yields, errors=errors)

        return ratios, ratio_errors

    def get_mc_yields(self, df: pd.DataFrame, yield_parameter: str) -> np.ndarray:
        df_mc = df.loc[df[self.decay_mode_column_name].isin(self.all_mode_ids)]

        assert isinstance(self.fit_case.template_setup, TemplateSetup)
        if isinstance(self.fit_case.template_setup.template_setup, Dict):
            sig_ids = tuple(
                [
                    self.fit_case.sig_queries[comp].id
                    for comp in self.fit_case.template_setup.template_setup[yield_parameter]
                ]
            )
        else:
            sig_ids = (self.fit_case.sig_queries[yield_parameter].id,)
        sig_id_condition = df_mc[self.fit_case.sig_id_column_name].isin(sig_ids)
        df_mc = df_mc.loc[sig_id_condition]

        scopes = self.fitting_setup.fit_in_bins_of_scopes

        mc_yields = []
        for i in range(0, self.variable.n_bins):
            df_cut = df_mc.query(f"{scopes[i]} <= {self.variable.df_label} < {scopes[i + 1]}")
            mc_yields.append(np.sum(df_cut[self.fit_case.weight_column_name]))

        return np.array(mc_yields)

    def calculate_systematic_error(self, df: pd.DataFrame, mc_yields: np.ndarray) -> np.ndarray:
        df_mc = df.loc[df[self.decay_mode_column_name].isin(self.all_mode_ids)]
        dim = len(mc_yields)
        dss_ff_systematics = DSSFFSystematics(df=df_mc, fitting_setup=self.fitting_setup)

        cov = dss_ff_systematics.get_total_cov()
        jacobi = self.generate_jacobi_matrix(yields=mc_yields)
        dov = np.matmul(np.matmul(jacobi, cov), np.transpose(jacobi))
        ratio_ff_errors = np.array([np.sqrt(dov[i][i]) for i in range(0, dim)])

        return ratio_ff_errors

    def calculate_chi_square(
        self,
        yields: np.ndarray,
        errors: np.ndarray,
        mc_yields: np.ndarray,
        mc_errors: np.ndarray,
        ff_cov: np.ndarray,
    ) -> float:
        assert len(yields) == len(mc_yields) == len(errors) == len(mc_errors)

        ratios, ratio_errors = self.calculate_ratios(yields=yields, errors=errors)
        mc_ratios, mc_ratio_errors = self.calculate_ratios(yields=mc_yields, errors=mc_errors)

        ratio_diff = (ratios - mc_ratios)[:-1]

        cov = self.generate_cov_matrix(errors=errors)
        jacobi = self.generate_jacobi_matrix(yields=yields)
        dov = np.matmul(np.matmul(jacobi, cov), np.transpose(jacobi))

        mc_cov = self.generate_cov_matrix(errors=mc_errors)
        mc_jacobi = self.generate_jacobi_matrix(yields=mc_yields)
        mc_dov = np.matmul(np.matmul(mc_jacobi, mc_cov), np.transpose(mc_jacobi))

        ff_dov = np.matmul(np.matmul(mc_jacobi, ff_cov), np.transpose(mc_jacobi))

        total_dov = (dov + mc_dov + ff_dov)[:-1, :-1]
        inv_total_dov = np.linalg.inv(total_dov)

        chi_2 = np.matmul(np.matmul(ratio_diff, inv_total_dov), np.transpose(ratio_diff))

        return float(chi_2)

    def run_calculation(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dict = self.generate_output_dict()

        if not dry_run:
            yield_dict = self.generate_yield_dict()

            bins = self.variable.n_bins
            scopes = self.fitting_setup.fit_in_bins_of_scopes
            bin_centers = [scopes[i + 1] - ((scopes[i + 1] - scopes[i]) / 2) for i in range(0, bins)]
            bin_widths = [scopes[i + 1] - scopes[i] for i in range(0, bins)]
            assert len(bin_centers) == len(scopes) - 1
            assert len(bin_widths) == len(scopes) - 1

            df_mc = pd.read_hdf(self.input_dict[f"combined_mc_prepared_{self.selection}"])
            assert isinstance(df_mc, pd.DataFrame)

            for yield_parameter in self.yield_parameters:
                plot_style.set_matplotlibrc_params()
                fig, ax = plt.subplots()

                mc_yields = self.get_mc_yields(df=df_mc, yield_parameter=yield_parameter)
                mc_errors = np.array([np.sqrt(mc_yields[i]) for i in range(0, bins)])
                mc_ratios, mc_ratio_errors = self.calculate_ratios(yields=mc_yields, errors=mc_errors)

                mc_sys_error = self.calculate_systematic_error(df=df_mc, mc_yields=mc_yields)

                logger.debug(f"FF errors: {mc_sys_error}")
                plot_colors = self.fit_case.plot_colors
                assert isinstance(plot_colors, dict)
                ax.bar(
                    x=bin_centers,
                    height=2 * (mc_ratio_errors + mc_sys_error),
                    width=bin_widths,
                    bottom=(mc_ratios / np.array(bin_widths) - (mc_ratio_errors + mc_sys_error)),
                    color=plot_colors[yield_parameter],
                    alpha=0.4,
                    label="FF Variation",
                )

                ax.bar(
                    x=bin_centers,
                    height=2 * mc_ratio_errors,
                    width=bin_widths,
                    bottom=(mc_ratios / np.array(bin_widths) - mc_ratio_errors),
                    color=plot_colors[yield_parameter],
                    alpha=1.0,
                    label="MC",
                )

                yields = self.get_yields(yield_dict=yield_dict, yield_parameter=yield_parameter)
                errors = self.get_errors(yield_dict=yield_dict, yield_parameter=yield_parameter)
                ratios, ratio_errors = self.calculate_ratios(yields=yields, errors=errors)

                ax.errorbar(
                    x=bin_centers,
                    y=ratios / np.array(bin_widths),
                    yerr=ratio_errors,
                    xerr=np.array(bin_widths) / 2,
                    ls="",
                    marker=".",
                    color="black",
                    capsize=3,
                    label="Data",
                )

                ax.set_xlabel(f"{self.variable.variable_name} in {self.variable.unit}", plot_style.xlabel_pos)
                ax.set_ylabel(
                    r"$\mathcal{R} = (N_{D^{**}}^{bin} / N_{D^{**}}^{total})$"
                    + rf"({self.variable.unit} / $\Delta${self.variable.variable_name})",
                    plot_style.ylabel_pos,
                )

                ax.set_title(self.fit_case.plot_info_text, loc="left")
                ax.set_title(self.yield_parameter_labels[yield_parameter], loc="right")

                dss_ff_systematics = DSSFFSystematics(df=df_mc, fitting_setup=self.fitting_setup)
                ff_cov = dss_ff_systematics.get_total_cov()
                chi_2 = self.calculate_chi_square(
                    yields=yields, errors=errors, mc_yields=mc_yields, mc_errors=mc_errors, ff_cov=ff_cov
                )
                text = r"$\chi^2 \mathrm{-Value} = $" + f"{round(chi_2, 2)}"
                ax.text(0.1, 0.1, text, horizontalalignment="left", verticalalignment="center", transform=ax.transAxes)
                ax.legend(loc=0)

                output_key = f"{self.fitting_setup_name}_{yield_parameter}_ratio"
                export(fig=fig, filename=f"{output_key}", target_dir=self.base_dir)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_calculation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.run_calculation(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})
        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for bin_number in range(0, self.fitting_setup.fit_in_bins_of.n_bins):
            yield self.clone(FittingTask, bin_number=bin_number)

        yield self.clone(CopyToSSDTask)
