"""
Tools to calculate effect of systematics on efficiencies

Felix Metzner 2021
"""

import os
import json
import copy
import numpy as np
import pandas as pd

from collections.abc import Mapping
from dataclasses import dataclass, asdict
from typing import Union, Optional, Tuple, List, Sequence, Dict, Generator, Iterator, Any, ClassVar

from rdstar.utilities import PathType
from rdstar.offline_analysis.systematics import SysColsInfo, multiplicative_eff_sys_info_names
from rdstar.offline_analysis.weight_management.weight_information import WeightTypeCollection, WeightManager


__all__ = [
    "EfficiencySysInfo",
    "EfficiencySysInfoContainer",
    "EfficiencySysCalculator",
]


@dataclass(frozen=True)
class EfficiencySysInfo:
    sys_name: str
    rms_d: float
    sig_ids: Optional[Tuple[int, ...]]
    all_sig_ids: Tuple[int, ...]
    nominal_sum_of_weights: float
    varied_sum_of_weights: Tuple[float, ...]
    nominal_col_name: str
    varied_col_names: Tuple[str, ...]

    def __post_init__(self) -> None:
        assert self.sig_ids is None or all(v in self.all_sig_ids for v in self.sig_ids), (self.sig_ids, self.all_sig_ids)
        assert len(self.varied_col_names) == len(self.varied_sum_of_weights), (
            len(self.varied_col_names),
            len(self.varied_sum_of_weights),
        )

    @property
    def systematic_uncert_value(self) -> float:
        return self.rms_d

    @property
    def relative_rms_dev(self) -> float:
        n_events_dev: np.ndarray = np.array(self.varied_sum_of_weights) - self.nominal_sum_of_weights
        relative_n_events_dev: np.ndarray = n_events_dev / self.nominal_sum_of_weights
        return np.sqrt(relative_n_events_dev.dot(relative_n_events_dev) / relative_n_events_dev.size)

    @property
    def rms_dev_all(self) -> float:
        n_events_dev: np.ndarray = self.all_sum_of_weights - self.nominal_sum_of_weights
        return np.sqrt(n_events_dev.dot(n_events_dev) / n_events_dev.size)

    @property
    def relative_rms_dev_all(self) -> float:
        n_events_dev: np.ndarray = self.all_sum_of_weights - self.nominal_sum_of_weights
        relative_n_events_dev: np.ndarray = n_events_dev / self.nominal_sum_of_weights
        return np.sqrt(relative_n_events_dev.dot(relative_n_events_dev) / relative_n_events_dev.size)

    @property
    def all_sum_of_weights(self) -> np.ndarray:
        return np.array([self.nominal_sum_of_weights, *self.varied_sum_of_weights])

    @property
    def deviation_of_all_sum_of_weights(self) -> np.ndarray:
        return self.all_sum_of_weights - self.nominal_sum_of_weights

    @property
    def relative_deviation_of_all_sum_of_weights(self) -> np.ndarray:
        return self.deviation_of_all_sum_of_weights / self.nominal_sum_of_weights

    def as_dict(self) -> Dict[str, Any]:
        return asdict(self)

    @staticmethod
    def from_dict(input_dict: Dict[str, Any]) -> "EfficiencySysInfo":
        return EfficiencySysInfo(**{k: tuple(v) if isinstance(v, list) else v for k, v in input_dict.items()})


class EfficiencySysInfoContainer(Mapping):
    relevant_sys_infos: ClassVar[Tuple[str, ...]] = copy.copy(multiplicative_eff_sys_info_names)
    all_sys_combined_key: ClassVar[str] = "AllSystematics"
    all_sys_combined_label: ClassVar[str] = "All Systematics"

    own_eff_weight_column_name: ClassVar[str] = "tmp_efficiency_sys_event_weight"

    def __init__(
        self,
        sig_ids: Sequence[Union[int, Tuple[int, ...]]],
        sys_infos: Sequence[SysColsInfo],
        weight_col_name: Optional[str],
        sig_id_column_name: str,
        use_only_given_sig_ids: bool = False,
        use_new_fei_calibration_weights: bool = False,
        additional_scaling_weight_columns: Optional[Tuple[str, ...]] = None,
    ) -> None:
        assert isinstance(sig_ids, (tuple, list)), (type(sig_ids), sig_ids)
        assert all(isinstance(v, (int, tuple)) for v in sig_ids), (
            [type(v) for v in sig_ids if not isinstance(v, (int, tuple))],
            sig_ids,
        )
        assert all(all(isinstance(vv, int) for vv in v) for v in sig_ids if isinstance(v, tuple)), (
            [type(vv) for v in sig_ids if isinstance(v, tuple) for vv in v if not isinstance(vv, int)],
            sig_ids,
        )

        assert isinstance(sys_infos, (tuple, list)), (type(sys_infos), sys_infos)

        self._sig_ids: Tuple[Union[int, Tuple[int, ...]], ...] = tuple(sig_ids)
        self._sys_infos: Tuple[SysColsInfo, ...] = tuple(sys_infos)

        self._weight_col_name: str = self.own_eff_weight_column_name if weight_col_name is None else weight_col_name
        self._sig_id_column_name: str = sig_id_column_name

        self._use_only_known_sig_ids: bool = use_only_given_sig_ids
        self._use_new_fei_calibration_weights: bool = use_new_fei_calibration_weights
        self._additional_scaling_weight_columns: Optional[Tuple[str, ...]] = additional_scaling_weight_columns

        self._sys_infos_by_sys_and_sig_id_map: Dict[Tuple[str, Optional[Tuple[int, ...]]], EfficiencySysInfo] = dict()

    def add_own_weight_column(self, df: pd.DataFrame) -> None:
        if self._weight_col_name != self.own_eff_weight_column_name:
            if self._weight_col_name not in df.columns:
                raise KeyError(
                    f"Provided weight column of name {self._weight_col_name} is not available in the to be evaluated data frame."
                )
        else:
            assert self._weight_col_name not in df.columns, self._weight_col_name

            eff_weight_cols: List[str] = list(WeightTypeCollection.efficiency_weight_columns)

            if self._use_new_fei_calibration_weights:
                eff_weight_cols.append(WeightManager.used_fei_calib_col)

            if self._additional_scaling_weight_columns is not None:
                assert all(c in df.columns for c in self._additional_scaling_weight_columns), [
                    c for c in self._additional_scaling_weight_columns if c not in df.columns
                ]
                eff_weight_cols.extend(list(self._additional_scaling_weight_columns))

            assert all(c in df.columns for c in eff_weight_cols), [c for c in eff_weight_cols if c not in df.columns]
            assert len(set(eff_weight_cols)) == len(eff_weight_cols), eff_weight_cols

            df.loc[:, self.own_eff_weight_column_name] = df[eff_weight_cols].product(axis=1)

    def remove_own_weight_column(self, df: pd.DataFrame) -> None:
        if self._weight_col_name == self.own_eff_weight_column_name and self._weight_col_name in df.columns:
            del df[self._weight_col_name]

    def fill(self, df: pd.DataFrame) -> None:
        self.add_own_weight_column(df=df)

        for sys_info in self._sys_infos:
            eff_sys_calculator = EfficiencySysCalculator(
                sys_info=sys_info,
                weight_col_name=self._weight_col_name,
                sig_id_column_name=self._sig_id_column_name,
                use_only_given_sig_ids=self._use_only_known_sig_ids,
            )
            for eff_sys_info in eff_sys_calculator.calculate_for(df=df, sig_ids=self._sig_ids):
                _current_key: Tuple[str, Optional[Tuple[int, ...]]] = (sys_info.name, eff_sys_info.sig_ids)
                assert _current_key not in self._sys_infos_by_sys_and_sig_id_map, _current_key

                self._sys_infos_by_sys_and_sig_id_map[_current_key] = copy.copy(eff_sys_info)

        self.remove_own_weight_column(df=df)

    def get_all_systematics_for(self, sig_ids: Optional[Union[int, Tuple[int, ...]]]) -> Dict[str, EfficiencySysInfo]:
        this_sig_id: Optional[Tuple[int, ...]] = (sig_ids,) if isinstance(sig_ids, int) else sig_ids
        sys_dict: Dict[str, EfficiencySysInfo] = {
            sys_name: v
            for (sys_name, sig_id_key), v in self._sys_infos_by_sys_and_sig_id_map.items()
            if sig_id_key == this_sig_id
        }
        assert all(k == v.sys_name for k, v in sys_dict.items()), [
            (k, v.sys_name) for k, v in sys_dict.items() if k != v.sys_name
        ]
        return sys_dict

    def get_sys_covariance_matrix_for(
        self,
        sig_ids_a: Optional[Tuple[int, ...]],
        sig_ids_b: Optional[Tuple[int, ...]],
        additional_sys_covariance_mat_dict: Optional[Dict[str, np.ndarray]] = None,
    ) -> Dict[str, np.ndarray]:
        assert all(isinstance(v, tuple) or v is None for v in [sig_ids_a, sig_ids_b]), (sig_ids_a, sig_ids_b)
        assert sig_ids_a is None or all(isinstance(v, int) for v in sig_ids_a), sig_ids_a
        assert sig_ids_b is None or all(isinstance(v, int) for v in sig_ids_b), sig_ids_b

        sys_a_dict: Dict[str, EfficiencySysInfo] = self.get_all_systematics_for(sig_ids=sig_ids_a)
        sys_b_dict: Dict[str, EfficiencySysInfo] = self.get_all_systematics_for(sig_ids=sig_ids_b)
        assert set(sys_a_dict.keys()) == set(sys_b_dict.keys()), (sys_a_dict.keys(), sys_b_dict.keys())

        sys_cov_dict: Dict[str, np.ndarray] = dict()
        for sys_name in sys_a_dict.keys():
            sys_cov_matrix = np.cov(
                sys_a_dict[sys_name].all_sum_of_weights,
                sys_b_dict[sys_name].all_sum_of_weights,
            )
            sys_cov_dict.update({sys_name: sys_cov_matrix})

        if additional_sys_covariance_mat_dict is not None:
            assert all(k not in sys_cov_dict for k in additional_sys_covariance_mat_dict), (
                additional_sys_covariance_mat_dict.keys(),
                sys_cov_dict.keys(),
            )
            assert all(isinstance(v, np.ndarray) for v in additional_sys_covariance_mat_dict.values()), [
                (k, type(v)) for k, v in additional_sys_covariance_mat_dict.items()
            ]
            assert all(v.shape == (2, 2) for v in additional_sys_covariance_mat_dict.values())
            sys_cov_dict.update(additional_sys_covariance_mat_dict)

        assert self.all_sys_combined_key not in sys_cov_dict, (self.all_sys_combined_key, sys_cov_dict.keys())
        sum_of_sys_cov_matrices: np.ndarray = np.sum([v for v in sys_cov_dict.values()], axis=0)
        sys_cov_dict.update({self.all_sys_combined_key: sum_of_sys_cov_matrices})
        return sys_cov_dict

    def __getitem__(self, item: Tuple[str, Optional[Union[int, Tuple[int, ...]]]]) -> EfficiencySysInfo:
        sys_name: str = item[0]
        sig_id_info: Optional[Union[int, Tuple[int, ...]]] = item[1]
        if isinstance(sig_id_info, tuple):
            _item: Tuple[str, Optional[Tuple[int, ...]]] = (sys_name, sig_id_info)
        elif isinstance(sig_id_info, int):
            _item = (sys_name, (sig_id_info,))
        elif sig_id_info is None:
            _item = (sys_name, sig_id_info)
        else:
            raise KeyError(f"Unknown key format {item} of [{type(item[0]).__name__}, {type(item[1]).__name__}]")

        return self._sys_infos_by_sys_and_sig_id_map[_item]

    def __iter__(self) -> Iterator[Tuple[str, Optional[Tuple[int, ...]]]]:
        return iter(self._sys_infos_by_sys_and_sig_id_map)

    def __len__(self) -> int:
        return len(self._sys_infos_by_sys_and_sig_id_map)

    @staticmethod
    def get_relevant_sys_infos(sys_cols=Sequence[SysColsInfo]) -> Tuple[SysColsInfo, ...]:
        return tuple(s for s in sys_cols if s.name in EfficiencySysInfoContainer.relevant_sys_infos)

    def encode_content_for_json(self) -> Dict[str, Any]:
        assert all(isinstance(k, tuple) for i, (k, v) in enumerate(self._sys_infos_by_sys_and_sig_id_map.items()))
        assert all(
            isinstance(v, EfficiencySysInfo) for i, (k, v) in enumerate(self._sys_infos_by_sys_and_sig_id_map.items())
        )
        return {
            str(key_index): {"key": k, "value": v.as_dict()}
            for key_index, (k, v) in enumerate(self._sys_infos_by_sys_and_sig_id_map.items())
        }

    @staticmethod
    def decode_content_for_json(
        encoded_content: Dict[str, Any]
    ) -> Dict[Tuple[str, Optional[Tuple[int, ...]]], EfficiencySysInfo]:
        assert all(isinstance(k, str) for k in encoded_content), encoded_content.keys()
        assert all(isinstance(v, dict) for v in encoded_content.values()), [type(v) for v in encoded_content.values()]
        assert all(len(v) == 2 for v in encoded_content.values()), [(type(v), len(v)) for v in encoded_content.values()]
        assert all(all(k in v.keys() for k in ["key", "value"]) for v in encoded_content.values()), [
            v.keys() for v in encoded_content.values()
        ]

        return {
            EfficiencySysInfoContainer.decode_key_tuple_from_json(key_input=v["key"]): EfficiencySysInfo.from_dict(
                input_dict=v["value"]
            )
            for v in encoded_content.values()
        }

    @staticmethod
    def decode_key_tuple_from_json(key_input: Any) -> Tuple[str, Optional[Tuple[int, ...]]]:
        assert isinstance(key_input, (list, tuple)), (type(key_input), key_input)
        assert len(key_input) == 2, (key_input, len(key_input))
        _sys_name = key_input[0]
        _sig_id_info = key_input[1]
        assert isinstance(_sys_name, str), (type(_sys_name), _sys_name)
        sys_name: str = _sys_name
        assert _sig_id_info is None or (
            isinstance(_sig_id_info, (tuple, list)) and all(isinstance(v, int) for v in _sig_id_info)
        ), _sig_id_info
        sig_id_info: Optional[Tuple[int, ...]] = _sig_id_info if _sig_id_info is None else tuple(_sig_id_info)

        return sys_name, sig_id_info

    def as_dict(self) -> Dict[str, Any]:
        return {
            "sig_ids": self._sig_ids,
            "sys_infos": [v.as_dict() for v in self._sys_infos],
            "weight_col_name": self._weight_col_name,
            "sig_id_column_name": self._sig_id_column_name,
            "use_only_given_sig_ids": self._use_only_known_sig_ids,
            "use_new_fei_calibration_weights": self._use_new_fei_calibration_weights,
            "additional_scaling_weight_columns": self._additional_scaling_weight_columns,
            "sys_infos_by_sys_and_sig_id_map": self.encode_content_for_json(),
        }

    @staticmethod
    def from_dict(input_dict: Dict[str, Any]) -> "EfficiencySysInfoContainer":
        _scl_cols: Optional[List[str]] = input_dict["additional_scaling_weight_columns"]

        new_instance = EfficiencySysInfoContainer(
            sig_ids=input_dict["sig_ids"],
            sys_infos=tuple(SysColsInfo.from_dict(v) for v in input_dict["sys_infos"]),
            weight_col_name=input_dict["weight_col_name"],
            sig_id_column_name=input_dict["sig_id_column_name"],
            use_only_given_sig_ids=input_dict["use_only_given_sig_ids"],
            use_new_fei_calibration_weights=input_dict["use_new_fei_calibration_weights"],
            additional_scaling_weight_columns=_scl_cols if _scl_cols is None else tuple(_scl_cols),
        )

        content = EfficiencySysInfoContainer.decode_content_for_json(
            encoded_content=input_dict["sys_infos_by_sys_and_sig_id_map"],
        )
        new_instance._sys_infos_by_sys_and_sig_id_map = content
        return new_instance

    def write_to_file(self, target_file_path: PathType) -> None:
        if os.path.exists(target_file_path):
            raise FileExistsError(f"Trying to overwrite {target_file_path}!")

        result_dict: Dict[str, Any] = self.as_dict()

        tmp_target_path = os.path.join(os.path.dirname(target_file_path), f"tmp_{os.path.basename(target_file_path)}")
        with open(tmp_target_path, "w") as output_file:
            json.dump(result_dict, output_file, indent=2)
        os.rename(tmp_target_path, target_file_path)

    @staticmethod
    def read_from_file(file_path: PathType) -> "EfficiencySysInfoContainer":
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"EfficiencySysInfoContainer file {file_path} does not exist!")

        with open(file_path, "r") as in_file:
            input_dict = json.load(in_file)

        return EfficiencySysInfoContainer.from_dict(input_dict=input_dict)


class EfficiencySysCalculator:
    def __init__(
        self,
        sys_info: SysColsInfo,
        weight_col_name: str,
        sig_id_column_name: str,
        use_only_given_sig_ids: bool,
    ) -> None:

        self._sys_info: SysColsInfo = sys_info

        self._weight_col_name: str = weight_col_name
        self._sig_id_column_name: str = sig_id_column_name

        self._use_only_known_sig_ids: bool = use_only_given_sig_ids

    @property
    def sig_id_column_name(self) -> str:
        return self._sig_id_column_name

    def calculate_for(
        self,
        df: pd.DataFrame,
        sig_ids: Sequence[Union[int, Tuple[int, ...]]],
    ) -> Generator[EfficiencySysInfo, None, None]:
        assert self._weight_col_name in df.columns, self._weight_col_name
        assert self.sig_id_column_name in df.columns, self.sig_id_column_name

        nominal_weight_col, var_cols = self.get_nominal_and_variation_cols(df=df)  # type: str, Tuple[str, ...]
        sig_id_tuples: Tuple[Tuple[int, ...], ...] = tuple(v if isinstance(v, tuple) else (v,) for v in sig_ids)
        all_sig_ids: Tuple[int, ...] = tuple(vv for v in sig_id_tuples for vv in v)
        assert len(all_sig_ids) == len(set(all_sig_ids)), (len(all_sig_ids), len(set(all_sig_ids)), all_sig_ids, sig_ids)

        for sig_id_tuple in [None, *sig_id_tuples]:
            nominal_count, varied_counts = self.calculate_sum_of_weights_for(
                df=df,
                sig_id_tuple=sig_id_tuple,
                all_sig_ids=all_sig_ids,
                nominal_weight_col=nominal_weight_col,
                variation_weight_cols=var_cols,
            )  # type:  float, Tuple[float, ...]
            n_events_dev: np.ndarray = np.array(varied_counts) - nominal_count
            n_events_rms_d = np.sqrt(n_events_dev.dot(n_events_dev) / n_events_dev.size)

            yield EfficiencySysInfo(
                sys_name=self._sys_info.name,
                rms_d=n_events_rms_d,
                sig_ids=sig_id_tuple,
                all_sig_ids=all_sig_ids,
                nominal_sum_of_weights=nominal_count,
                varied_sum_of_weights=varied_counts,
                nominal_col_name=nominal_weight_col,
                varied_col_names=var_cols,
            )

    def calculate_sum_of_weights_for(
        self,
        df: pd.DataFrame,
        sig_id_tuple: Optional[Tuple[int, ...]],
        all_sig_ids: Sequence[int],
        nominal_weight_col: str,
        variation_weight_cols: Tuple[str, ...],
    ) -> Tuple[float, Tuple[float, ...]]:
        if sig_id_tuple is None:
            if not self._use_only_known_sig_ids:
                _df: pd.DataFrame = df
            else:
                _df = df[df[self.sig_id_column_name].isin(all_sig_ids)]
        else:
            _df = df[df[self.sig_id_column_name].isin(sig_id_tuple)]

        _nom_col: str = nominal_weight_col

        nominal_event_count: float = _df[self._weight_col_name].sum()
        event_counts: List[float] = list()

        for var_col in variation_weight_cols:
            v: np.ndarray = _df[self._weight_col_name].values / _df[_nom_col].values * _df[var_col].values
            v[np.isnan(v)] = _df[self._weight_col_name].values[np.isnan(v)]
            event_counts.append(v.sum())

        return nominal_event_count, tuple(event_counts)

    def get_nominal_and_variation_cols(self, df: pd.DataFrame) -> Tuple[str, Tuple[str, ...]]:
        nominal_weight_col, var_cols = self._sys_info.get_systematics_input_for_df(df=df)
        assert isinstance(nominal_weight_col, str), (type(nominal_weight_col), nominal_weight_col)
        assert isinstance(var_cols, (tuple, list)), (type(var_cols), var_cols)
        assert all(isinstance(c, str) for c in var_cols), (set([type(c) for c in var_cols]), var_cols)

        assert nominal_weight_col in df.columns, nominal_weight_col
        assert all(c in df.columns for c in var_cols), [c for c in var_cols if c not in df.columns]

        return nominal_weight_col, tuple(var_cols)

    # TODO: For plots: https://seaborn.pydata.org/tutorial/distributions.html
