"""
Base fitter class for extraction of Continuum calibration factors.

Felix Metzner 2023
"""

import copy
import numpy as np
import pandas as pd

from abc import ABC, abstractmethod
from typing import Union, Optional, Tuple, Dict

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.observables.common_observables import Cols

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fitter_base import FitterBase
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import FitSetupInfoContainer

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.fit_setup_utilities import FitSetupTriplet


__all__ = [
    "ContCalFitterBase",
    "ContCalMappingType",
    "SimpleContCalMappingType",
]


logger = log_to_default(name=__name__)


ContCalMappingType = Dict[Union[Tuple[int, ...], Tuple[Tuple[int, ...], Tuple[int, ...]]], int]
SimpleContCalMappingType = Dict[Tuple[int, ...], int]


class ContCalFitterBase(FitterBase, ABC):

    this_fitter_setup: Optional[FitSetupTriplet] = None

    secondary_reco_id_col: Optional[str] = None

    def __init__(
        self,
        *ignore,
        fit_setup_info: FitSetupInfoContainer,
        mc_df: pd.DataFrame,
        data_df: Optional[pd.DataFrame],
        file_dict_dir_path: Optional[PathType] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            *ignore,
            fit_setup_info=fit_setup_info,
            mc_df=mc_df,
            data_df=data_df,
            file_dict_dir_path=file_dict_dir_path,
            **kwargs,
        )

    # region Cost function
    @abstractmethod
    def cost_fcn(self, params: np.ndarray) -> float:
        raise NotImplementedError("The 'cost_fcn' must be implemented in the derived class.")

    @abstractmethod
    def chi2_fcn(self, params: np.ndarray) -> float:
        raise NotImplementedError("The 'chi2_fcn' must be implemented in the derived class.")

    @abstractmethod
    def gaussian_constrains(self, params: np.ndarray) -> float:
        raise NotImplementedError("The 'gaussian_constrains_fcn' must be implemented in the derived class.")

    # endregion

    # region Expected bin count
    def expected_bin_counts(self, params: np.ndarray) -> np.ndarray:
        return np.sum(
            self.get_yields(params=params) * self.local_efficiencies * self.template_shapes,
            axis=1,
        )

    # endregion

    # region Yields relations
    @abstractmethod
    def get_yields(self, params: np.ndarray) -> np.ndarray:
        raise NotImplementedError("The 'get_yields' must be implemented in the derived class.")

    # endregion

    # region Correction application info
    @classmethod
    @abstractmethod
    def get_correction_mapping(cls) -> ContCalMappingType:
        raise NotImplementedError("The 'get_correction_mapping' must be implemented in the derived class.")

    @classmethod
    @abstractmethod
    def get_simple_correction_mapping(cls) -> SimpleContCalMappingType:
        raise NotImplementedError("The 'get_asimov_correction_mapping' must be implemented in the derived class.")

    @classmethod
    def get_calibrated_df(
        cls,
        df: pd.DataFrame,
        ratio_result_values: np.ndarray,
    ) -> pd.DataFrame:
        _sample = copy.deepcopy(df)

        assert cls.this_fitter_setup is not None
        reco_id_col: str = cls.this_fitter_setup.asimov.reco_mode_col
        component_id_col: str = cls.this_fitter_setup.asimov.component_id_column_name
        weight_col: str = cls.this_fitter_setup.asimov.weight_info.col_name

        if component_id_col not in _sample.columns:
            cls.this_fitter_setup.asimov.prepare_mc_df(df=_sample)

        test_mask: np.ndarray = np.full(shape=len(_sample), fill_value=False)

        for ids, ratio_index in cls.get_correction_mapping().items():

            if all(isinstance(_e, tuple) for _e in ids) and len(ids) == 2:
                primary_reco_ids, secondary_reco_ids = ids

                assert isinstance(primary_reco_ids, tuple), (primary_reco_ids, type(primary_reco_ids))
                assert all(isinstance(_i, int) for _i in primary_reco_ids), (primary_reco_ids, ids)

                assert isinstance(secondary_reco_ids, tuple), (secondary_reco_ids, type(secondary_reco_ids))
                assert all(isinstance(_i, int) for _i in secondary_reco_ids), (secondary_reco_ids, ids)

                assert cls.secondary_reco_id_col is not None
                assert cls.secondary_reco_id_col in _sample.columns, cls.secondary_reco_id_col

                _first_reco_mask: np.ndarray = _sample[reco_id_col].isin(primary_reco_ids)
                _second_reco_mask: np.ndarray = _sample[cls.secondary_reco_id_col].isin(secondary_reco_ids)
                reco_mask: np.ndarray = _first_reco_mask & _second_reco_mask
            elif all(isinstance(_i, int) for _i in ids):
                assert isinstance(ids, tuple), (ids, type(ids))
                assert all(isinstance(_i, int) for _i in ids), (ids, type(ids))
                reco_mask = _sample[reco_id_col].isin(ids)
            else:
                raise TypeError(
                    f"Unsupported type for first elements of mapping:\n\ttype(ids) = {type(ids)}\n\twith ids = {ids}"
                )

            is_continuum: np.ndarray = np.equal(_sample[Cols.is_continuum_evt].values, np.ones(shape=reco_mask.shape))

            mask: np.ndarray = reco_mask & is_continuum

            scale_factor: float = ratio_result_values[ratio_index]
            _sample.loc[mask, weight_col] *= scale_factor

            assert not np.any(test_mask & mask), np.sum(test_mask & mask)
            test_mask |= mask

        return _sample

    # endregion
