"""
Fit setup info for dedicated tagging efficiency calibration fit.

Felix Metzner 2022
"""

from typing import Tuple
from uncertainties import ufloat

from rdstar.utilities.logging import log_to_default

from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses

from rdstar.offline_analysis.reco_modes.reco_mode_infos import TagRecoModeInfos
from rdstar.offline_analysis.mc_matching.old_mc_matching_info import OldMCMatchingIDsQueryInfo

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.fit_setup_utilities import (
    FitSetupTriplet,
    base_asimov_data_setup_info,
    tag_cal_per_tag_mode_fit_obs_setup,
    base_tag_cal_weight_info,
    tag_cal_bbbar_bkg_sig_ids,
    tag_cal_dstst_l_bkg_sig_ids,
    fei_tag_b_reco_based_fit_component_id_column_name,
    add_fei_tag_b_reco_based_fit_component_id,
    get_sig_ids_with_tag_b_reco_info,
)
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import (
    FitSetupInfoContainer,
    ComponentInfo,
    TagEfficiencyCorrectionInfo,
    RecoChannelInfo,
)


__all__ = [
    "tag_cal_fit_setup",
]


logger = log_to_default(name=__name__)


# Definition of components:


sig_d_sig_ids: Tuple[int, ...] = (
    OldMCMatchingIDsQueryInfo.Dl.id,
    OldMCMatchingIDsQueryInfo.Dtau.id,
)
sig_ds_sig_ids: Tuple[int, ...] = (
    OldMCMatchingIDsQueryInfo.DSl.id,
    OldMCMatchingIDsQueryInfo.DStau.id,
)
bkg_sig_ids: Tuple[int, ...] = (
    tag_cal_bbbar_bkg_sig_ids + tag_cal_dstst_l_bkg_sig_ids + (OldMCMatchingIDsQueryInfo.CBKG.id,)
)


# D Signal Components:

component_cb_tag_d_sig: ComponentInfo = ComponentInfo(
    name="cb_tag_d_sig",
    comp_index=1,
    latex_label="D Sig in " + LatexDecayProcesses.charged_tag_reco_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=sig_d_sig_ids, fei_reco_type="all", fei_reco_charge_type="charged"),
)

component_nb_tag_d_sig: ComponentInfo = ComponentInfo(
    name="nb_tag_d_sig",
    comp_index=2,
    latex_label="D Sig in " + LatexDecayProcesses.neutral_tag_reco_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=sig_d_sig_ids, fei_reco_type="all", fei_reco_charge_type="neutral"),
)

# D* Signal Components:

component_cb_tag_ds_sig: ComponentInfo = ComponentInfo(
    name="cb_tag_ds_sig",
    comp_index=11,
    latex_label="D$^*$ Sig in " + LatexDecayProcesses.charged_tag_reco_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=sig_ds_sig_ids, fei_reco_type="all", fei_reco_charge_type="charged"),
)

component_nb_tag_ds_sig: ComponentInfo = ComponentInfo(
    name="nb_tag_ds_sig",
    comp_index=12,
    latex_label="D$^*$ Sig in " + LatexDecayProcesses.neutral_tag_reco_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=sig_ds_sig_ids, fei_reco_type="all", fei_reco_charge_type="neutral"),
)

# Background Components:

component_cb_tag_bkg: ComponentInfo = ComponentInfo(
    name="cb_tag_bkg",
    comp_index=21,
    latex_label="Bkg in " + LatexDecayProcesses.charged_tag_reco_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=bkg_sig_ids, fei_reco_type="all", fei_reco_charge_type="charged"),
)

component_nb_tag_bkg: ComponentInfo = ComponentInfo(
    name="nb_tag_bkg",
    comp_index=22,
    latex_label="Bkg in " + LatexDecayProcesses.neutral_tag_reco_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=bkg_sig_ids, fei_reco_type="all", fei_reco_charge_type="neutral"),
)


tag_cal_fit_comps: Tuple[ComponentInfo, ...] = (
    component_cb_tag_d_sig,
    component_nb_tag_d_sig,
    #
    component_cb_tag_ds_sig,
    component_nb_tag_ds_sig,
    #
    component_cb_tag_bkg,
    component_nb_tag_bkg,
)


tag_cal_fit_chs: Tuple[RecoChannelInfo, ...] = tuple(
    [
        RecoChannelInfo(
            name="ChargedBTag",
            ch_index=0,
            reco_mode_info=TagRecoModeInfos.charged_tag_reco_modes,
            components=tag_cal_fit_comps,
            observable_infos=tag_cal_per_tag_mode_fit_obs_setup.tag_cal_ptm_obs_infos,
            tag_eff_corr_factor=ufloat(0.0, 0.00),
        ),
        RecoChannelInfo(
            name="NeutralBTag",
            ch_index=1,
            reco_mode_info=TagRecoModeInfos.neutral_tag_reco_modes,
            components=tag_cal_fit_comps,
            observable_infos=tag_cal_per_tag_mode_fit_obs_setup.tag_cal_ptm_obs_infos,
            tag_eff_corr_factor=ufloat(0.0, 0.00),
        ),
    ]
)


tag_cal_fit_tag_eff_corr_info: TagEfficiencyCorrectionInfo = TagEfficiencyCorrectionInfo(
    items=[],
    key_order=(),
)


_tag_cal_asimov_fit_setup = FitSetupInfoContainer(
    components=tag_cal_fit_comps,
    reco_channels=tag_cal_fit_chs,
    observable_infos=tag_cal_per_tag_mode_fit_obs_setup.tag_cal_ptm_obs_infos,
    weight_info=base_tag_cal_weight_info,
    data_setup_info=base_asimov_data_setup_info,
    remove_empty_bins=True,
    use_alternative_gap=True,
    tag_efficiency_correction_info=tag_cal_fit_tag_eff_corr_info,
    reco_mode_column_name=TagRecoModeInfos.fei_reco_mode_col_name,
    component_id_column_name=fei_tag_b_reco_based_fit_component_id_column_name,
    component_id_function=add_fei_tag_b_reco_based_fit_component_id,
)


tag_cal_fit_setup: FitSetupTriplet = FitSetupTriplet(asimov=_tag_cal_asimov_fit_setup)
