"""
MC SigID groups for extended mc matching information.

Felix Metzner 2023
"""

import pandas as pd

from dataclasses import dataclass
from typing import Tuple, Dict, Union, Optional, Type

from templatefitter.plotter import plot_style

from rdstar.utilities.logging import log_to_default

from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses

from rdstar.offline_analysis.observables.common_observables import Cols

from rdstar.offline_analysis.mc_matching.mc_matching_id_groups import MCMatchingGroups
from rdstar.offline_analysis.mc_matching.old_mc_matching_info import OldMCMatchingIDsQueryInfo
from rdstar.offline_analysis.mc_matching.mc_matching_info import ExtendedMCMatchingIDsQueryInfo
from rdstar.offline_analysis.mc_matching.abstract_mc_matching_info import AbstractBaseMCMatchingIDsQueryInfo


__all__ = [
    "MCMatchingCollection",
    "MCMatchingCollectionItem",
    "MCMatchingCollections",
]


logger = log_to_default(name=__name__)


@dataclass(frozen=True)
class MCMatchingCollectionItem:
    name: str
    sig_ids: Tuple[int, ...]
    new_id: int
    latex_label: str
    color: Optional[str]


@dataclass(frozen=True)
class MCMatchingCollection:
    collection_name: str
    sig_id_col_name: str
    original_sig_id_col_name: Optional[str]
    collection_definition: Tuple[MCMatchingCollectionItem, ...]
    base_mc_matching_id_query_infos: Type[AbstractBaseMCMatchingIDsQueryInfo]
    is_reduced_subset_of_original: bool = False

    def get_group_definition(self, name: str) -> MCMatchingCollectionItem:
        defs: Tuple[MCMatchingCollectionItem, ...] = tuple([d for d in self.collection_definition if d.name == name])
        assert len(defs) == 1, (len(defs), defs)
        return defs[0]

    def get_group_sig_id_to_latex_label_mapping(self) -> Dict[int, str]:
        return {v.new_id: v.latex_label for v in self.collection_definition}

    def get_group_name_to_latex_label_mapping(self) -> Dict[str, str]:
        return {v.name: v.latex_label for v in self.collection_definition}

    def get_group_sig_id_to_color_mapping(self) -> Dict[int, str]:
        assert all(v.color is not None for v in self.collection_definition), (
            self.__class__.__name__,
            self.collection_name,
            [v.name for v in self.collection_definition if v.color is None],
        )
        return {v.new_id: v.color for v in self.collection_definition if v.color is not None}

    def get_group_name_to_color_mapping(self) -> Dict[str, str]:
        assert all(v.color is not None for v in self.collection_definition), (
            self.__class__.__name__,
            self.collection_name,
            [v.name for v in self.collection_definition if v.color is None],
        )
        return {v.name: v.color for v in self.collection_definition if v.color is not None}

    @property
    def plot_order(self) -> Tuple[str, ...]:
        return tuple(reversed([v.name for v in self.collection_definition]))

    def get_sig_id_mapping_for(self, sig_id_column: str) -> Dict[str, Union[int, Tuple[int, ...]]]:
        original_sig_id_col: Optional[str] = self.original_sig_id_col_name

        if sig_id_column == self.sig_id_col_name:
            return {group.name: group.new_id for group in self.collection_definition}
        elif original_sig_id_col is not None and sig_id_column == original_sig_id_col:
            return {group.name: group.sig_ids for group in self.collection_definition}
        else:
            if original_sig_id_col is None:
                error_info: str = (
                    f"the column '{self.sig_id_col_name}' or an `original_sig_id_col_name` (which is None) "
                    f"to be defined for the {self.__class__.__name__} instance "
                )
            else:
                error_info = f"one of the column names '{self.sig_id_col_name}' or '{original_sig_id_col}'"
            raise RuntimeError(
                f"Unexpected sig_id_column {sig_id_column} "
                f"for {self.__class__.__name__} instance with the name {self.collection_name}! "
                f"Expected {error_info}!"
            )

    def get_sig_id_column_name_for_df(self, df: pd.DataFrame) -> str:
        if self.sig_id_col_name in df.columns:
            return self.sig_id_col_name
        elif self.original_sig_id_col_name is not None and self.original_sig_id_col_name in df.columns:
            return self.original_sig_id_col_name
        else:
            if self.original_sig_id_col_name is None:
                error_info: str = f"the column '{self.sig_id_col_name}' or an `original_sig_id_col_name` to be defined"
            else:
                error_info = f"one of the columns '{self.sig_id_col_name}' or '{self.original_sig_id_col_name}' to be available in the provided data frame `df`"
            raise RuntimeError(
                f"No sig id column found in the provided data frame `df` for the {self.__class__.__name__} "
                f"instance with the name '{self.collection_name}'. Expected {error_info}!"
            )


class MCMatchingCollections:
    __original_sig_id__base__collection_definition: Tuple[MCMatchingCollectionItem, ...] = tuple(
        MCMatchingCollectionItem(
            name=k,
            sig_ids=(v.id,),
            new_id=v.id,
            latex_label=v.latex_label,
            color=v.get_color(),
        )
        for k, v in OldMCMatchingIDsQueryInfo.get_sig_id_mapping().items()
    )
    original__base__collection: MCMatchingCollection = MCMatchingCollection(
        collection_name="original_sig_id__base",
        sig_id_col_name=Cols.base_sig_id_col,
        original_sig_id_col_name=None,
        collection_definition=__original_sig_id__base__collection_definition,
        base_mc_matching_id_query_infos=OldMCMatchingIDsQueryInfo,
    )

    #
    __original_sig_id__off_resonance__collection_definition: Tuple[MCMatchingCollectionItem, ...] = (
        MCMatchingCollectionItem(
            name="CBKG",
            sig_ids=(OldMCMatchingIDsQueryInfo.CBKG.id,),
            new_id=OldMCMatchingIDsQueryInfo.CBKG.id,
            latex_label=OldMCMatchingIDsQueryInfo.CBKG.latex_label,
            color=OldMCMatchingIDsQueryInfo.CBKG.get_color(),
        ),
    )
    original__off_resonance__collection: MCMatchingCollection = MCMatchingCollection(
        collection_name="original_sig_id__off_resonance",
        sig_id_col_name=Cols.base_sig_id_col,
        original_sig_id_col_name=None,
        collection_definition=__original_sig_id__off_resonance__collection_definition,
        base_mc_matching_id_query_infos=OldMCMatchingIDsQueryInfo,
        is_reduced_subset_of_original=True,
    )

    #
    #
    # Base Sig ID Collection
    __extended_sig_id__base__collection_definition: Tuple[MCMatchingCollectionItem, ...] = tuple(
        MCMatchingCollectionItem(
            name=k,
            sig_ids=(v.id,),
            new_id=v.id,
            latex_label=v.latex_label,
            color=v.get_color(),
        )
        for k, v in ExtendedMCMatchingIDsQueryInfo.get_sig_id_mapping().items()
    )
    extended_sig_id__base__collection: MCMatchingCollection = MCMatchingCollection(
        collection_name="extended_sig_id__base",
        sig_id_col_name=Cols.extended_sig_id__base_collection__col,
        original_sig_id_col_name=None,
        collection_definition=__extended_sig_id__base__collection_definition,
        base_mc_matching_id_query_infos=ExtendedMCMatchingIDsQueryInfo,
    )

    #
    # fmt: off
    __extended_sig_id__fine__collection_definition: Tuple[MCMatchingCollectionItem, ...] = (
        MCMatchingCollectionItem(name="Dtaunu", sig_ids=MCMatchingGroups.group_Dtaunu, new_id=1, latex_label=LatexDecayProcesses.b_to_d_tau_nu, color=plot_style.KITColors.kit_green),
        MCMatchingCollectionItem(name="DStaunu", sig_ids=MCMatchingGroups.group_DStaunu, new_id=2, latex_label=LatexDecayProcesses.b_to_d_st_tau_nu, color=plot_style.KITColors.kit_maygreen),
        MCMatchingCollectionItem(name="Dlnu", sig_ids=MCMatchingGroups.group_Dlnu, new_id=3, latex_label=LatexDecayProcesses.b_to_d_l_nu, color=plot_style.KITColors.kit_blue),
        MCMatchingCollectionItem(name="DSlnu", sig_ids=MCMatchingGroups.group_DSlnu, new_id=4, latex_label=LatexDecayProcesses.b_to_d_st_l_nu, color=plot_style.KITColors.kit_cyan),
        MCMatchingCollectionItem(name="DSSpi0lnu", sig_ids=MCMatchingGroups.group_DSSlnu_via_pi0, new_id=5, latex_label=LatexDecayProcesses.b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_zero, color=plot_style.KITColors.kit_red),
        MCMatchingCollectionItem(name="DSSpiplnu", sig_ids=MCMatchingGroups.group_DSSlnu_via_pip, new_id=6, latex_label=LatexDecayProcesses.b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_pm, color=plot_style.KITColors.kit_orange),
        MCMatchingCollectionItem(name="HadronicTau_HBKG", sig_ids=MCMatchingGroups.group_hadronic_tau_bkg, new_id=7, latex_label=LatexDecayProcesses.hadronic_tau_bkg, color=plot_style.KITColors.kit_brown),
        MCMatchingCollectionItem(name="DSSHadronicTau_HBKG", sig_ids=MCMatchingGroups.group_DSS_hadronic_tau_bkg, new_id=8, latex_label=LatexDecayProcesses.b_to_dstst_hadronic_tau_bkg, color=plot_style.TangoColors.chocolate1),
        MCMatchingCollectionItem(name="LeptonFake_HBKG", sig_ids=MCMatchingGroups.group_lepton_fake_bkg, new_id=9, latex_label=LatexDecayProcesses.lepton_fake_bkg, color=plot_style.KITColors.kit_purple),
        MCMatchingCollectionItem(name="LeptonCrossFeed_GBKG", sig_ids=MCMatchingGroups.group_lepton_cross_feed_bkg, new_id=10, latex_label=LatexDecayProcesses.lepton_cross_feed_bkg, color=plot_style.TangoColors.plum1),
        MCMatchingCollectionItem(name="DststDaughterMismatch_GBKG", sig_ids=MCMatchingGroups.group_DSS_daughter_mismatch, new_id=11, latex_label=LatexDecayProcesses.b_to_dstst_with_dstst_daughter_mismatch_bkg, color=plot_style.TangoColors.scarletred1),
        MCMatchingCollectionItem(name="PiLNu", sig_ids=MCMatchingGroups.group_pi_l_nu, new_id=12, latex_label=LatexDecayProcesses.b_to_pi_l_nu, color=plot_style.KITColors.kit_yellow),
        MCMatchingCollectionItem(name="Described_GBKG", sig_ids=MCMatchingGroups.group_described_generic_bkg, new_id=13, latex_label=LatexDecayProcesses.described_generic_bkg, color=plot_style.KITColors.dark_grey),
        MCMatchingCollectionItem(name="GBKG", sig_ids=MCMatchingGroups.group_generic_bkg, new_id=14, latex_label=LatexDecayProcesses.generic_bbbar_bkg, color=plot_style.KITColors.grey),
        MCMatchingCollectionItem(name="CBKG", sig_ids=MCMatchingGroups.group_continuum_bkg, new_id=15, latex_label=LatexDecayProcesses.continuum_bkg, color=plot_style.KITColors.light_grey),
    )
    # fmt: on
    extended_sig_id__fine__collection: MCMatchingCollection = MCMatchingCollection(
        collection_name="extended_sig_id__fine",
        sig_id_col_name=Cols.extended_sig_id__fine_collection__col,
        original_sig_id_col_name=Cols.extended_sig_id_col,
        collection_definition=__extended_sig_id__fine__collection_definition,
        base_mc_matching_id_query_infos=ExtendedMCMatchingIDsQueryInfo,
    )

    #
    # fmt: off
    __extended_sig_id__simple__collection_definition: Tuple[MCMatchingCollectionItem, ...] = (
        MCMatchingCollectionItem(name="Dtaunu", sig_ids=MCMatchingGroups.group_Dtaunu, new_id=1, latex_label=LatexDecayProcesses.b_to_d_tau_nu, color=plot_style.KITColors.kit_green),
        MCMatchingCollectionItem(name="DStaunu", sig_ids=MCMatchingGroups.group_DStaunu, new_id=2, latex_label=LatexDecayProcesses.b_to_d_st_tau_nu, color=plot_style.KITColors.kit_maygreen),
        MCMatchingCollectionItem(name="Dlnu", sig_ids=MCMatchingGroups.group_Dlnu, new_id=3, latex_label=LatexDecayProcesses.b_to_d_l_nu, color=plot_style.KITColors.kit_blue),
        MCMatchingCollectionItem(name="DSlnu", sig_ids=MCMatchingGroups.group_DSlnu, new_id=4, latex_label=LatexDecayProcesses.b_to_d_st_l_nu, color=plot_style.KITColors.kit_cyan),
        MCMatchingCollectionItem(name="DSSpi0lnu", sig_ids=MCMatchingGroups.group_DSSlnu_via_pi0, new_id=5, latex_label=LatexDecayProcesses.b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_zero, color=plot_style.KITColors.kit_red),
        MCMatchingCollectionItem(name="DSSpiplnu", sig_ids=MCMatchingGroups.group_DSSlnu_via_pip, new_id=6, latex_label=LatexDecayProcesses.b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_pm, color=plot_style.KITColors.kit_orange),
        MCMatchingCollectionItem(name="HadronicTau_HBKG", sig_ids=MCMatchingGroups.group_hadronic_tau_bkg, new_id=7, latex_label=LatexDecayProcesses.hadronic_tau_bkg, color=plot_style.KITColors.kit_brown),
        MCMatchingCollectionItem(name="DSSHadronicTau_HBKG", sig_ids=MCMatchingGroups.group_DSS_hadronic_tau_bkg, new_id=8, latex_label=LatexDecayProcesses.b_to_dstst_hadronic_tau_bkg, color=plot_style.TangoColors.chocolate1),
        MCMatchingCollectionItem(name="LeptonFake_HBKG", sig_ids=MCMatchingGroups.group_lepton_fake_bkg, new_id=9, latex_label=LatexDecayProcesses.lepton_fake_bkg, color=plot_style.KITColors.kit_purple),
        MCMatchingCollectionItem(name="LeptonCrossFeed_GBKG", sig_ids=MCMatchingGroups.group_lepton_cross_feed_bkg, new_id=10, latex_label=LatexDecayProcesses.lepton_cross_feed_bkg, color=plot_style.TangoColors.plum1),
        MCMatchingCollectionItem(name="DststDaughterMismatch_GBKG", sig_ids=MCMatchingGroups.group_DSS_daughter_mismatch, new_id=11, latex_label=LatexDecayProcesses.b_to_dstst_with_dstst_daughter_mismatch_bkg, color=plot_style.TangoColors.scarletred1),
        # MCMatchingCollectionItem(name="PiLNu", sig_ids=MCMatchingGroups.group_pi_l_nu, new_id=12, latex_label=LatexDecayProcesses.b_to_pi_l_nu, color=plot_style.KITColors.kit_yellow),
        # MCMatchingCollectionItem(name="Described_GBKG", sig_ids=MCMatchingGroups.group_described_generic_bkg, new_id=13, latex_label=LatexDecayProcesses.described_generic_bkg, color=plot_style.KITColors.dark_grey),
        MCMatchingCollectionItem(name="Described_GBKG", sig_ids=MCMatchingGroups.group_pi_l_nu_and_described_generic_bkg, new_id=12, latex_label=LatexDecayProcesses.described_generic_bkg, color=plot_style.KITColors.dark_grey),
        MCMatchingCollectionItem(name="GBKG", sig_ids=MCMatchingGroups.group_generic_bkg, new_id=13, latex_label=LatexDecayProcesses.generic_bbbar_bkg, color=plot_style.KITColors.grey),
        MCMatchingCollectionItem(name="CBKG", sig_ids=MCMatchingGroups.group_continuum_bkg, new_id=14, latex_label=LatexDecayProcesses.continuum_bkg, color=plot_style.KITColors.light_grey),
    )
    # fmt: on
    extended_sig_id__simple__collection: MCMatchingCollection = MCMatchingCollection(
        collection_name="extended_sig_id__simple",
        sig_id_col_name=Cols.extended_sig_id__simple_collection__col,
        original_sig_id_col_name=Cols.extended_sig_id_col,
        collection_definition=__extended_sig_id__simple__collection_definition,
        base_mc_matching_id_query_infos=ExtendedMCMatchingIDsQueryInfo,
    )

    #
    # fmt: off
    __extended_sig_id__basic__collection_definition: Tuple[MCMatchingCollectionItem, ...] = (
        MCMatchingCollectionItem(name="Dtaunu", sig_ids=MCMatchingGroups.group_Dtaunu, new_id=1, latex_label=LatexDecayProcesses.b_to_d_tau_nu, color=plot_style.KITColors.kit_green),
        MCMatchingCollectionItem(name="DStaunu", sig_ids=MCMatchingGroups.group_DStaunu, new_id=2, latex_label=LatexDecayProcesses.b_to_d_st_tau_nu, color=plot_style.KITColors.kit_maygreen),
        MCMatchingCollectionItem(name="Dlnu", sig_ids=MCMatchingGroups.group_Dlnu, new_id=3, latex_label=LatexDecayProcesses.b_to_d_l_nu, color=plot_style.KITColors.kit_blue),
        MCMatchingCollectionItem(name="DSlnu", sig_ids=MCMatchingGroups.group_DSlnu, new_id=4, latex_label=LatexDecayProcesses.b_to_d_st_l_nu, color=plot_style.KITColors.kit_cyan),
        MCMatchingCollectionItem(name="DSSpi0lnu", sig_ids=MCMatchingGroups.group_DSSlnu_via_pi0, new_id=5, latex_label=LatexDecayProcesses.b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_zero, color=plot_style.KITColors.kit_red),
        MCMatchingCollectionItem(name="DSSpiplnu", sig_ids=MCMatchingGroups.group_DSSlnu_via_pip, new_id=6, latex_label=LatexDecayProcesses.b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_pm, color=plot_style.KITColors.kit_orange),
        MCMatchingCollectionItem(name="HadronicTau_HBKG", sig_ids=MCMatchingGroups.group_combined_hadronic_tau_bkg, new_id=7, latex_label=LatexDecayProcesses.b_to_any_d_hadronic_tau_bkg, color=plot_style.KITColors.kit_brown),
        MCMatchingCollectionItem(name="Lepton_GBKG", sig_ids=MCMatchingGroups.group_combined_lepton_bkg, new_id=8, latex_label=LatexDecayProcesses.long_lepton_fake_bkg, color=plot_style.KITColors.dark_grey),  # kit_purple
        MCMatchingCollectionItem(name="GBKG", sig_ids=MCMatchingGroups.group_combined_generic_bkg, new_id=9, latex_label=LatexDecayProcesses.generic_bbbar_bkg, color=plot_style.KITColors.grey),
        MCMatchingCollectionItem(name="CBKG", sig_ids=MCMatchingGroups.group_continuum_bkg, new_id=10, latex_label=LatexDecayProcesses.continuum_bkg, color=plot_style.KITColors.light_grey),
    )
    # fmt: on
    extended_sig_id__basic__collection: MCMatchingCollection = MCMatchingCollection(
        collection_name="extended_sig_id__basic",
        sig_id_col_name=Cols.extended_sig_id__basic_collection__col,
        original_sig_id_col_name=Cols.extended_sig_id_col,
        collection_definition=__extended_sig_id__basic__collection_definition,
        base_mc_matching_id_query_infos=ExtendedMCMatchingIDsQueryInfo,
    )

    #
    # fmt: off
    __extended_sig_id__charge__collection_definition: Tuple[MCMatchingCollectionItem, ...] = (
        MCMatchingCollectionItem(name="Neutral_B_To_Dtaunu", sig_ids=MCMatchingGroups.group_charged_Dtaunu, new_id=1, latex_label=LatexDecayProcesses.bz_to_dm_tau_nu, color=plot_style.KITColors.kit_green),
        MCMatchingCollectionItem(name="Charged_B_To_Dtaunu", sig_ids=MCMatchingGroups.group_neutral_Dtaunu, new_id=2, latex_label=LatexDecayProcesses.bp_to_dz_tau_nu, color="mediumspringgreen"),
        MCMatchingCollectionItem(name="Neutral_B_To_DStaunu", sig_ids=MCMatchingGroups.group_charged_DStaunu, new_id=3, latex_label=LatexDecayProcesses.bz_to_dm_st_tau_nu, color=plot_style.KITColors.kit_maygreen),
        MCMatchingCollectionItem(name="Charged_B_To_DStaunu", sig_ids=MCMatchingGroups.group_neutral_DStaunu, new_id=4, latex_label=LatexDecayProcesses.bp_to_dz_st_tau_nu, color=plot_style.TangoColors.chameleon2),
        MCMatchingCollectionItem(name="Neutral_B_To_Dlnu", sig_ids=MCMatchingGroups.group_charged_Dlnu, new_id=5, latex_label=LatexDecayProcesses.bz_to_dm_l_nu, color=plot_style.KITColors.kit_blue),
        MCMatchingCollectionItem(name="Charged_B_To_Dlnu", sig_ids=MCMatchingGroups.group_neutral_Dlnu, new_id=6, latex_label=LatexDecayProcesses.bp_to_dz_l_nu, color=plot_style.TangoColors.skyblue1),
        MCMatchingCollectionItem(name="Neutral_B_To_DSlnu", sig_ids=MCMatchingGroups.group_charged_DSlnu, new_id=7, latex_label=LatexDecayProcesses.bz_to_dm_st_l_nu, color=plot_style.KITColors.kit_cyan),
        MCMatchingCollectionItem(name="Charged_B_To_DSlnu", sig_ids=MCMatchingGroups.group_neutral_DSlnu, new_id=8, latex_label=LatexDecayProcesses.bp_to_dz_st_l_nu, color="cyan"),
        MCMatchingCollectionItem(name="DSSpi0lnu", sig_ids=MCMatchingGroups.group_DSSlnu_via_pi0, new_id=9, latex_label=LatexDecayProcesses.b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_zero, color=plot_style.KITColors.kit_red),
        MCMatchingCollectionItem(name="DSSpiplnu", sig_ids=MCMatchingGroups.group_DSSlnu_via_pip, new_id=10, latex_label=LatexDecayProcesses.b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_pm, color=plot_style.KITColors.kit_orange),
        MCMatchingCollectionItem(name="HadronicTau_HBKG", sig_ids=MCMatchingGroups.group_combined_hadronic_tau_bkg, new_id=11, latex_label=LatexDecayProcesses.b_to_any_d_hadronic_tau_bkg, color=plot_style.KITColors.kit_brown),
        MCMatchingCollectionItem(name="Lepton_GBKG", sig_ids=MCMatchingGroups.group_combined_lepton_bkg, new_id=12, latex_label=LatexDecayProcesses.long_lepton_fake_bkg, color=plot_style.KITColors.dark_grey),
        MCMatchingCollectionItem(name="GBKG", sig_ids=MCMatchingGroups.group_combined_generic_bkg, new_id=13, latex_label=LatexDecayProcesses.generic_bbbar_bkg, color=plot_style.KITColors.grey),
        MCMatchingCollectionItem(name="CBKG", sig_ids=MCMatchingGroups.group_continuum_bkg, new_id=14, latex_label=LatexDecayProcesses.continuum_bkg, color=plot_style.KITColors.light_grey),
    )
    # fmt: on
    extended_sig_id__charge__collection: MCMatchingCollection = MCMatchingCollection(
        collection_name="extended_sig_id__charge",
        sig_id_col_name=Cols.extended_sig_id__charge_collection__col,
        original_sig_id_col_name=Cols.extended_sig_id_col,
        collection_definition=__extended_sig_id__charge__collection_definition,
        base_mc_matching_id_query_infos=ExtendedMCMatchingIDsQueryInfo,
    )

    #
    # fmt: off
    __extended_sig_id__charge_and_dstar__collection_definition: Tuple[MCMatchingCollectionItem, ...] = (
        MCMatchingCollectionItem(name="Neutral_B_To_Dtaunu", sig_ids=MCMatchingGroups.group_charged_Dtaunu, new_id=1, latex_label=LatexDecayProcesses.bz_to_dm_tau_nu, color=plot_style.KITColors.kit_green),
        MCMatchingCollectionItem(name="Charged_B_To_Dtaunu", sig_ids=MCMatchingGroups.group_neutral_Dtaunu, new_id=2, latex_label=LatexDecayProcesses.bp_to_dz_tau_nu, color="mediumspringgreen"),
        MCMatchingCollectionItem(name="Neutral_B_To_DStaunu", sig_ids=MCMatchingGroups.group_charged_DStaunu, new_id=3, latex_label=LatexDecayProcesses.bz_to_dm_st_tau_nu, color=plot_style.KITColors.kit_maygreen),
        MCMatchingCollectionItem(name="Cahrged_B_To_DStaunu_via_pi", sig_ids=MCMatchingGroups.group_neutral_DStaunu_via_pi, new_id=4, latex_label=LatexDecayProcesses.bp_to_dz_st_tau_nu_dst_via_pi, color=plot_style.TangoColors.chameleon2),
        MCMatchingCollectionItem(name="Charged_B_To_DStaunu_via_ph", sig_ids=MCMatchingGroups.group_neutral_DStaunu_via_photon, new_id=5, latex_label=LatexDecayProcesses.bp_to_dz_st_tau_nu_dst_via_ph, color=plot_style.TangoColors.chameleon3),
        MCMatchingCollectionItem(name="Neutral_B_To_Dlnu", sig_ids=MCMatchingGroups.group_charged_Dlnu, new_id=6, latex_label=LatexDecayProcesses.bz_to_dm_l_nu, color=plot_style.KITColors.kit_blue),
        MCMatchingCollectionItem(name="Charged_B_To_Dlnu", sig_ids=MCMatchingGroups.group_neutral_Dlnu, new_id=7, latex_label=LatexDecayProcesses.bp_to_dz_l_nu, color=plot_style.TangoColors.skyblue1),
        MCMatchingCollectionItem(name="Neutral_B_To_DSlnu", sig_ids=MCMatchingGroups.group_charged_DSlnu, new_id=8, latex_label=LatexDecayProcesses.bz_to_dm_st_l_nu, color=plot_style.KITColors.kit_cyan),
        MCMatchingCollectionItem(name="Charged_B_To_DSlnu_via_pi", sig_ids=MCMatchingGroups.group_neutral_DSlnu_via_pi, new_id=9, latex_label=LatexDecayProcesses.bp_to_dz_st_l_nu_dst_via_pi, color="cyan"),
        MCMatchingCollectionItem(name="Charged_B_To_DSlnu_via_ph", sig_ids=MCMatchingGroups.group_neutral_DSlnu_via_photon, new_id=10, latex_label=LatexDecayProcesses.bp_to_dz_st_l_nu_dst_via_ph, color=plot_style.TangoColors.skyblue3),
        MCMatchingCollectionItem(name="DSSpi0lnu", sig_ids=MCMatchingGroups.group_DSSlnu_via_pi0, new_id=11, latex_label=LatexDecayProcesses.b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_zero, color=plot_style.KITColors.kit_red),
        MCMatchingCollectionItem(name="DSSpiplnu", sig_ids=MCMatchingGroups.group_DSSlnu_via_pip, new_id=12, latex_label=LatexDecayProcesses.b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_pm, color=plot_style.KITColors.kit_orange),
        MCMatchingCollectionItem(name="HadronicTau_HBKG", sig_ids=MCMatchingGroups.group_combined_hadronic_tau_bkg, new_id=13, latex_label=LatexDecayProcesses.b_to_any_d_hadronic_tau_bkg, color=plot_style.KITColors.kit_brown),
        MCMatchingCollectionItem(name="Lepton_GBKG", sig_ids=MCMatchingGroups.group_combined_lepton_bkg, new_id=14, latex_label=LatexDecayProcesses.long_lepton_fake_bkg, color=plot_style.KITColors.dark_grey),
        MCMatchingCollectionItem(name="GBKG", sig_ids=MCMatchingGroups.group_combined_generic_bkg, new_id=15, latex_label=LatexDecayProcesses.generic_bbbar_bkg, color=plot_style.KITColors.grey),
        MCMatchingCollectionItem(name="CBKG", sig_ids=MCMatchingGroups.group_continuum_bkg, new_id=16, latex_label=LatexDecayProcesses.continuum_bkg, color=plot_style.KITColors.light_grey),
    )
    # fmt: on
    extended_sig_id__charge_and_dstar__collection: MCMatchingCollection = MCMatchingCollection(
        collection_name="extended_sig_id__charge_dstar",
        sig_id_col_name=Cols.extended_sig_id__charge_dstar_collection__col,
        original_sig_id_col_name=Cols.extended_sig_id_col,
        collection_definition=__extended_sig_id__charge_and_dstar__collection_definition,
        base_mc_matching_id_query_infos=ExtendedMCMatchingIDsQueryInfo,
    )

    #
    # fmt: off
    __extended_sig_id__charge_and_dstar_and_lepton__collection_definition: Tuple[MCMatchingCollectionItem, ...] = (
        MCMatchingCollectionItem(name="Neutral_B_To_Dtaunu", sig_ids=MCMatchingGroups.group_charged_Dtaunu, new_id=1, latex_label=LatexDecayProcesses.bz_to_dm_tau_nu, color=plot_style.KITColors.kit_green),
        MCMatchingCollectionItem(name="Charged_B_To_Dtaunu", sig_ids=MCMatchingGroups.group_neutral_Dtaunu, new_id=2, latex_label=LatexDecayProcesses.bp_to_dz_tau_nu, color="mediumspringgreen"),
        MCMatchingCollectionItem(name="Neutral_B_To_DStaunu", sig_ids=MCMatchingGroups.group_charged_DStaunu, new_id=3, latex_label=LatexDecayProcesses.bz_to_dm_st_tau_nu, color=plot_style.KITColors.kit_maygreen),
        MCMatchingCollectionItem(name="Charged_B_To_DStaunu", sig_ids=MCMatchingGroups.group_neutral_DStaunu, new_id=4, latex_label=LatexDecayProcesses.bp_to_dz_st_tau_nu, color=plot_style.TangoColors.chameleon2),
        MCMatchingCollectionItem(name="Neutral_B_To_Denu", sig_ids=MCMatchingGroups.group_chargedD_enu, new_id=5, latex_label=LatexDecayProcesses.bz_to_dm_e_nu, color=plot_style.KITColors.kit_blue),
        MCMatchingCollectionItem(name="Charged_B_To_Denu", sig_ids=MCMatchingGroups.group_neutralD_enu, new_id=6, latex_label=LatexDecayProcesses.bp_to_dz_e_nu, color=plot_style.TangoColors.skyblue1),
        MCMatchingCollectionItem(name="Neutral_B_To_DSenu", sig_ids=MCMatchingGroups.group_chargedDS_e_nu, new_id=7, latex_label=LatexDecayProcesses.bz_to_dm_st_e_nu, color=plot_style.KITColors.kit_cyan),
        MCMatchingCollectionItem(name="Charged_B_To_DSenu_via_pi", sig_ids=MCMatchingGroups.group_neutralDS_e_nu_DS_via_pi, new_id=8, latex_label=LatexDecayProcesses.bp_to_dz_st_e_nu_dst_via_pi, color="cyan"),
        MCMatchingCollectionItem(name="Charged_B_To_DSenu_via_ph", sig_ids=MCMatchingGroups.group_neutralDS_e_nu_DS_via_photon, new_id=9, latex_label=LatexDecayProcesses.bp_to_dz_st_e_nu_dst_via_ph, color=plot_style.TangoColors.skyblue3),
        MCMatchingCollectionItem(name="Neutral_B_To_Dmunu", sig_ids=MCMatchingGroups.group_chargedD_munu, new_id=10, latex_label=LatexDecayProcesses.bz_to_dm_l_nu, color=plot_style.KITColors.kit_blue),
        MCMatchingCollectionItem(name="Charged_B_To_Dmunu", sig_ids=MCMatchingGroups.group_neutralD_munu, new_id=11, latex_label=LatexDecayProcesses.bp_to_dz_l_nu, color=plot_style.TangoColors.skyblue1),
        MCMatchingCollectionItem(name="Neutral_B_To_DSmunu", sig_ids=MCMatchingGroups.group_chargedDS_mu_nu, new_id=12, latex_label=LatexDecayProcesses.bz_to_dm_st_l_nu, color=plot_style.KITColors.kit_cyan),
        MCMatchingCollectionItem(name="Charged_B_To_DSmunu_via_pi", sig_ids=MCMatchingGroups.group_neutralDS_mu_nu_DS_via_pi, new_id=13, latex_label=LatexDecayProcesses.bp_to_dz_st_mu_nu_dst_via_pi, color="cyan"),
        MCMatchingCollectionItem(name="Charged_B_To_DSmunu_via_ph", sig_ids=MCMatchingGroups.group_neutralDS_mu_nu_DS_via_photon, new_id=14, latex_label=LatexDecayProcesses.bp_to_dz_st_mu_nu_dst_via_ph, color=plot_style.TangoColors.skyblue3),
        MCMatchingCollectionItem(name="DSSpi0lnu", sig_ids=MCMatchingGroups.group_DSSlnu_via_pi0, new_id=15, latex_label=LatexDecayProcesses.b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_zero, color=plot_style.KITColors.kit_red),
        MCMatchingCollectionItem(name="DSSpiplnu", sig_ids=MCMatchingGroups.group_DSSlnu_via_pip, new_id=16, latex_label=LatexDecayProcesses.b_to_dstst_l_nu_with_dstst_to_d_wwost_pi_pm, color=plot_style.KITColors.kit_orange),
        MCMatchingCollectionItem(name="HadronicTau_HBKG", sig_ids=MCMatchingGroups.group_combined_hadronic_tau_bkg, new_id=17, latex_label=LatexDecayProcesses.b_to_any_d_hadronic_tau_bkg, color=plot_style.KITColors.kit_brown),
        MCMatchingCollectionItem(name="Lepton_GBKG", sig_ids=MCMatchingGroups.group_combined_lepton_bkg, new_id=18, latex_label=LatexDecayProcesses.long_lepton_fake_bkg, color=plot_style.KITColors.dark_grey),
        MCMatchingCollectionItem(name="GBKG", sig_ids=MCMatchingGroups.group_combined_generic_bkg, new_id=19, latex_label=LatexDecayProcesses.generic_bbbar_bkg, color=plot_style.KITColors.grey),
        MCMatchingCollectionItem(name="CBKG", sig_ids=MCMatchingGroups.group_continuum_bkg, new_id=20, latex_label=LatexDecayProcesses.continuum_bkg, color=plot_style.KITColors.light_grey),
    )
    # fmt: on
    extended_sig_id__charge_and_dstar_and_lepton__collection: MCMatchingCollection = MCMatchingCollection(
        collection_name="extended_sig_id__charge_dstar_lepton",
        sig_id_col_name=Cols.extended_sig_id__charge_dstar_lepton_collection__col,
        original_sig_id_col_name=Cols.extended_sig_id_col,
        collection_definition=__extended_sig_id__charge_and_dstar_and_lepton__collection_definition,
        base_mc_matching_id_query_infos=ExtendedMCMatchingIDsQueryInfo,
    )

    # fmt: off
    __charm_c_bkg_info = ExtendedMCMatchingIDsQueryInfo.Charm_CBKG
    __uds_c_bkg_info = ExtendedMCMatchingIDsQueryInfo.UDS_CBKG
    __extended_sig_id__continuum__collection_definition: Tuple[MCMatchingCollectionItem, ...] = (
        MCMatchingCollectionItem(name="Charm_CBKG", sig_ids=(__charm_c_bkg_info.id,), new_id=__charm_c_bkg_info.id, latex_label=LatexDecayProcesses.to_charm_continuum_txt, color=plot_style.KITColors.dark_grey),
        MCMatchingCollectionItem(name="UDS_CBKG", sig_ids=(__uds_c_bkg_info.id,), new_id=__uds_c_bkg_info.id, latex_label=LatexDecayProcesses.to_uds_continuum_txt, color=plot_style.KITColors.light_grey),
    )
    # fmt: on
    extended_sig_id__continuum__collection: MCMatchingCollection = MCMatchingCollection(
        collection_name="extended_sig_id__continuum",
        sig_id_col_name=Cols.extended_sig_id_col,
        original_sig_id_col_name=None,
        collection_definition=__extended_sig_id__continuum__collection_definition,
        base_mc_matching_id_query_infos=ExtendedMCMatchingIDsQueryInfo,
        is_reduced_subset_of_original=True,
    )

    @staticmethod
    def test_collection_definition(collection: MCMatchingCollection) -> None:
        _all_ext_sig_ids: Tuple[int, ...] = collection.base_mc_matching_id_query_infos.get_all_sig_ids()

        collection_definition: Tuple[MCMatchingCollectionItem, ...] = collection.collection_definition
        collection_name: str = collection.collection_name

        all_ids_in_group: Tuple[int, ...] = tuple(sig_id for group in collection_definition for sig_id in group.sig_ids)
        assert len(all_ids_in_group) == len(set(all_ids_in_group)), (
            collection_name,
            len(all_ids_in_group),
            len(set(all_ids_in_group)),
            set([i for i in set(all_ids_in_group) if all_ids_in_group.count(i) > 1]),
        )

        if not collection.is_reduced_subset_of_original:
            assert set(all_ids_in_group) == set(_all_ext_sig_ids), (
                collection_name,
                all_ids_in_group,
                [sig_id for sig_id in all_ids_in_group if sig_id not in _all_ext_sig_ids],
                [sig_id for sig_id in _all_ext_sig_ids if sig_id not in all_ids_in_group],
            )
        else:
            logger.warning(
                f"MCMatchingCollection {collection.collection_name} is subset of original "
                f"Signal ID definition and covers only the following IDs:\n\t-"
                + "\n\t-".join([f"{i.name} ({i.new_id}): {i.sig_ids}" for i in collection_definition])
                + f"\nNot covered are the following IDs:\n{[i for i in _all_ext_sig_ids if i not in all_ids_in_group]}"
            )

        assert all(v in _all_ext_sig_ids for group in collection_definition for v in group.sig_ids), (
            collection_name,
            [
                (group.name, [miss_v for miss_v in group.sig_ids if miss_v not in _all_ext_sig_ids])
                for group in collection_definition
                if not all(v in _all_ext_sig_ids for v in group.sig_ids)
            ],
        )

        _latex_labels: Tuple[str, ...] = tuple(collection.get_group_sig_id_to_latex_label_mapping().values())
        assert len(_latex_labels) == len(set(_latex_labels)), (
            len(_latex_labels),
            len(set(_latex_labels)),
            set([latex_label for latex_label in set(_latex_labels) if _latex_labels.count(latex_label) > 1]),
        )

    @staticmethod
    def get_collections_dict() -> Dict[str, MCMatchingCollection]:
        _method_names: Tuple[str, ...] = (
            "get_collections_dict",
            "get_all_collections",
            "test_collection_definition",
            "run_checks",
        )

        return {
            attr_name: val
            for attr_name, val in MCMatchingCollections.__dict__.items()
            if attr_name.endswith("__collection")
            and not ((attr_name in _method_names) or (attr_name.endswith("__") or attr_name.startswith("__")))
        }

    @staticmethod
    def get_all_collections() -> Tuple[MCMatchingCollection, ...]:
        return tuple(MCMatchingCollections.get_collections_dict().values())

    @staticmethod
    def run_checks() -> None:
        __all_collections: Tuple[MCMatchingCollection, ...] = MCMatchingCollections.get_all_collections()
        __all_collection_names: Tuple[str, ...] = tuple(c.collection_name for c in __all_collections)
        __all_generated_collection_column_names: Tuple[str, ...] = tuple(
            c.sig_id_col_name for c in __all_collections if c.original_sig_id_col_name is not None
        )

        assert len(set(__all_collection_names)) == len(__all_collection_names), (
            len(__all_collection_names),
            len(set(__all_collection_names)),
            set([n for n in set(__all_collection_names) if __all_collection_names.count(n) > 1]),
        )
        assert len(set(__all_generated_collection_column_names)) == len(__all_generated_collection_column_names), (
            len(__all_generated_collection_column_names),
            len(set(__all_generated_collection_column_names)),
            set(
                [
                    c
                    for c in set(__all_generated_collection_column_names)
                    if __all_generated_collection_column_names.count(c) > 1
                ]
            ),
        )

        for collection_def in __all_collections:
            MCMatchingCollections.test_collection_definition(collection=collection_def)


if __name__ == "__main__":
    MCMatchingCollections.run_checks()
    print("Running the assertions above... and nothing failed if this is printed!")

    for mc_coll in MCMatchingCollections.get_all_collections():
        print(f"Looging at MCMatchingCollection {mc_coll.collection_name}")
        color_dict: Dict[int, str] = mc_coll.get_group_sig_id_to_color_mapping()
        for coll_key, col_value in mc_coll.get_group_sig_id_to_latex_label_mapping().items():
            print(f"{coll_key}: ({color_dict[coll_key]}) {col_value}")
