"""
Utilities for MC matching.

Felix Metzner 2023
"""


from dataclasses import dataclass
from typing import Dict, Callable, Tuple, Union, Optional

from rdstar.utilities.logging import log_to_default


__all__ = [
    "MCSigID",
    "MatchQueryInfo",
    #
    "MCSigIDNameToQueryMappingType",
    #
    "MCSigIDToLatexLabelMappingType",
    "MCSigIDNameToLatexLabelMappingType",
    #
    "MCSigIDToColorMappingType",
    "MCSigIDNameToColorMappingType",
    #
    "RecoModeToExpectedSignalSigIDMappingType",
    "SigIDToExpectedSignalForRecoModeMappingType",
]


logger = log_to_default(name=__name__)


@dataclass(frozen=True)
class MCSigID:
    id: int
    color: str
    expected_signal_in_decay_ids: Union[None, int, Tuple[int, ...]]

    def get_expected_signal_decay_mode_ids(self) -> Optional[Tuple[int, ...]]:
        if self.expected_signal_in_decay_ids is None:
            return None
        elif isinstance(self.expected_signal_in_decay_ids, int):
            return (self.expected_signal_in_decay_ids,)
        else:
            return self.expected_signal_in_decay_ids

    def get_tuple_with_expected_signal_decay_mode_ids(self) -> Tuple[int, ...]:
        expected_ids: Optional[Tuple[int, ...]] = self.get_expected_signal_decay_mode_ids()
        if expected_ids is not None:
            return expected_ids
        else:
            return tuple()


@dataclass(frozen=True)
class MatchQueryInfo:
    id_info: Union[int, MCSigID]
    query: Callable
    latex_label: str

    @property
    def has_full_mc_sig_id(self) -> bool:
        if isinstance(self.id_info, int):
            return False
        assert isinstance(self.id_info, MCSigID), (type(self.id_info), self.id_info)
        return True

    @property
    def id(self) -> int:
        if isinstance(self.id_info, int):
            return self.id_info
        else:
            return self.id_info.id

    @property
    def expected_signal_decay_mode_ids(self) -> Optional[Tuple[int, ...]]:
        if self.has_full_mc_sig_id:
            assert isinstance(self.id_info, MCSigID), (type(self.id_info), self.id_info)
            return self.id_info.get_expected_signal_decay_mode_ids()
        else:
            return None

    def get_color(self) -> str:
        assert isinstance(self.id_info, MCSigID), (type(self.id_info), self.id_info)
        return self.id_info.color


MCSigIDNameToQueryMappingType = Dict[str, MatchQueryInfo]

MCSigIDToLatexLabelMappingType = Dict[int, str]
MCSigIDNameToLatexLabelMappingType = Dict[str, str]

MCSigIDToColorMappingType = Dict[int, str]
MCSigIDNameToColorMappingType = Dict[str, str]

RecoModeToExpectedSignalSigIDMappingType = Dict[int, Tuple[int, ...]]
SigIDToExpectedSignalForRecoModeMappingType = Dict[int, Tuple[int, ...]]
