"""
Tools to modify MC matching IDs.

Felix Metzner 2023
"""


import copy
import pandas as pd

from typing import Dict, Tuple

from rdstar.offline_analysis.observables.common_observables import Cols

from rdstar.offline_analysis.reco_modes.reco_mode_utilities import RecoModeDictType
from rdstar.offline_analysis.mc_matching.mc_matching_utilities import MatchQueryInfo, MCSigIDNameToQueryMappingType


__all__ = [
    "SigIDAndGenIDCombinerInputType",
    "SigIDAndGenIDCombiner",
]


SigIDAndGenIDCombinerInputType = Dict[Tuple[str, int], Dict[str, Tuple[int, ...]]]


class SigIDAndGenIDCombiner:
    def __init__(
        self,
        customizing_dict: SigIDAndGenIDCombinerInputType,
        sig_id_column_name: str,
        decay_mode_dict: RecoModeDictType,
    ) -> None:
        self._customizing_dict = customizing_dict
        self._sig_id_column_name = sig_id_column_name
        self._decay_mode_dict = decay_mode_dict

    def customize_sig_id(
        self,
        df: pd.DataFrame,
    ) -> None:
        for new_sig_id_tuple, mode_sigID_dict in self.customizing_dict.items():
            _, new_sig_id = new_sig_id_tuple

            for mode_name, old_sig_ids in mode_sigID_dict.items():
                mode_ids: Tuple[int, ...] = self.decay_mode_dict[mode_name]

                sig_id_cond = df[self.sig_id_column_name].isin(old_sig_ids)

                mode_cond = df[Cols.decay_mode_column_name].isin(mode_ids)
                cond = sig_id_cond & mode_cond
                print(f"Overwrite IDs {old_sig_ids} in mode {mode_name} with ID {new_sig_id}")
                df.loc[cond, self.sig_id_column_name] = new_sig_id

    def get_updated_component_dict(
        self,
        initial_sig_query_dict: MCSigIDNameToQueryMappingType,
    ) -> MCSigIDNameToQueryMappingType:

        updated_component_dict: MCSigIDNameToQueryMappingType = {
            new_comp_name: MatchQueryInfo(
                query=lambda df: pd.Series(data=True, index=df.index),
                id_info=new_sig_id,
                latex_label=f"{new_comp_name}_{new_sig_id}",
            )
            for new_comp_name, new_sig_id in self.customizing_dict.keys()
        }

        sig_query_dict = copy.copy(initial_sig_query_dict)
        sig_query_dict.update(updated_component_dict)
        return sig_query_dict

    @property
    def customizing_dict(self) -> SigIDAndGenIDCombinerInputType:
        return self._customizing_dict

    @property
    def sig_id_column_name(self) -> str:
        return self._sig_id_column_name

    @property
    def decay_mode_dict(self) -> RecoModeDictType:
        return self._decay_mode_dict
