"""
New implementation for scaling weights of MC samples.

Felix Metzner 2021
"""

import os
import numpy as np
import pandas as pd

from typing import Tuple, List, Dict

from analysistools.offline_analysis import sample_weights as at_sample_weights

from rdstar.utilities import PathType
from rdstar.offline_analysis.tasks.preparation_tasks import DataPreparationTask
from rdstar.offline_analysis.rdstar_sample_weights import scale_weight_column_name, signal_mc_bfs
from rdstar.offline_analysis.rdstar_samples import (
    data_type_combinations,
    get_b_flavour_from_data_type,
    get_number_of_produced_events,
)

from rdstar.branching_fractions import b_meson_bf_calculation as bf_calc
from rdstar.branching_fractions import belle_neutral_b_meson_decay_info as belle_mixed_bfs
from rdstar.branching_fractions import belle_charged_b_meson_decay_info as belle_charged_bfs
from rdstar.branching_fractions import belle_special_b_meson_decay_info as belle_special_bfs

from rdstar.branching_fractions.old_new_bf_comparison_mixed_b import MixedBBFComparison
from rdstar.branching_fractions.old_new_bf_comparison_charged_b import ChargedBBFComparison


__all__ = [
    "MCSampleScaler",
]


class MCSampleScaler:
    mc_scaling_weight_column_name = "__mc_sample_scaling_weight__"  # type: str

    _required_cols = (DataPreparationTask.scale_corr_column,)  # type: Tuple[str, ...]

    _generic_data_types = (
        "Mixed",
        "Charged",
    )  # type: Tuple[str, ...]

    _dstst_data_types = tuple(
        [
            *data_type_combinations["Dss"],
            *data_type_combinations["DssTau"],
        ]
    )  # type: Tuple[str, ...]

    _gap_done_pipi_data_types = tuple(data_type_combinations["GapD1"])  # type: Tuple[str, ...]

    _gap_data_types = tuple(
        [v for k, vs in data_type_combinations.items() if k.startswith("GapD") and k != "GapD1" for v in vs]
    )  # type: Tuple[str, ...]

    all_covered_data_types = (
        *_generic_data_types,
        *at_sample_weights.generic_charmless_types,
        *_dstst_data_types,
        *_gap_done_pipi_data_types,
        *_gap_data_types,
    )  # type: Tuple[str, ...]

    old_scaling_weight_col_name = scale_weight_column_name  # type: str

    def __init__(
        self,
        data_type_column_name: str,
        file_dict_dir_path: PathType,
    ) -> None:
        self.data_type_col = data_type_column_name

        assert os.path.isdir(file_dict_dir_path), file_dict_dir_path
        self._file_dict_dir_path = file_dict_dir_path  # type: PathType

        self.data_type_scaling_map = self.generate_data_type_scaling_map(
            file_dict_dir_path=self._file_dict_dir_path
        )  # type: Dict[str, float]

    @property
    def required_cols(self) -> Tuple[str, ...]:
        return tuple([*self._required_cols, self.data_type_col])

    def add_scaling_weights(self, df: pd.DataFrame) -> None:
        assert self.mc_scaling_weight_column_name not in df.columns, self.mc_scaling_weight_column_name
        assert all(c in df.columns for c in self.required_cols), [c for c in self.required_cols if c not in df.columns]

        unique_dts = list(df[self.data_type_col].unique())  # type: List[str]

        if df.empty:
            df[self.mc_scaling_weight_column_name] = 1.0
            return

        df.loc[:, self.mc_scaling_weight_column_name] = np.NaN

        for data_type in unique_dts:
            dt_cond = df[self.data_type_col] == data_type  # type: pd.Series

            scaling_weights = self._get_scaling_weight(df=df.loc[dt_cond], data_type=data_type)  # type: np.ndarray

            assert len(scaling_weights) == sum(dt_cond), (len(scaling_weights), sum(dt_cond), len(df.loc[dt_cond]))
            df.loc[dt_cond, self.mc_scaling_weight_column_name] = scaling_weights

        assert df[self.mc_scaling_weight_column_name].dtype == np.float64, df[self.mc_scaling_weight_column_name].dtype
        assert not pd.isnull(df[self.mc_scaling_weight_column_name]).any(), sum(
            pd.isnull(df[self.mc_scaling_weight_column_name])
        )
        assert not np.any(np.isnan(df[self.mc_scaling_weight_column_name].values)), np.sum(
            np.isnan(df[self.mc_scaling_weight_column_name].values)
        )

    def add_scaling_weight_for_data_type(
        self,
        df: pd.DataFrame,
        data_type: str,
    ) -> None:
        assert self.mc_scaling_weight_column_name not in df.columns, self.mc_scaling_weight_column_name
        assert all(c in df.columns for c in self.required_cols), [c for c in self.required_cols if c not in df.columns]

        if df.empty:
            df[self.mc_scaling_weight_column_name] = 1.0
            return

        unique_dts = list(df[self.data_type_col].unique())  # type: List[str]
        assert len(unique_dts) == 1, unique_dts
        assert unique_dts[0] == data_type, (unique_dts, data_type)

        scaling_weights = self._get_scaling_weight(df=df, data_type=data_type)  # type: np.ndarray
        assert len(scaling_weights) == len(np.unique(df.index)), (len(scaling_weights), len(np.unique(df.index)), len(df))
        df.loc[:, self.mc_scaling_weight_column_name] = scaling_weights

    def _get_scaling_weight(
        self,
        df: pd.DataFrame,
        data_type: str,
    ) -> np.ndarray:
        if data_type in self.data_type_scaling_map.keys():
            scaling_weight = self.data_type_scaling_map[data_type]  # type: float
        else:
            old_scaling_values = np.unique(df[self.old_scaling_weight_col_name].values)
            assert len(old_scaling_values) == 1, old_scaling_values
            scaling_weight = float(old_scaling_values[0])

        assert not np.isnan(scaling_weight), (
            scaling_weight,
            type(scaling_weight),
            type(scaling_weight).__name__,
            data_type,
        )
        return np.ones_like(df[self.data_type_col].values, dtype=float) * scaling_weight

    @staticmethod
    def generate_data_type_scaling_map(file_dict_dir_path: PathType) -> Dict[str, float]:
        scaling_map = dict()  # type: Dict[str, float]

        for data_type in MCSampleScaler.all_covered_data_types:
            is_charged = MCSampleScaler.is_charged_dt(data_type=data_type)  # type: bool
            n_bb_flavour = MCSampleScaler.get_n_bb_by_flavour(data_type=data_type)  # type: float
            n_bb_have_base = get_number_of_produced_events(
                data_type=data_type,
                file_dict_dir_path=file_dict_dir_path,
            )  # type: int

            bf_rare = belle_special_bfs.sum_bf_bp_rare if is_charged else belle_special_bfs.sum_bf_b0_rare  # type: float
            bf_ulnu = belle_special_bfs.sum_bf_bp_ulnu if is_charged else belle_special_bfs.sum_bf_b0_ulnu  # type: float
            bf_r_u = bf_rare + bf_ulnu  # type: float

            bf_old_dstst = (
                belle_charged_bfs.sum_of_belle_bp_to_dss_l_nu_bfs
                if is_charged
                else belle_mixed_bfs.sum_of_belle_b0_to_dss_l_nu_bfs
            )  # type: float
            bf_old_gap = 2 * (
                belle_charged_bfs.sum_of_belle_removed_bp_bfs
                if is_charged
                else belle_mixed_bfs.sum_of_belle_removed_b0_bfs
            )  # type: float

            bf_new_dstst = MCSampleScaler._get_new_dstst_bf_sum(charged_b=is_charged)  # type: float
            bf_new_gap_done = MCSampleScaler._get_new_gap_done_bf_sum(charged_b=is_charged)  # type: float
            bf_new_gap = MCSampleScaler._get_new_gap_bf_sum(charged_b=is_charged)  # type: float

            bf_never_covered = MCSampleScaler._get_new_missing_bf_sum(charged_b=is_charged)  # type: float

            if data_type in MCSampleScaler._generic_data_types:
                bf_removed = bf_old_gap + bf_old_dstst  # type: float
                bf_not_covered = bf_r_u + bf_new_gap + bf_new_dstst + bf_new_gap_done + bf_never_covered  # type: float
                n_bb_have = n_bb_have_base * (1.0 - (2 * bf_removed - bf_removed ** 2))  # type: float
                n_bb_want = n_bb_flavour * (1.0 - (2 * bf_not_covered - bf_not_covered ** 2))  # type: float
                scaling_weight = n_bb_want / n_bb_have  # type: float

            elif data_type in at_sample_weights.generic_charmless_types:
                bf_removed = bf_old_gap
                n_bb_have = n_bb_have_base * (1.0 - bf_removed)

                dt_bf = MCSampleScaler.get_data_type_bf(data_type=data_type)  # type: float
                n_bb_want = n_bb_flavour * (2 * dt_bf - 2 * dt_bf * (bf_r_u + bf_new_gap + bf_never_covered))
                scaling_weight = n_bb_want / n_bb_have

            elif data_type in MCSampleScaler._dstst_data_types:
                bf_removed = bf_old_gap
                n_bb_have = n_bb_have_base * (1.0 - bf_removed)

                dt_bf = MCSampleScaler.get_data_type_bf(data_type=data_type)
                bf_not_covered = bf_r_u + bf_new_gap + bf_never_covered
                bf_cross_feed = bf_new_dstst + bf_new_gap_done - dt_bf  # type: float
                n_bb_want = n_bb_flavour * (2 * dt_bf - dt_bf ** 2 - 2 * dt_bf * bf_not_covered - dt_bf * bf_cross_feed)
                scaling_weight = n_bb_want / n_bb_have

            elif data_type in MCSampleScaler._gap_done_pipi_data_types:
                bf_removed = bf_old_gap
                n_bb_have = n_bb_have_base * (1.0 - bf_removed)

                dt_bf = MCSampleScaler.get_data_type_bf(data_type=data_type)
                bf_not_covered = bf_r_u + bf_new_gap + bf_never_covered
                bf_cross_feed = bf_new_dstst
                n_bb_want = n_bb_flavour * (2 * dt_bf - dt_bf ** 2 - 2 * dt_bf * bf_not_covered - dt_bf * bf_cross_feed)
                scaling_weight = n_bb_want / n_bb_have

            elif data_type in MCSampleScaler._gap_data_types:
                bf_removed = bf_old_gap
                n_bb_have = n_bb_have_base * (1.0 - bf_removed)

                dt_bf = MCSampleScaler.get_data_type_bf(data_type=data_type)
                bf_not_covered = bf_r_u + bf_never_covered
                n_bb_want = n_bb_flavour * (2 * dt_bf - 2 * dt_bf ** 2 - 2 * dt_bf * bf_not_covered)
                scaling_weight = n_bb_want / n_bb_have
            else:
                raise KeyError(f"No scaling definition available for data type {data_type}...")

            assert data_type not in scaling_map.keys(), data_type
            scaling_map[data_type] = scaling_weight

        assert all(dt in scaling_map.keys() for dt in MCSampleScaler.all_covered_data_types), [
            dt for dt in MCSampleScaler.all_covered_data_types if dt not in scaling_map.keys()
        ]
        return scaling_map

    @staticmethod
    def get_n_bb_by_flavour(data_type: str) -> float:
        flavour = get_b_flavour_from_data_type(data_type=data_type)  # type: str
        return at_sample_weights.expected_number_of_generic_MC.get(flavour)

    @staticmethod
    def is_charged_dt(data_type: str) -> bool:
        return get_b_flavour_from_data_type(data_type=data_type) == "Charged"

    @staticmethod
    def get_data_type_bf(data_type: str) -> float:
        is_charged = MCSampleScaler.is_charged_dt(data_type=data_type)  # type: bool

        if data_type in at_sample_weights.generic_charmless_types:
            if data_type.startswith("rare"):
                return belle_special_bfs.sum_bf_bp_rare if is_charged else belle_special_bfs.sum_bf_b0_rare
            elif data_type.startswith("ulnu"):
                return belle_special_bfs.sum_bf_bp_ulnu if is_charged else belle_special_bfs.sum_bf_b0_ulnu
            else:
                raise KeyError(f"No bf info available for data type {data_type}!")
        elif data_type in signal_mc_bfs.keys():
            return signal_mc_bfs[data_type].bf
        else:
            raise KeyError(f"No bf info available for data type {data_type}!")

    @staticmethod
    def _get_new_dstst_bf_sum(charged_b: bool) -> float:
        if charged_b:
            return (
                2 * bf_calc.bf_b0_to_d0st_l_nu_our_cal.nominal_value
                + 2 * bf_calc.bf_b0_to_dprime1_l_nu_our_cal.nominal_value
                + 2 * bf_calc.bf_b0_to_dst_2_l_nu_our_cal.nominal_value
                + 2 * bf_calc.bf_b0_to_d_1_l_nu_our_cal.nominal_value
                + bf_calc.bf_b0_to_d0st_tau_nu_cal.nominal_value
                + bf_calc.bf_b0_to_dprime1_tau_nu_cal.nominal_value
                + bf_calc.bf_b0_to_dst_2_tau_nu_cal.nominal_value
                + bf_calc.bf_b0_to_d_1_tau_nu_cal.nominal_value
            )
        else:
            return (
                2 * bf_calc.bf_bp_to_d0st_l_nu_our_cal.nominal_value
                + 2 * bf_calc.bf_bp_to_dprime1_l_nu_our_cal.nominal_value
                + 2 * bf_calc.bf_bp_to_dst_2_l_nu_our_cal.nominal_value
                + 2 * bf_calc.bf_bp_to_d_1_l_nu_our_cal.nominal_value
                + bf_calc.bf_bp_to_d0st_tau_nu_cal.nominal_value
                + bf_calc.bf_bp_to_dprime1_tau_nu_cal.nominal_value
                + bf_calc.bf_bp_to_dst_2_tau_nu_cal.nominal_value
                + bf_calc.bf_bp_to_d_1_tau_nu_cal.nominal_value
            )

    @staticmethod
    def _get_new_gap_done_bf_sum(charged_b: bool) -> float:
        if charged_b:
            return 2 * bf_calc.bf_bp_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap.nominal_value
        else:
            return 2 * bf_calc.bf_b0_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap.nominal_value

    @staticmethod
    def _get_new_gap_bf_sum(charged_b: bool) -> float:
        if charged_b:
            bf_gap = (
                bf_calc.bf_bp_to_d_eta_l_nu_gap.nominal_value
                + bf_calc.bf_bp_to_d_st_eta_l_nu_gap.nominal_value
                + bf_calc.bf_bp_to_d_pi_pi_l_nu_our_gap.nominal_value
                + bf_calc.bf_bp_to_d_st_pi_pi_l_nu_our_gap.nominal_value
            )  # type: float

            bf_alt_gap = (
                bf_calc.bf_bp_to_d_eta_l_nu_gap_via_d_st_0.nominal_value
                + bf_calc.bf_bp_to_d_st_eta_l_nu_gap_via_d_prime_1.nominal_value
                + bf_calc.bf_bp_to_d_pi_pi_l_nu_gap_via_d_st_0.nominal_value
                + bf_calc.bf_bp_to_d_pi_pi_l_nu_gap_via_d_prime_1.nominal_value
                + bf_calc.bf_bp_to_d_st_pi_pi_l_nu_gap_via_d_st_0.nominal_value
                + bf_calc.bf_bp_to_d_st_pi_pi_l_nu_gap_via_d_prime_1.nominal_value
            )  # type: float
        else:
            bf_gap = (
                bf_calc.bf_b0_to_d_eta_l_nu_gap.nominal_value
                + bf_calc.bf_b0_to_d_st_eta_l_nu_gap.nominal_value
                + bf_calc.bf_b0_to_d_pi_pi_l_nu_our_gap.nominal_value
                + bf_calc.bf_b0_to_d_st_pi_pi_l_nu_our_gap.nominal_value
            )
            bf_alt_gap = (
                bf_calc.bf_b0_to_d_eta_l_nu_gap_via_d_st_0.nominal_value
                + bf_calc.bf_b0_to_d_st_eta_l_nu_gap_via_d_prime_1.nominal_value
                + bf_calc.bf_b0_to_d_pi_pi_l_nu_gap_via_d_st_0.nominal_value
                + bf_calc.bf_b0_to_d_pi_pi_l_nu_gap_via_d_prime_1.nominal_value
                + bf_calc.bf_b0_to_d_st_pi_pi_l_nu_gap_via_d_st_0.nominal_value
                + bf_calc.bf_b0_to_d_st_pi_pi_l_nu_gap_via_d_prime_1.nominal_value
            )

        assert np.isclose(bf_gap, bf_alt_gap), (bf_gap, bf_alt_gap)
        return 2 * bf_gap  # Factor 2 because of l = e, mu

    @staticmethod
    def _get_new_missing_bf_sum(charged_b: bool) -> float:
        rescaling_other_generic_bfs = True

        bf_diff_factor = 0.0 if rescaling_other_generic_bfs else 1.0  # type: float

        if charged_b:
            return (
                bf_calc.bf_bp_to_d_1_tau_nu_with_d_1_to_d_pi_pi_gap_missing.nominal_value
                + bf_calc.bf_bp_to_d_eta_tau_nu_gap_missing.nominal_value
                + bf_calc.bf_bp_to_d_st_eta_tau_nu_gap_missing.nominal_value
                + bf_calc.bf_bp_to_d_pi_pi_tau_nu_our_gap_missing.nominal_value
                + bf_calc.bf_bp_to_d_st_pi_pi_tau_nu_our_gap_missing.nominal_value
                + 2 * bf_calc.bf_bp_to_dst_2_l_nu_with_dst_2_to_dst_pi_pi_missing.nominal_value
                + 2 * bf_calc.bf_bp_to_d_s_kaon_l_nu_missing.nominal_value
                + 2 * bf_calc.bf_bp_to_d_st_s_kaon_l_nu_missing.nominal_value
                + bf_diff_factor * ChargedBBFComparison().get_total_bp_bf_diff()
            )
        else:
            return (
                bf_calc.bf_b0_to_d_1_tau_nu_with_d_1_to_d_pi_pi_gap_missing.nominal_value
                + bf_calc.bf_b0_to_d_eta_tau_nu_gap_missing.nominal_value
                + bf_calc.bf_b0_to_d_st_eta_tau_nu_gap_missing.nominal_value
                + bf_calc.bf_b0_to_d_pi_pi_tau_nu_our_gap_missing.nominal_value
                + bf_calc.bf_b0_to_d_st_pi_pi_tau_nu_our_gap_missing.nominal_value
                + 2 * bf_calc.bf_b0_to_dst_2_l_nu_with_dst_2_to_dst_pi_pi_missing.nominal_value
                + 2 * bf_calc.bf_b0_to_d_s_kaon_l_nu_missing.nominal_value
                + 2 * bf_calc.bf_b0_to_d_st_s_kaon_l_nu_missing.nominal_value
                + bf_diff_factor * MixedBBFComparison().get_total_b0_bf_diff()
            )
