# Quick calculation of uncertainties on sigmas of mm2 resolution correction

from typing import List
from dataclasses import dataclass

from uncertainties.umath import sqrt
from uncertainties import ufloat, UFloat


@dataclass(frozen=True)
class ResFitResult:
    bin_i: int
    data_sig: float
    data_sig_uncert: float
    mc_sig: float
    mc_sig_uncert: float
    sigma_diff: float
    plain_sigma_diff: float

    @property
    def data_val(self) -> UFloat:
        return ufloat(nominal_value=self.data_sig, std_dev=self.data_sig_uncert)

    @property
    def mc_val(self) -> UFloat:
        return ufloat(nominal_value=self.mc_sig, std_dev=self.mc_sig_uncert)

    def calc_diff(self) -> UFloat:
        if self.data_sig - self.mc_sig <= 0.0:
            assert self.sigma_diff == 0.0, (
                self.sigma_diff,
                self.data_sig - self.mc_sig,
                self.data_sig,
                self.mc_sig,
                self.plain_sigma_diff,
            )
            alt_res = sqrt(self.mc_val ** 2 - self.data_val ** 2)
            return ufloat(0.0, alt_res.std_dev)

        return sqrt(self.data_val ** 2 - self.mc_val ** 2)


fit_results: List[ResFitResult] = [
    ResFitResult(
        bin_i=0,
        data_sig=0.07517276645427605,
        data_sig_uncert=0.01666,
        mc_sig=0.07056725016631132,
        mc_sig_uncert=0.0108,
        sigma_diff=0.025907682651182165,
        plain_sigma_diff=0.0046055162879647255,
    ),
    ResFitResult(
        bin_i=1,
        data_sig=0.056580202471170384,
        data_sig_uncert=0.003204,
        mc_sig=0.053922401633745176,
        mc_sig_uncert=0.002616,
        sigma_diff=0.017137500218168077,
        plain_sigma_diff=0.0026578008374252074,
    ),
    ResFitResult(
        bin_i=2,
        data_sig=0.05560491145034885,
        data_sig_uncert=0.002708,
        mc_sig=0.051191254055461426,
        mc_sig_uncert=0.002147,
        sigma_diff=0.021710865612184622,
        plain_sigma_diff=0.004413657394887423,
    ),
    ResFitResult(
        bin_i=3,
        data_sig=0.05576085980341307,
        data_sig_uncert=0.003406,
        mc_sig=0.05499257793257762,
        mc_sig_uncert=0.002791,
        sigma_diff=0.009224416423019086,
        plain_sigma_diff=0.000768281870835448,
    ),
    ResFitResult(
        bin_i=4,
        data_sig=0.05812313249736217,
        data_sig_uncert=0.009539,
        mc_sig=0.0640039344716097,
        mc_sig_uncert=0.01246,
        sigma_diff=0.0,
        plain_sigma_diff=0.005880801974247521,
    ),
]


if __name__ == "__main__":

    for r in fit_results:
        print(f"{r.bin_i}: {r.calc_diff().nominal_value} +/- {r.calc_diff().std_dev}")
