"""
My copy of scipy.stats.laplace_asymmetric_gen,
as scipy provides it only since version 1.6, which requires python version 3.7 and I am bound to 3.6.

Felix Metzner, 2021
"""


import numpy as np
from scipy.stats._distn_infrastructure import rv_continuous


__all__ = [
    "laplace_asymmetric",
]


class laplace_asymmetric_gen(rv_continuous):
    r"""An asymmetric Laplace continuous random variable.
    %(before_notes)s
    See Also
    --------
    laplace : Laplace distribution
    Notes
    -----
    The probability density function for `laplace_asymmetric` is
    .. math::
       f(x, \kappa) &= \frac{1}{\kappa+\kappa^{-1}}\exp(-x\kappa),\quad x\ge0\\
                    &= \frac{1}{\kappa+\kappa^{-1}}\exp(x/\kappa),\quad x<0\\
    for :math:`-\infty < x < \infty`, :math:`\kappa > 0`.
    `laplace_asymmetric` takes ``kappa`` as a shape parameter for
    :math:`\kappa`. For :math:`\kappa = 1`, it is identical to a
    Laplace distribution.
    %(after_notes)s
    References
    ----------
    .. [1] "Asymmetric Laplace distribution", Wikipedia
            https://en.wikipedia.org/wiki/Asymmetric_Laplace_distribution
    .. [2] Kozubowski TJ and Podgórski K. A Multivariate and
           Asymmetric Generalization of Laplace Distribution,
           Computational Statistics 15, 531--540 (2000).
           :doi:`10.1007/PL00022717`
    %(example)s
    """

    def _pdf(self, x, kappa):
        return np.exp(self._logpdf(x, kappa))

    def _logpdf(self, x, kappa):
        kapinv = 1 / kappa
        lPx = x * np.where(x >= 0, -kappa, kapinv)
        lPx -= np.log(kappa + kapinv)
        return lPx

    def _cdf(self, x, kappa):
        kapinv = 1 / kappa
        kappkapinv = kappa + kapinv
        return np.where(x >= 0, 1 - np.exp(-x * kappa) * (kapinv / kappkapinv), np.exp(x * kapinv) * (kappa / kappkapinv))

    def _sf(self, x, kappa):
        kapinv = 1 / kappa
        kappkapinv = kappa + kapinv
        return np.where(x >= 0, np.exp(-x * kappa) * (kapinv / kappkapinv), 1 - np.exp(x * kapinv) * (kappa / kappkapinv))

    def _ppf(self, q, kappa):
        kapinv = 1 / kappa
        kappkapinv = kappa + kapinv
        return np.where(
            q >= kappa / kappkapinv,
            -np.log((1 - q) * kappkapinv * kappa) * kapinv,
            np.log(q * kappkapinv / kappa) * kappa,
        )

    def _isf(self, q, kappa):
        kapinv = 1 / kappa
        kappkapinv = kappa + kapinv
        return np.where(
            q <= kapinv / kappkapinv,
            -np.log(q * kappkapinv * kappa) * kapinv,
            np.log((1 - q) * kappkapinv / kappa) * kappa,
        )

    def _stats(self, kappa):
        kapinv = 1 / kappa
        mn = kapinv - kappa
        var = kapinv * kapinv + kappa * kappa
        g1 = 2.0 * (1 - np.power(kappa, 6)) / np.power(1 + np.power(kappa, 4), 1.5)
        g2 = 6.0 * (1 + np.power(kappa, 8)) / np.power(1 + np.power(kappa, 4), 2)
        return mn, var, g1, g2

    def _entropy(self, kappa):
        return 1 + np.log(kappa + 1 / kappa)


laplace_asymmetric = laplace_asymmetric_gen(name="laplace_asymmetric")
