"""
Visualization of effect of different smearing techniques.

Felix Metzner 2021
"""

import os
import numpy as np
import matplotlib.pyplot as plt
import scipy.stats as scipy_stats

from typing import Tuple, Dict
from scipy.special import expit

from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.plot_utilities import FigureType, AxesType, export
from templatefitter.plotter.plot_style import KITColors, set_matplotlibrc_params, xlabel_pos, ylabel_pos

from rdstar.utilities import PathType
from rdstar.offline_analysis.mms_resolution_correction.shift_stats import laplace_asymmetric
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction import my_lap


__all__ = [
    "SmearingComparisonPlot",
]


class SmearingComparisonPlot:
    base_file_name = "smearing_illustration"  # type: str
    minimal_pdf_value = 0.00000001  # type: float
    number_of_included_neighbours = 5  # type: int

    def __init__(
        self,
        target_dir_path: PathType,
        base_hist_var: HistVariable,
    ) -> None:

        self.hist_var = base_hist_var  # type: HistVariable
        self._target_dir_path = target_dir_path  # type: PathType

    def get_output_dict(self, n_sigmas: int) -> Dict[str, PathType]:
        return dict([self.get_output_key_path_pair_for(sigma_index=i + 1) for i in range(n_sigmas)])

    def get_output_key_path_pair_for(self, sigma_index: int) -> Tuple[str, PathType]:
        key = f"{self.base_file_name}_for_sigma_index_{sigma_index}_value"  # type: str
        file_path = os.path.join(self._target_dir_path, f"{key}.pdf")  # type: PathType
        return key, file_path

    def create_plot(
        self,
        sigma_index: int,
        sigma_value: float,
    ) -> None:
        set_matplotlibrc_params()
        fig, ax = plt.subplots(1, 1, dpi=200)  # type: FigureType, AxesType

        _, file_path = self.get_output_key_path_pair_for(sigma_index=sigma_index)
        dir_path = os.path.dirname(file_path)  # type: PathType
        pure_file_name = os.path.splitext(os.path.basename(file_path))[0]  # type: str

        plot_range = self.hist_var.scope  # type: Tuple[float, float]

        test_x = np.linspace(*plot_range, 15)[1:-1]  # type: np.ndarray

        x = np.array(sorted(list(np.linspace(*plot_range, 1000)) + list(test_x)))  # type: np.ndarray

        lp_folding_norm = 1.0 / scipy_stats.norm.pdf(
            0.0, loc=my_lap.lp_folding_loc, scale=my_lap.lp_folding_scale
        )  # type: float
        lp_folded_sigmas = (my_lap.lp_expit_other + expit(my_lap.lp_expit_scale * (test_x + my_lap.lp_expit_asym))) * (
            sigma_value
            * lp_folding_norm
            * scipy_stats.norm.pdf(
                test_x,
                loc=my_lap.lp_folding_loc,
                scale=my_lap.lp_folding_scale,
            )
        )  # type: np.ndarray

        kappa_values = my_lap.min_kappa + (my_lap.max_kappa - my_lap.min_kappa) / (
            1 + np.exp(my_lap.kappa_shift + my_lap.kappa_scale * test_x)
        )  # type: np.ndarray

        for test_i, test_x_value in enumerate(test_x):
            norm_fac = 1.0 / laplace_asymmetric.pdf(
                test_x_value,
                kappa=kappa_values[test_i],
                loc=test_x_value,
                scale=0.5 * lp_folded_sigmas[test_i],
            )  # type: float

            smearing = norm_fac * laplace_asymmetric.pdf(
                x,
                kappa=kappa_values[test_i],
                loc=test_x_value,
                scale=0.5 * lp_folded_sigmas[test_i],
            )  # type: np.ndarray

            selected_x_mask = self.get_mask_for_significant_values(a=smearing)  # type: np.ndarray
            ax.plot(
                x[selected_x_mask],
                smearing[selected_x_mask],
                "-",
                lw=2,
                alpha=1.0,
                color=KITColors.kit_blue,
                label=rf"Laplace Smearing with $\sigma$ {sigma_value:.3f}" if test_i == 0 else None,
            )

            gauss_norm_fac = 1.0 / scipy_stats.norm.pdf(test_x_value, loc=test_x_value, scale=sigma_value)
            gauss_smeared_values = scipy_stats.norm.pdf(x, loc=test_x_value, scale=sigma_value)

            selected_gauss_x_mask = self.get_mask_for_significant_values(a=gauss_smeared_values)  # type: np.ndarray

            ax.plot(
                x[selected_gauss_x_mask],
                gauss_norm_fac * gauss_smeared_values[selected_gauss_x_mask],
                "--",
                lw=1,
                alpha=1.0,
                color=KITColors.kit_orange,
                label=r"Gaussian Smearing (same $\sigma$)" if test_i == 0 else None,
            )

        ax.set_xlim(*plot_range)
        ax.set_ylim(0.0, 1.24)
        ax.set_xlabel(self.hist_var.x_label, xlabel_pos)
        ax.set_ylabel("Arbitrary Scale", ylabel_pos)
        ax.legend(loc=1, frameon=False)

        export(
            fig=fig,
            filename=pure_file_name,
            target_dir=dir_path,
            close_figure=True,
        )

    def get_mask_for_significant_values(self, a: np.ndarray) -> np.ndarray:
        selected_x_mask = a > self.minimal_pdf_value  # type: np.ndarray
        for i_neighbour in range(self.number_of_included_neighbours):
            selected_x_mask = selected_x_mask | (
                (selected_x_mask != np.roll(selected_x_mask, (i_neighbour + 1)))
                | (selected_x_mask != np.roll(selected_x_mask, -(i_neighbour + 1)))
            )
        return selected_x_mask
