"""
Other definitions of the Missing Mass Squared

Felix Metzner 2020
"""

import numpy as np
import pandas as pd

from typing import Tuple, List, Dict

from analysistools.physics_vectors.fourvector import FourVector


__all__ = [
    "MissingMassSquaredWOVertexFitVariable",
]


class MissingMassSquaredWOVertexFitVariable:
    mm2_wo_vf_col_name = "m2RecoilSignalSideWOVertexFit"  # type: str

    # Beam Energy in CMS Frame
    e_beam_cms = "Ecms"  # type: str

    four_momentum_components = (
        "E",
        "px",
        "py",
        "pz",
    )  # type: Tuple[str, ...]
    # Tag Side
    tag_4mom_vars = tuple(
        [f"daughter__bo0__cm__spuseCMSFrame__bo{v}__bc__bc" for v in four_momentum_components]
    )  # type: Tuple[str, ...]
    # Signal Side
    sig_4mom_vars = tuple(
        [f"daughter__bo1__cm__spextraInfo__bo{v}_cms_prefit__bc__bc" for v in four_momentum_components]
    )  # type: Tuple[str, ...]

    required_variables = [
        e_beam_cms,
    ]  # type: List[str]
    required_variables.extend(tag_4mom_vars)
    required_variables.extend(sig_4mom_vars)

    definitions = {
        mm2_wo_vf_col_name: (e_beam_cms, tag_4mom_vars, sig_4mom_vars),
    }  # type: Dict[str, Tuple[str, Tuple[str, ...], Tuple[str, ...]]]

    @staticmethod
    def add_mm2_definitions(df: pd.DataFrame) -> None:
        for new_col_name, _ in MissingMassSquaredWOVertexFitVariable.definitions.items():
            MissingMassSquaredWOVertexFitVariable.add_mm2_variant(df=df, name=new_col_name)

    @staticmethod
    def add_mm2_variant(df: pd.DataFrame, name: str) -> None:
        new_col_name = name  # type: str
        e_beam_cms, tag_vars, sig_vars = MissingMassSquaredWOVertexFitVariable.definitions[name]

        req_cols = MissingMassSquaredWOVertexFitVariable.get_required_columns_list_for_definition(
            name=name
        )  # type: List[str]
        assert all(c in df.columns for c in req_cols), (name, [c for c in req_cols if c not in df.columns])

        assert new_col_name not in df.columns, new_col_name

        e_beam_cms_values = df[e_beam_cms].values  # type: np.ndarray

        tag_4_moms = FourVector(p=df[list(tag_vars)].values)  # type: FourVector
        sig_4_moms = FourVector(p=df[list(sig_vars)].values)  # type: FourVector

        tag_4_moms.T = -0.5 * e_beam_cms_values
        minus_tag_4_moms = FourVector(p=-tag_4_moms.as_array)  # type: FourVector

        df.loc[:, new_col_name] = (minus_tag_4_moms - sig_4_moms).M2()

        assert not pd.isnull(df[new_col_name]).any(), new_col_name

    @staticmethod
    def get_required_columns_list() -> List[str]:
        return [c for c in MissingMassSquaredWOVertexFitVariable.required_variables]

    @staticmethod
    def get_required_columns_list_for_definition(name: str) -> List[str]:
        c1, c2, c3 = MissingMassSquaredWOVertexFitVariable.definitions[name]
        return [c1] + list(c2) + list(c3)

    @staticmethod
    def get_new_columns_list() -> List[str]:
        return list(MissingMassSquaredWOVertexFitVariable.definitions.keys())
