"""
Utilities for handling decay reconstruction modes IDs.
"""

from dataclasses import dataclass
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Mapping
from typing import Dict, Tuple, Set, List, Union, Optional, Type, Iterator

from rdstar.utilities.logging import log_to_default

from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses


__all__ = [
    "RecoModeDictType",
    "RecoModeIDToTitleDictType",
    "RecoModeIDTupleToTitleDictType",
    "RecoModeGroupNameToTitleDictType",
    "RecoModeIDInfo",
    "RecoModeGroupInfo",
    "AbstractRecoModeInfosBaseClass",
    "RecoModeCollection",
]


logger = log_to_default(name=__name__)


RecoModeDictType = Dict[str, Tuple[int, ...]]
RecoModeIDToTitleDictType = Dict[int, str]
RecoModeIDTupleToTitleDictType = Dict[Tuple[int, ...], str]
RecoModeGroupNameToTitleDictType = Dict[str, str]


@dataclass(frozen=True)
class RecoModeIDInfo:
    id: int
    name: str
    latex_label: str

    @property
    def ids(self) -> Tuple[int, ...]:
        return (self.id,)


@dataclass(frozen=True)
class RecoModeGroupInfo:
    name: str
    ids: Tuple[int, ...]
    latex_label: str


class AbstractRecoModeInfosBaseClass(ABC):
    _all_reco_modes_latex_label_key: str = "all_modes"
    _all_reco_modes_latex_label: str = LatexDecayProcesses.all_modes_label
    _all_reco_modes_latex_label_short: str = LatexDecayProcesses.all_modes_label_short

    _base_method_names: Tuple[str, ...] = (
        "_get_additional_method_names",
        "get_all_reco_modes_latex_label",
        "get_all_reco_modes_latex_label_short",
        "get_method_names",
        "get_reco_ids_dict",
        "get_reco_id_groups_dict",
        "get_all_reco_mode_ids",
        "get_all_reco_mode_id_group_names",
        "get_reco_mode_name_to_info_mapping",
        "get_reco_mode_name_to_ids_mapping",
        "get_reco_id_to_latex_label",
        "get_reco_mode_name_to_latex_label",
        "get_reco_mode_name_to_info_mapping_for_group",
        "get_all_reco_mode_ids_for_group",
        "get_reco_mode_name_to_ids_mapping_for_group",
        "get_reco_mode_name_to_latex_label_for_group",
        "find_duplicates",
        "print_duplicates",
        "run_dedicated_checks",
        "run_checks",
    )

    @classmethod
    @abstractmethod
    def _get_additional_method_names(cls) -> Tuple[str, ...]:
        raise NotImplementedError(
            f"Method _get_additional_method_names is not implemented for class {cls.__name__}! "
            f"It should return a tuple of method names in addition to the method names stored "
            f"in '_base_method_names'... The tuple can be empty."
        )

    @classmethod
    def get_method_names(cls) -> Tuple[str, ...]:
        return AbstractRecoModeInfosBaseClass._base_method_names + cls._get_additional_method_names()

    @staticmethod
    def get_all_reco_modes_latex_label_key() -> str:
        return AbstractRecoModeInfosBaseClass._all_reco_modes_latex_label_key

    @classmethod
    def get_all_reco_modes_latex_label(cls) -> str:
        return cls._all_reco_modes_latex_label

    @classmethod
    def get_all_reco_modes_latex_label_short(cls) -> str:
        return cls._all_reco_modes_latex_label_short

    @classmethod
    def get_reco_ids_dict(cls) -> Dict[str, RecoModeIDInfo]:
        _method_names: Tuple[str, ...] = cls.get_method_names()
        return {
            attr_name: val
            for attr_name, val in cls.__dict__.items()
            if isinstance(val, RecoModeIDInfo)
            if not ((attr_name in _method_names) or (attr_name.endswith("__") or attr_name.startswith("_")))
        }

    @classmethod
    def get_reco_id_groups_dict(cls) -> Dict[str, RecoModeGroupInfo]:
        _method_names: Tuple[str, ...] = cls.get_method_names()
        return {
            attr_name: val
            for attr_name, val in cls.__dict__.items()
            if isinstance(val, RecoModeGroupInfo)
            if not ((attr_name in _method_names) or (attr_name.endswith("__") or attr_name.startswith("_")))
        }

    @classmethod
    def get_all_reco_mode_ids(cls) -> Tuple[int, ...]:
        return tuple(reco_id_info.id for reco_id_info in cls.get_reco_ids_dict().values())

    @classmethod
    def get_all_reco_mode_id_group_names(cls) -> Tuple[str, ...]:
        return tuple(reco_id_group_info.name for reco_id_group_info in cls.get_reco_id_groups_dict().values())

    @classmethod
    def get_reco_mode_name_to_info_mapping(cls) -> Dict[str, Union[RecoModeIDInfo, RecoModeGroupInfo]]:
        reco_mode_infos: Dict[str, Union[RecoModeIDInfo, RecoModeGroupInfo]] = {
            k: v for k, v in cls.get_reco_ids_dict().items()
        }
        reco_mode_group_infos: Dict[str, RecoModeGroupInfo] = {k: v for k, v in cls.get_reco_id_groups_dict().items()}
        reco_mode_infos.update(reco_mode_group_infos)
        return reco_mode_infos

    @classmethod
    def get_reco_mode_name_to_ids_mapping(cls) -> RecoModeDictType:
        reco_mode_ids: RecoModeDictType = {k: (v.id,) for k, v in cls.get_reco_ids_dict().items()}
        reco_mode_id_groups: RecoModeDictType = {k: v.ids for k, v in cls.get_reco_id_groups_dict().items()}
        reco_mode_ids.update(reco_mode_id_groups)
        return reco_mode_ids

    @classmethod
    def get_reco_id_to_latex_label(cls) -> RecoModeIDToTitleDictType:
        return {v.id: v.latex_label for v in cls.get_reco_ids_dict().values()}

    @classmethod
    def get_reco_mode_name_to_latex_label(cls) -> RecoModeGroupNameToTitleDictType:
        _r: RecoModeGroupNameToTitleDictType = {
            k: v.latex_label for k, v in cls.get_reco_mode_name_to_info_mapping().items()
        }

        assert AbstractRecoModeInfosBaseClass.get_all_reco_modes_latex_label_key() not in _r
        _r.update(
            {AbstractRecoModeInfosBaseClass.get_all_reco_modes_latex_label_key(): cls.get_all_reco_modes_latex_label()}
        )
        return _r

    @classmethod
    def get_reco_mode_name_to_info_mapping_for_group(
        cls,
        group: RecoModeGroupInfo,
    ) -> Dict[str, RecoModeIDInfo]:
        assert group in cls.get_reco_id_groups_dict().values(), (group.name, group.ids)
        reco_mode_id_mapping: Dict[str, RecoModeIDInfo] = cls.get_reco_ids_dict()
        return {k: v for k, v in reco_mode_id_mapping.items() if v.id in group.ids}

    @classmethod
    def get_all_reco_mode_ids_for_group(
        cls,
        group: RecoModeGroupInfo,
    ) -> Tuple[int, ...]:
        assert group in cls.get_reco_id_groups_dict().values(), (group.name, group.ids)
        return group.ids

    @classmethod
    def get_reco_mode_name_to_ids_mapping_for_group(
        cls,
        group: RecoModeGroupInfo,
    ) -> RecoModeDictType:
        reco_mode_id_infos: Dict[str, RecoModeIDInfo] = cls.get_reco_mode_name_to_info_mapping_for_group(group=group)
        return {k: v.ids for k, v in reco_mode_id_infos.items()}

    @classmethod
    def get_reco_mode_name_to_latex_label_for_group(
        cls,
        group: RecoModeGroupInfo,
        all_modes_label: Optional[str] = None,
    ) -> RecoModeGroupNameToTitleDictType:
        _r: RecoModeGroupNameToTitleDictType = {
            k: v.latex_label for k, v in cls.get_reco_mode_name_to_info_mapping_for_group(group=group).items()
        }
        assert AbstractRecoModeInfosBaseClass.get_all_reco_modes_latex_label_key() not in _r

        _all_label: str = cls.get_all_reco_modes_latex_label() if all_modes_label is None else all_modes_label
        _r.update({AbstractRecoModeInfosBaseClass.get_all_reco_modes_latex_label_key(): _all_label})
        return _r

    @classmethod
    def find_duplicates(cls) -> Dict[Tuple[int, ...], List[str]]:
        reco_group_mapping: Dict[str, RecoModeGroupInfo] = cls.get_reco_id_groups_dict()
        all_reco_group_contents: Tuple[Set[int], ...] = tuple(set(g.ids) for g in reco_group_mapping.values())

        duplicates_container: Dict[Tuple[int, ...], List[str]] = defaultdict(list)
        for k, v in reco_group_mapping.items():
            if all_reco_group_contents.count(set(v.ids)) > 1:
                duplicates_container[tuple(set(v.ids))].append(k)

        return duplicates_container

    @classmethod
    def print_duplicates(cls) -> None:
        duplicates_dict: Dict[Tuple[int, ...], List[str]] = cls.find_duplicates()
        print(
            f"Found {len(duplicates_dict)} duplicated reco mode id groups of {cls.__name__}{':' if len(duplicates_dict) else '.'}"
        )
        for k, v in duplicates_dict.items():
            print(f"\t {k}: {v}")

    @classmethod
    def run_dedicated_checks(cls) -> None:
        print(f"No dedicated checks defined for {cls.__name__}...")

    @classmethod
    def run_checks(cls) -> None:
        reco_mode_id_infos: Dict[str, RecoModeIDInfo] = cls.get_reco_ids_dict()
        reco_mode_id_group_infos: Dict[str, RecoModeGroupInfo] = cls.get_reco_id_groups_dict()

        assert all(k == v.name for k, v in reco_mode_id_infos.items()), [
            (k, v) for k, v in reco_mode_id_infos.items() if k != v.name
        ]
        assert all(k == v.name for k, v in reco_mode_id_group_infos.items()), [
            (k, v) for k, v in reco_mode_id_group_infos.items() if k != v.name
        ]

        reco_mode_ids: Tuple[int, ...] = cls.get_all_reco_mode_ids()
        assert len(set(reco_mode_ids)) == len(reco_mode_ids), (
            len(reco_mode_ids),
            len(set(reco_mode_ids)),
            set([i for i in set(reco_mode_ids) if reco_mode_ids.count(i) > 1]),
        )

        _ids_used_in_groups: Tuple[int, ...] = tuple(i for g in reco_mode_id_group_infos.values() for i in g.ids)

        assert all(i in reco_mode_ids for i in _ids_used_in_groups), [
            (g.name, i) for g in reco_mode_id_group_infos.values() for i in g.ids if i not in reco_mode_ids
        ]

        for reco_group in reco_mode_id_group_infos.values():
            assert len(set(reco_group.ids)) == len(reco_group.ids), (
                reco_group.ids,
                set([i for i in set(reco_group.ids) if reco_group.ids.count(i) > 1]),
            )

        all_infos_map: Dict[str, Union[RecoModeIDInfo, RecoModeGroupInfo]] = cls.get_reco_mode_name_to_info_mapping()

        reco_mode_id_to_label_map: RecoModeIDToTitleDictType = cls.get_reco_id_to_latex_label()
        _reco_mode_name_to_label_map: RecoModeGroupNameToTitleDictType = cls.get_reco_mode_name_to_latex_label()
        all_modes_latex_label_key: str = AbstractRecoModeInfosBaseClass.get_all_reco_modes_latex_label_key()
        assert all_modes_latex_label_key in _reco_mode_name_to_label_map, all_modes_latex_label_key

        reco_mode_name_to_label_map: RecoModeGroupNameToTitleDictType = {
            k: v for k, v in _reco_mode_name_to_label_map.items() if k != all_modes_latex_label_key
        }

        assert all(i in reco_mode_ids for i in reco_mode_id_to_label_map), [
            i for i in reco_mode_id_to_label_map if i not in reco_mode_ids
        ]
        assert all(i in all_infos_map for i in reco_mode_name_to_label_map), [
            i for i in reco_mode_name_to_label_map if i not in all_infos_map
        ]

        all_labels: Tuple[str, ...] = tuple(reco_mode_name_to_label_map.values())
        assert len(all_labels) == len(set(all_labels)), (
            len(all_labels),
            len(set(all_labels)),
            set([latex_label for latex_label in set(all_labels) if all_labels.count(latex_label) > 1]),
            [k for k, v in reco_mode_name_to_label_map.items() if all_labels.count(v) > 1],
        )

        cls.run_dedicated_checks()
        cls.print_duplicates()


class RecoModeCollection(Mapping):
    def __init__(
        self,
        reco_mode_info_class: Type[AbstractRecoModeInfosBaseClass],
        reco_mode_infos: Tuple[Union[RecoModeIDInfo, RecoModeGroupInfo], ...],
        all_reco_modes_latex_label: str,
        all_reco_modes_latex_label_short: Optional[str] = None,
        main_reco_mode_infos: Optional[Tuple[Union[RecoModeIDInfo, RecoModeGroupInfo], ...]] = None,
        allow_duplicates: bool = False,
    ) -> None:
        self._reco_mode_info_class: Type[AbstractRecoModeInfosBaseClass] = reco_mode_info_class
        self._reco_mode_infos: Tuple[Union[RecoModeIDInfo, RecoModeGroupInfo], ...] = reco_mode_infos
        self._all_reco_modes_latex_label: str = all_reco_modes_latex_label
        self._all_reco_modes_latex_label_short: Optional[str] = all_reco_modes_latex_label_short
        self._allow_duplicates: bool = allow_duplicates

        if main_reco_mode_infos is not None:
            assert all(v in reco_mode_infos for v in main_reco_mode_infos), (
                [v.name for v in main_reco_mode_infos if v not in reco_mode_infos],
            )
        self._provided_main_reco_mode_infos: Optional[
            Tuple[Union[RecoModeIDInfo, RecoModeGroupInfo], ...]
        ] = main_reco_mode_infos

        self._all_modes_key: str = AbstractRecoModeInfosBaseClass.get_all_reco_modes_latex_label_key()

        all_names: Tuple[str, ...] = tuple(i.name for i in self._reco_mode_infos)
        assert len(all_names) == len(set(all_names)), (
            len(all_names),
            len(set(all_names)),
            set([i for i in set(all_names) if all_names.count(i) > 1]),
        )

        self._mapping: Dict[str, Union[RecoModeIDInfo, RecoModeGroupInfo]] = {i.name: i for i in self._reco_mode_infos}
        assert len(self._mapping) == len(self._reco_mode_infos), (len(self._mapping), len(self._reco_mode_infos))

        _all_ids: Tuple[int, ...] = tuple(i for rmi in self._reco_mode_infos for i in rmi.ids)

        if allow_duplicates:
            assert self._provided_main_reco_mode_infos is not None
            _all_ids = tuple(set(_all_ids))

        self._all_ids: Tuple[int, ...] = _all_ids

        _all_modes_group_info: RecoModeGroupInfo = RecoModeGroupInfo(
            name=self._all_modes_key,
            ids=self._all_ids,
            latex_label=self._all_reco_modes_latex_label,
        )
        self._mapping.update({self.all_modes_key: _all_modes_group_info})

        if self._provided_main_reco_mode_infos is not None:
            assert all(i in _all_ids for g in self._provided_main_reco_mode_infos for i in g.ids), [
                (g.name, [i for i in g.ids if i not in _all_ids])
                for g in self._provided_main_reco_mode_infos
                if not all(i in _all_ids for i in g.ids)
            ]

        _all_ids_from_class: Tuple[int, ...] = self._reco_mode_info_class.get_all_reco_mode_ids()
        assert all(i in _all_ids_from_class for i in _all_ids), (i for i in _all_ids if i not in _all_ids_from_class)
        assert len(_all_ids) == len(set(_all_ids)), (
            len(_all_ids),
            len(set(_all_ids)),
            set([i for i in set(_all_ids) if _all_ids.count(i) > 1]),
        )

    @property
    def reco_mode_info_class(self) -> Type[AbstractRecoModeInfosBaseClass]:
        return self._reco_mode_info_class

    @property
    def reco_mode_infos(self) -> Tuple[Union[RecoModeIDInfo, RecoModeGroupInfo], ...]:
        return self._reco_mode_infos

    @property
    def provided_main_reco_mode_infos(self) -> Optional[Tuple[Union[RecoModeIDInfo, RecoModeGroupInfo], ...]]:
        return self._provided_main_reco_mode_infos

    @property
    def main_reco_mode_infos(self) -> Tuple[Union[RecoModeIDInfo, RecoModeGroupInfo], ...]:
        if self.provided_main_reco_mode_infos is not None:
            return self.provided_main_reco_mode_infos
        else:
            return self.reco_mode_infos

    @property
    def all_modes_key(self) -> str:
        return self._all_modes_key

    @property
    def all_reco_modes_latex_label(self) -> str:
        return self._all_reco_modes_latex_label

    @property
    def all_reco_modes_latex_label_short(self) -> str:
        if self._all_reco_modes_latex_label_short is None:
            return self.all_reco_modes_latex_label
        return self._all_reco_modes_latex_label_short

    @property
    def all_reco_mode_ids(self) -> Tuple[int, ...]:
        return self._all_ids

    def get_latex_label_for_all_modes(self, short: bool = False) -> str:
        if short:
            return self.all_reco_modes_latex_label_short
        return self.all_reco_modes_latex_label

    def get_reco_ids_to_latex_label_mapping(self, short: bool = False) -> RecoModeIDTupleToTitleDictType:
        m: RecoModeIDTupleToTitleDictType = {v.ids: v.latex_label for v in self.values()}
        if short:
            assert self.all_reco_mode_ids in m.keys(), self.all_modes_key
            assert m.get(self.all_reco_mode_ids) == self.all_reco_modes_latex_label, repr(m.get(self.all_reco_mode_ids))
            m[self.all_reco_mode_ids] = self.all_reco_modes_latex_label_short
        return m

    def get_reco_mode_name_to_latex_label_mapping(self, short: bool = False) -> RecoModeGroupNameToTitleDictType:
        _map: RecoModeGroupNameToTitleDictType = {v.name: v.latex_label for v in self.values()}
        if short:
            assert self.all_modes_key in _map.keys(), self.all_modes_key
            assert _map.get(self.all_modes_key) == self.all_reco_modes_latex_label, repr(_map.get(self.all_modes_key))
            _map[self.all_modes_key] = self.all_reco_modes_latex_label_short
        return _map

    def get_reco_mode_title(self, mode_name: str, use_short_titles: bool) -> str:
        return self.get_reco_mode_name_to_latex_label_mapping(short=use_short_titles)[mode_name]

    @property
    def reco_mode_info_mapping_as_tuple(self) -> Tuple[Tuple[str, Tuple[int, ...]], ...]:
        reco_mode_infos = [(k, v.ids) for k, v in self.items()]
        return tuple(reco_mode_infos)

    @property
    def main_reco_mode_infos_as_tuple(self) -> Tuple[Tuple[str, Tuple[int, ...]], ...]:
        return tuple((v.name, v.ids) for v in self.main_reco_mode_infos)

    @property
    def main_reco_mode_infos_mapping(self) -> RecoModeDictType:
        return {k: v for k, v in self.main_reco_mode_infos_as_tuple}

    @property
    def special_reco_mode_id_group_names(self) -> Tuple[str, ...]:
        main_reco_mode_group_names: Tuple[str, ...] = tuple(v.name for v in self.main_reco_mode_infos)
        return tuple(v.name for v in self.reco_mode_infos if v.name not in main_reco_mode_group_names)

    def __getitem__(self, key: str) -> Union[RecoModeIDInfo, RecoModeGroupInfo]:
        return self._mapping[key]

    def __len__(self) -> int:
        return len(self._mapping)

    def __iter__(self) -> Iterator[str]:
        return iter(self._mapping)

    def __repr__(self) -> str:
        rep: str = (
            f"{self.__class__.__name__} based on {self.reco_mode_info_class.__class__.__name__} "
            f"with reco_mode_info_mapping_as_tuple = {repr(self.reco_mode_info_mapping_as_tuple)}"
        )
        if self.main_reco_mode_infos is not None:
            rep += f"\nmain_reco_mode_infos_as_tuple: {repr(self.main_reco_mode_infos_as_tuple)}"
        if self._allow_duplicates:
            rep += "\nDuplicates are allowed!"
        return rep
