"""
Continuum Suppression Tasks

Felix Metzner 2019
"""

import os
import json
import numpy as np
import pandas as pd
import b2luigi as luigi

from PyFastBDT import FastBDT
from recordclass import RecordClass
from luigi.util import inherits, requires
from sklearn.model_selection import train_test_split
from typing import Optional, List, Tuple, Dict, Generator, Any

from analysistools.dataframe_tools import hdf5_utils

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.tasks import utilities
from rdstar import continuum_suppression as cs_tools
from rdstar.offline_analysis.validation import cs_validation
from rdstar.offline_analysis.rdstar_samples import mva_samples
from rdstar.offline_analysis.tasks.logistics_tasks import DataCopyTask
from rdstar.offline_analysis.tasks.preparation_tasks import DataPreparationTask, DataPreparationMergerTask


__all__ = [
    "CSDataPreparationTask",
    "CSTrainTask",
    "CSTestTask",
    "CSValidationTask",
]


logger = log_to_default(name=__name__)


@inherits(DataCopyTask)
class CSDataPreparationTask(utilities.BaseOfflineTask):
    cs_data_tag = luigi.Parameter(positional=False)
    cs_use_signal_side_vars = luigi.BoolParameter(positional=False)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        offline_config = utilities.load_yaml_config(yaml_config_path=self.offline_config_path)
        self.working_dir = offline_config["working_dir"]  # type: PathType
        self.cs_dir = os.path.join(self.working_dir, "CS_Trainings")

        self.cs_base_tuple_type = "Y4S_ntuple"

        self.cs_sig_data_types = mva_samples  # type: List[str]
        self.cs_sig_stream = -999

        self.cs_bkg_data_types = ["UDS", "Charm"]  # type: List[str]
        self.cs_bkg_stream = 0

        self.cs_signal_query = "isSignalAcceptMissingNeutrino == 1"

        self.cs_target_var = offline_config["cs_config"]["cs_target_var"]

        assert isinstance(self.cs_use_signal_side_vars, bool), type(self.cs_use_signal_side_vars)
        cs_variable_dict = cs_tools.get_cs_variable_dict(signal_side=self.cs_use_signal_side_vars)
        self.cs_vars = list(cs_variable_dict.keys())

        self.signal_vars = [
            "m2RecoilSignalSide",
            "daughter__bo1__cm__spuseRestFrame__bodaughter__bo1__cm__spp__bc__bc__bc",
            "isSignalAcceptMissingNeutrino",
        ]

        _vars = self.cs_vars + [self.cs_target_var] + self.signal_vars + cs_tools.cs_additional_variables
        self.vars = list(set(_vars))

        self.cs_vars_tag = "SigSideBased" if self.cs_use_signal_side_vars else "TagSideBased"
        self.cs_info_tag = f"{self.cs_data_tag}_{self.cs_vars_tag}"

        self._input_dict = None  # type: Optional[Dict[str, PathType]]

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    def get_key_pair(self, data_type: str, extension: Optional[str] = None) -> Tuple[str, str]:
        assert data_type in self.cs_sig_data_types + self.cs_bkg_data_types

        if data_type in self.cs_sig_data_types:
            hdf5_key = f"{data_type}_{self.cs_base_tuple_type}"
            file_key = f"{hdf5_key}_{DataPreparationTask.base_file_tag}"
        elif extension is None and data_type in self.cs_bkg_data_types:
            hdf5_key = f"{data_type}_{self.cs_bkg_stream}_{self.cs_base_tuple_type}"
            file_key = f"{hdf5_key}_{DataPreparationTask.base_file_tag}"
        else:
            assert extension in ["test", "train"]
            hdf5_key = f"{data_type}_{self.cs_bkg_stream}_{self.cs_base_tuple_type}"
            file_key = f"{hdf5_key}_forCS_{extension}"

        return file_key, hdf5_key

    def get_data_frames(self, data_types: List[str], extension: Optional[str] = None) -> List[pd.DataFrame]:
        dfs = []
        for file_key, hdf5_key in [self.get_key_pair(d, extension) for d in data_types]:
            dfs.append(pd.read_hdf(path_or_buf=self.input_dict[file_key], key=hdf5_key)[self.vars])
        return dfs

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for data_type in self.cs_bkg_data_types:
            yield self.clone(
                DataPreparationMergerTask,
                data_type=data_type,
                stream=self.cs_bkg_stream,
                tuple_type=self.cs_base_tuple_type,
            )

        for data_type in self.cs_sig_data_types:
            yield self.clone(
                DataPreparationMergerTask,
                data_type=data_type,
                stream=self.cs_sig_stream,
                tuple_type=self.cs_base_tuple_type,
            )

    def cs_preparation_process(self, dry_run: bool = False) -> Dict[str, PathType]:
        cs_train_data_file_name = os.path.join(self.cs_dir, f"cs_train_data_{self.cs_info_tag}.h5")
        cs_test_data_file_name = os.path.join(self.cs_dir, f"cs_test_data_{self.cs_info_tag}.h5")
        tmp_cs_train_data_file_name = os.path.join(self.cs_dir, f"tmp_cs_train_data_{self.cs_info_tag}.h5")
        tmp_cs_test_data_file_name = os.path.join(self.cs_dir, f"tmp_cs_test_data_{self.cs_info_tag}.h5")

        if not dry_run:
            os.makedirs(self.cs_dir, exist_ok=True)

            bkg_test_dfs = self.get_data_frames(self.cs_bkg_data_types, "test")
            bkg_train_dfs = self.get_data_frames(self.cs_bkg_data_types, "train")

            sig_dfs = self.get_data_frames(self.cs_sig_data_types)

            sig_df = pd.concat([df.query(self.cs_signal_query) for df in sig_dfs], ignore_index=True, sort=False)

            logger.info(f"Number of considered Signal events: {len(sig_df.index)}")
            sig_test_df, sig_train_df = train_test_split(sig_df, train_size=0.5)

            train_df = pd.concat(bkg_train_dfs + [sig_train_df], ignore_index=True, sort=False)
            test_df = pd.concat(bkg_test_dfs + [sig_test_df], ignore_index=True, sort=False)

            train_df.to_hdf(path_or_buf=tmp_cs_train_data_file_name, key="cs_train")
            test_df.to_hdf(path_or_buf=tmp_cs_test_data_file_name, key="cs_test")

            os.rename(src=tmp_cs_train_data_file_name, dst=cs_train_data_file_name)
            os.rename(src=tmp_cs_test_data_file_name, dst=cs_test_data_file_name)

        return {
            f"cs_train_{self.cs_info_tag}": cs_train_data_file_name,
            f"cs_test_{self.cs_info_tag}": cs_test_data_file_name,
        }

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.cs_preparation_process()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.cs_preparation_process(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})

        return output_dict


@requires(CSDataPreparationTask)
class CSTrainTask(utilities.BaseOfflineTask):
    cs_tag = luigi.Parameter(positional=False)  # type: str
    cs_fastbdt_settings = luigi.DictParameter(positional=False, hashed=True, significant=False)  # type: Dict[str, Any]

    default_fast_bdt_settings = {
        "nTrees": 200,
    }  # type: Dict[str, int]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        offline_config = utilities.load_yaml_config(yaml_config_path=self.offline_config_path)
        self.working_dir = offline_config["working_dir"]  # type: PathType
        self.cs_dir = os.path.join(self.working_dir, "CS_Trainings")  # type: PathType

        self.cs_target_var = offline_config["cs_config"]["cs_target_var"]

        assert isinstance(self.cs_use_signal_side_vars, bool), type(self.cs_use_signal_side_vars)
        cs_variable_dict = cs_tools.get_cs_variable_dict(signal_side=self.cs_use_signal_side_vars)
        self.cs_vars = list(cs_variable_dict.keys())

        self.cs_vars_tag = "SigSideBased" if self.cs_use_signal_side_vars else "TagSideBased"
        self.cs_info_tag = f"{self.cs_data_tag}_{self.cs_vars_tag}"
        self.cs_training_tag = f"{self.cs_tag}_{self.cs_vars_tag}"

        self._input_dict = None  # type: Optional[Dict[str, PathType]]

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for k, v in self.input().items()}
        return self._input_dict

    def get_sample(self) -> pd.DataFrame:
        train_data_path = self.input_dict[f"cs_train_{self.cs_info_tag}"]
        return hdf5_utils.load_hdf5(hdf5_file_path=train_data_path, hdf5_key="cs_train")

    def run_cs_training(self, dry_run: bool = False) -> Dict[str, PathType]:
        cs_output_weight_file = os.path.join(self.cs_dir, f"cs_classifier_weights_{self.cs_training_tag}.xml")
        cs_var_list_file = os.path.join(self.cs_dir, f"cs_mva_vars_{self.cs_training_tag}.pkl")
        tmp_cs_output_weight_file = os.path.join(self.cs_dir, f"tmp_cs_classifier_weights_{self.cs_training_tag}.xml")
        tmp_cs_var_list_file = os.path.join(self.cs_dir, f"tmp_cs_mva_vars_{self.cs_training_tag}.pkl")

        if not dry_run:
            cs_train_data = self.get_sample()

            cs_train_data.loc[:, "__cs_weights__"] = np.ones(len(cs_train_data.index))

            cs = cs_tools.ContinuumSuppression(
                cs_variables=self.cs_vars,
                target_variable=self.cs_target_var,
                weight="__cs_weights__",
            )

            if not self.cs_fastbdt_settings:
                fast_bdt_settings = CSTrainTask.default_fast_bdt_settings
            else:
                fast_bdt_settings = self.cs_fastbdt_settings

            cs.add_classifier(classifier=FastBDT.Classifier(**fast_bdt_settings))
            cs.train(data=cs_train_data)

            cs.save_classifier_weights(weight_file_path=tmp_cs_output_weight_file, overwrite=True)

            cs_tools.ContinuumSuppression.dump_list_of_variables(
                variable_list=self.cs_vars,
                pickle_file_path=tmp_cs_var_list_file,
            )

            os.rename(src=tmp_cs_output_weight_file, dst=cs_output_weight_file)
            os.rename(src=tmp_cs_var_list_file, dst=cs_var_list_file)

        return {
            f"cs_output_weight_file_{self.cs_training_tag}": cs_output_weight_file,
            f"cs_var_list_file_{self.cs_training_tag}": cs_var_list_file,
        }

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_cs_training()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.run_cs_training(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})

        return output_dict


class BestCSRecord(RecordClass):
    roc_auc_on_test: Optional[float] = None
    roc_auc_on_train: Optional[float] = None
    cs_use_signal_side_vars: Optional[bool] = None
    info_dict_path: Optional[PathType] = None
    cs_var_list_file: Optional[PathType] = None
    cs_output_weight_file: Optional[PathType] = None

    @property
    def parameter_names(self) -> Tuple[str, ...]:
        return tuple(self.__annotations__.keys())

    @property
    def as_dict(self) -> Dict[str, float]:
        return {name: getattr(self, name) for name in self.parameter_names}

    def check_if_all_filled(self) -> None:
        if any(v is None for v in self.as_dict.values()):
            raise RuntimeError(
                f"BestCSRecord is not set properly, as the following attributes are None:\n"
                f"{[k for k, v in self.as_dict.items() if v is None]}"
            )


@inherits(DataCopyTask)
class CSTestTask(utilities.BaseOfflineTask):
    cs_tag = luigi.Parameter(positional=False)  # type: str
    cs_fastbdt_settings = luigi.DictParameter(positional=False, hashed=True, significant=False)  # type: Dict[str, Any]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        offline_config = utilities.load_yaml_config(yaml_config_path=self.offline_config_path)
        self.working_dir = offline_config["working_dir"]  # type: PathType
        self.cs_dir = os.path.join(self.working_dir, "CS_Trainings")  # type: PathType

        self.cs_data_tag = offline_config["cs_config"]["cs_data_tag"]
        self.cs_target_var = offline_config["cs_config"]["cs_target_var"]
        self.cs_output_col_name = offline_config["cs_config"]["cs_column_name"]
        self.cs_use_tag_side_vars = offline_config["cs_config"]["cs_use_tag_side_vars"]

        if self.cs_use_tag_side_vars is None:
            # This means the best one is used.
            self.output_weight_file_to_use = None  # type: Optional[PathType]
            self.var_list_file_to_use = None  # type: Optional[PathType]
        else:
            vars_tag = "TagSideBased" if self.cs_use_tag_side_vars else "SigSideBased"
            train_tag_to_use = f"{self.cs_tag}_{vars_tag}"
            self.output_weight_file_to_use = os.path.join(self.cs_dir, f"cs_classifier_weights_{train_tag_to_use}.xml")
            self.var_list_file_to_use = os.path.join(self.cs_dir, f"cs_mva_vars_{train_tag_to_use}.pkl")

        self.use_sig_side_vars_values = [True, False]  # type: List[bool]

        self.best_cs_info = BestCSRecord()

        self._input_dict = None  # type: Optional[Dict[str, PathType]]

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    def run_cs_evaluation(self, use_sig_side_vars: bool, dry_run: bool = False) -> Dict[str, PathType]:
        cs_vars_tag = "SigSideBased" if use_sig_side_vars else "TagSideBased"
        output_tag = f"{self.cs_tag}_{self.cs_data_tag}_{cs_vars_tag}"

        roc_output_key = f"cs_{output_tag}_roc_validation"
        roc_output_path = os.path.join(self.cs_dir, f"cs_{output_tag}_roc_validation.json")

        cs_info_tag = f"{self.cs_data_tag}_{cs_vars_tag}"
        cs_training_tag = f"{self.cs_tag}_{cs_vars_tag}"
        cs_train_data_path = self.input_dict[f"cs_train_{cs_info_tag}"]
        cs_test_data_path = self.input_dict[f"cs_test_{cs_info_tag}"]
        cs_var_list_file = self.input_dict[f"cs_var_list_file_{cs_training_tag}"]
        cs_output_weight_file = self.input_dict[f"cs_output_weight_file_{cs_training_tag}"]

        cs_val_instance = cs_validation.CSValidation(
            path_train_df=cs_train_data_path,
            path_test_df=cs_test_data_path,
            path_cs_var_list=cs_var_list_file,
            path_cs_bdt_weights=cs_output_weight_file,
            cs_output_col_name=self.cs_output_col_name,
            cs_target_var=self.cs_target_var,
            target_dir_path=self.cs_dir,
            using_sig_side_vars=use_sig_side_vars,
        )

        output_info_list, roc_auc_on_test, roc_auc_on_train = cs_val_instance.run_validation(
            base_name=f"cs_{output_tag}",
            working_point=0.5,
            mva_output_latex_name=cs_validation.CSValidation.default_cs_mva_output_latex_name,
            dry_run=dry_run,
        )

        output_info_dict = {k: p for k, p in output_info_list}  # type: Dict[str, PathType]
        assert roc_output_key not in output_info_dict, roc_output_key
        output_info_dict.update({roc_output_key: roc_output_path})

        if dry_run:
            return output_info_dict

        if self.best_cs_info.roc_auc_on_test is None or roc_auc_on_test > self.best_cs_info.roc_auc_on_test:
            self.best_cs_info.roc_auc_on_test = roc_auc_on_test
            self.best_cs_info.roc_auc_on_train = roc_auc_on_train
            self.best_cs_info.cs_use_signal_side_vars = use_sig_side_vars
            self.best_cs_info.info_dict_path = roc_output_path
            self.best_cs_info.cs_var_list_file = cs_var_list_file
            self.best_cs_info.cs_output_weight_file = cs_output_weight_file

        fast_bdt_settings = (
            {k: v for k, v in self.cs_fastbdt_settings.items()}
            if self.cs_fastbdt_settings
            else CSTrainTask.default_fast_bdt_settings
        )

        result_dict = {
            "cs_tag": self.cs_tag,
            "cs_data_tag": self.cs_data_tag,
            "cs_use_signal_side_vars": use_sig_side_vars,
            "roc_auc_on_test": roc_auc_on_test,
            "roc_auc_on_train": roc_auc_on_train,
            "cs_fastbdt_settings": fast_bdt_settings,
            "cs_var_list_file": cs_var_list_file,
            "cs_output_weight_file": cs_output_weight_file,
        }

        with open(roc_output_path, "w") as result_fp:
            json.dump(result_dict, result_fp, indent=2)

        logger.info(
            f"ROC AUC for {cs_vars_tag} CS tag {self.cs_tag} with CS data {self.cs_data_tag} "
            f"is {roc_auc_on_test} on the test sample and {roc_auc_on_train} on the training sample"
        )

        return output_info_dict

    def run_cs_comparison(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dict = {}  # type: Dict[str, PathType]
        best_roc_output_key = "cs_best_roc_validation"
        best_roc_output_path = os.path.join(self.cs_dir, "cs_best_roc_validation.json")
        best_weight_file_key = "cs_best_output_weight_file"
        best_weight_file_path = os.path.join(self.cs_dir, f"cs_best_classifier_weights_{self.cs_tag}.xml")
        best_var_file_key = "cs_best_var_list_file"
        best_var_file_path = os.path.join(self.cs_dir, f"cs_best_mva_vars_{self.cs_tag}.pkl")
        weight_file_key = "cs_output_weight_file"
        weight_file_path = os.path.join(self.cs_dir, f"cs_classifier_weights_{self.cs_tag}.xml")
        var_file_key = "cs_var_list_file"
        var_file_path = os.path.join(self.cs_dir, f"cs_mva_vars_{self.cs_tag}.pkl")

        output_dict.update(
            {
                best_roc_output_key: best_roc_output_path,
                best_weight_file_key: best_weight_file_path,
                best_var_file_key: best_var_file_path,
                weight_file_key: weight_file_path,
                var_file_key: var_file_path,
            }
        )

        if not dry_run:
            for use_sig_side_vars in self.use_sig_side_vars_values:
                new_output = self.run_cs_evaluation(use_sig_side_vars=use_sig_side_vars, dry_run=dry_run)
                assert all(output_key not in output_dict for output_key in new_output.keys()), (output_dict, new_output)
                output_dict.update(new_output)

            self.best_cs_info.check_if_all_filled()

            assert self.best_cs_info.info_dict_path is not None
            assert self.best_cs_info.cs_output_weight_file is not None
            assert self.best_cs_info.cs_var_list_file is not None

            if os.path.exists(best_roc_output_path):
                os.remove(best_roc_output_path)
            os.symlink(self.best_cs_info.info_dict_path, best_roc_output_path)
            if os.path.exists(best_weight_file_path):
                os.remove(best_weight_file_path)
            os.symlink(self.best_cs_info.cs_output_weight_file, best_weight_file_path)
            if os.path.exists(best_var_file_path):
                os.remove(best_var_file_path)
            os.symlink(self.best_cs_info.cs_var_list_file, best_var_file_path)

            if self.output_weight_file_to_use is None:
                self.output_weight_file_to_use = self.best_cs_info.cs_output_weight_file
            if os.path.exists(weight_file_path):
                os.remove(weight_file_path)
            os.symlink(self.output_weight_file_to_use, weight_file_path)
            if self.var_list_file_to_use is None:
                self.var_list_file_to_use = self.best_cs_info.cs_var_list_file
            if os.path.exists(var_file_path):
                os.remove(var_file_path)
            os.symlink(self.var_list_file_to_use, var_file_path)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_cs_comparison()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.run_cs_comparison(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})

        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for use_sig_side_vars in self.use_sig_side_vars_values:
            yield self.clone(
                CSTrainTask,
                cs_data_tag=self.cs_data_tag,
                cs_use_signal_side_vars=use_sig_side_vars,
            )
            yield self.clone(
                CSDataPreparationTask,
                cs_data_tag=self.cs_data_tag,
                cs_use_signal_side_vars=use_sig_side_vars,
            )


@inherits(DataCopyTask)
class CSValidationTask(utilities.BaseOfflineTask):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        offline_config = utilities.load_yaml_config(yaml_config_path=self.offline_config_path)
        self.working_dir = offline_config["working_dir"]  # type: PathType
        self.cs_val_dir = os.path.join(self.working_dir, "CS_Validations")  # type: PathType

        self.cs_tag = offline_config["cs_config"]["cs_tag"]  # type: str
        self.cs_fastbdt_settings = offline_config["cs_config"]["cs_fastbdt_settings"]  # type: Dict[str, Any]
        self.cs_base_cut_value = offline_config["cs_config"]["cs_cut_value"]  # type: float

        self.cs_data_tag = offline_config["cs_config"]["cs_data_tag"]  # type: str
        self.cs_target_var = offline_config["cs_config"]["cs_target_var"]  # type: str
        self.cs_output_col_name = offline_config["cs_config"]["cs_column_name"]  # type: str

        self.use_sig_side_vars_values = [True, False]  # type: List[bool]

        self._input_dict = None  # type: Optional[Dict[str, PathType]]

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    def run_cs_evaluation(self, use_sig_side_vars: bool, dry_run: bool = False) -> Dict[str, PathType]:
        cs_vars_tag = "SigSideBased" if use_sig_side_vars else "TagSideBased"  # type: str
        output_tag = f"{self.cs_tag}_{self.cs_data_tag}_{cs_vars_tag}"  # type: str

        sub_output_path = os.path.join(self.cs_val_dir, cs_vars_tag)  # type: PathType

        roc_output_key = f"cs_{output_tag}_roc_validation"  # type: str
        roc_output_path = os.path.join(sub_output_path, f"{roc_output_key}.json")  # type: PathType

        os.makedirs(sub_output_path, exist_ok=True)

        cs_info_tag = f"{self.cs_data_tag}_{cs_vars_tag}"  # type: str
        cs_training_tag = f"{self.cs_tag}_{cs_vars_tag}"  # type: str
        cs_train_data_path = self.input_dict[f"cs_train_{cs_info_tag}"]  # type: PathType
        cs_test_data_path = self.input_dict[f"cs_test_{cs_info_tag}"]  # type: PathType
        cs_var_list_file = self.input_dict[f"cs_var_list_file_{cs_training_tag}"]  # type: PathType
        cs_output_weight_file = self.input_dict[f"cs_output_weight_file_{cs_training_tag}"]  # type: PathType

        cs_val_instance = cs_validation.CSValidation(
            path_train_df=cs_train_data_path,
            path_test_df=cs_test_data_path,
            path_cs_var_list=cs_var_list_file,
            path_cs_bdt_weights=cs_output_weight_file,
            cs_output_col_name=self.cs_output_col_name,
            cs_target_var=self.cs_target_var,
            target_dir_path=sub_output_path,
            using_sig_side_vars=use_sig_side_vars,
            additional_options={
                "correlation_figure_size": (12, 7),
                "correlation_aspect_ratio": "auto",
                "alternative_working_points": (self.cs_base_cut_value,),
            },
        )

        output_info_list, roc_auc_on_test, roc_auc_on_train = cs_val_instance.run_validation(
            base_name=f"cs_{output_tag}",
            working_point=0.5,
            mva_output_latex_name=cs_validation.CSValidation.default_cs_mva_output_latex_name,
            dry_run=dry_run,
        )

        output_info_dict = {k: p for k, p in output_info_list}  # type: Dict[str, PathType]
        assert roc_output_key not in output_info_dict, roc_output_key
        output_info_dict.update({roc_output_key: roc_output_path})

        if dry_run:
            return output_info_dict

        fast_bdt_settings = (
            {k: v for k, v in self.cs_fastbdt_settings.items()}
            if self.cs_fastbdt_settings
            else CSTrainTask.default_fast_bdt_settings
        )

        result_dict = {
            "cs_tag": self.cs_tag,
            "cs_data_tag": self.cs_data_tag,
            "cs_use_signal_side_vars": use_sig_side_vars,
            "roc_auc_on_test": roc_auc_on_test,
            "roc_auc_on_train": roc_auc_on_train,
            "cs_fastbdt_settings": fast_bdt_settings,
            "cs_var_list_file": cs_var_list_file,
            "cs_output_weight_file": cs_output_weight_file,
        }

        with open(roc_output_path, "w") as result_fp:
            json.dump(result_dict, result_fp, indent=2)

        return output_info_dict

    def run_cs_validation(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dict = {}  # type: Dict[str, PathType]

        for use_sig_side_vars in self.use_sig_side_vars_values:
            new_output = self.run_cs_evaluation(use_sig_side_vars=use_sig_side_vars, dry_run=dry_run)
            assert all(output_key not in output_dict for output_key in new_output.keys()), (output_dict, new_output)
            assert all(out_p not in output_dict.values() for out_p in new_output.values()), (output_dict, new_output)
            output_dict.update(new_output)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.cs_val_dir, exist_ok=True)
        self.run_cs_validation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.run_cs_validation(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})

        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for use_sig_side_vars in self.use_sig_side_vars_values:
            yield self.clone(
                CSTrainTask,
                cs_data_tag=self.cs_data_tag,
                cs_use_signal_side_vars=use_sig_side_vars,
            )
            yield self.clone(
                CSDataPreparationTask,
                cs_data_tag=self.cs_data_tag,
                cs_use_signal_side_vars=use_sig_side_vars,
            )
