"""
Tasks producing the Results of the Analysis

Felix Metzner 2019
"""

import os
import b2luigi as luigi

from luigi.util import inherits

from typing import Optional, Tuple, List, Dict, Generator

from rdstar.utilities import PathType
from rdstar.offline_analysis import rdstar_samples
from rdstar.offline_analysis import rdstar_sample_weights
from rdstar.offline_analysis.tasks.logistics_tasks import DataCopyTask
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.efficiency_calculation.efficiency_calculator import EfficiencyCalculator

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.selection_tasks import DataSelectionTask
from rdstar.offline_analysis.tasks.weight_tasks import WeightValidationWrapperTask, WeightInfoGathererTask


__all__ = [
    "EfficiencyEvaluationTask",
]


@inherits(WeightValidationWrapperTask)
class EfficiencyEvaluationTask(utilities.BaseOfflineTask):
    base_dir_name = "Efficiencies"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_path = os.path.join(self.working_dir, self.base_dir_name)

        self.selection_tags = SelectionSetups.get_selection_tags(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
        )  # type: Tuple[str, ...]

        self._file_dict_dir_path = DataCopyTask.get_local_file_dict_path_from_main_dir(
            main_work_dir_path=self.working_dir
        )  # type: PathType

        self._input_dict = None  # type: Optional[Dict[str, PathType]]

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    @property
    def flag_info_dict_path(self) -> PathType:
        return WeightManager.get_flag_info_dict_path_from_task_input_dict(
            input_dict=self.input_dict,
            selection_name=self.selection_name,
        )

    @staticmethod
    def get_relevant_data_types() -> List[str]:
        relevant_data_types = rdstar_samples.efficiency_samples
        assert all(k in rdstar_sample_weights.signal_mc_evts.keys() for k in relevant_data_types)
        return relevant_data_types

    def get_efficiency_samples(self, selection_tag: str) -> List[Tuple[str, PathType]]:
        sample_key_path_list = [
            (k.split("_", 1)[0], v)
            for k, v in self.input_dict.items()
            if k.endswith(selection_tag) and str(v).endswith(".h5")
        ]
        return sample_key_path_list

    def create_efficiency_dictionaries(self, dry_run: bool = False) -> List[Tuple[str, PathType]]:
        output_tuples = []  # type: List[Tuple[str, PathType]]

        for selection_tag in self.selection_tags:
            output_dir = os.path.join(self.base_path, f"{self.selection_name}_{selection_tag}")
            efficiency_calculator = EfficiencyCalculator(
                selection_tag=selection_tag,
                output_dir_path=output_dir,
                flag_info_dict_path=self.flag_info_dict_path,
                file_dict_dir_path=self._file_dict_dir_path,
                weight_column_names=EfficiencyCalculator.default_weight_columns,
            )

            output_tuples.append(efficiency_calculator.get_output_key_file_path_pair())
            if not dry_run:
                os.makedirs(output_dir, exist_ok=True)
                efficiency_dict = efficiency_calculator.evaluate_efficiencies(
                    efficiency_sample_tuples=self.get_efficiency_samples(selection_tag=selection_tag)
                )
                efficiency_calculator.atomic_save(efficiency_dict=efficiency_dict)

        return output_tuples

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)

        os.makedirs(self.base_path, exist_ok=True)
        self.create_efficiency_dictionaries()

    def output(self) -> Dict[str, PathType]:
        return {
            output_key: luigi.LocalTarget(output_path)
            for output_key, output_path in self.create_efficiency_dictionaries(dry_run=True)
        }

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(WeightInfoGathererTask, weight_key=WeightManager.pid_weight_info_key)

        used_data_types = []
        for _, _, relevant_configs in self.config_combinations:
            for params in relevant_configs:
                if params["data_type"] in self.get_relevant_data_types():
                    used_data_types.append(params["data_type"])
                    yield DataSelectionTask(selection_name=self.selection_name, **params)
        if len(used_data_types) < 1:
            raise RuntimeError(
                f"No data for efficiency evaluation available!\n"
                f"Please ensure that all of the following data types are processed:"
                f"\n {self.get_relevant_data_types()}"
            )
