"""
Analysis Flow Validation -- Confusion Matrices

Felix Metzner 2019
"""

import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt

from dataclasses import dataclass
from typing import NamedTuple, Union, Optional, Callable, Tuple, List, Dict, Any

from templatefitter.plotter.plot_utilities import FigureType, AxesType

from rdstar.latex_labels.latex_observable_names import LatexObsNames

from rdstar.offline_analysis import rdstar_samples
from rdstar.offline_analysis.mc_matching.old_mc_matching_info import OldMCMatchingIDsQueryInfo
from rdstar.offline_analysis.generator_level_mc_matching.gen_mc_matching_info import GenMCMatchingInfo

from rdstar.offline_analysis.rdstar_plots import rdstar_color_map
from rdstar.offline_analysis.reco_modes.reco_mode_infos import RecoModeInfos, U4SRecoModeInfos

from rdstar.offline_analysis.selection.selection_evaluation_information_base import EvalInfoContainerBase
from rdstar.offline_analysis.selection.selection_evaluation_info_container import (
    SelectionEvaluationContainer,
    SelectionEvaluationContainerEntry,
)


__all__ = [
    "ConfusionMatrixInfo",
    "confusion_matrix_info_collection",
    "create_confusion_matrix_plot",
    "get_confusion_matrix",
]


def gen_id_formatter(label: Union[str, int, float]) -> str:
    return GenMCMatchingInfo.gen_id_to_latex_label()[int(label)]


def sig_id_formatter(label: Union[str, int, float]) -> str:
    return OldMCMatchingIDsQueryInfo.get_sig_id_to_latex_label()[int(label)]


def per_gen_id_formatter(label: Union[str, int, float]) -> str:
    gen_id_to_latex_label_dict: Dict[int, str] = GenMCMatchingInfo.gen_id_to_latex_label()
    if isinstance(label, int):
        return gen_id_to_latex_label_dict[label]
    assert isinstance(label, str), (label, type(label).__name__)
    return gen_id_to_latex_label_dict[int(label.rsplit("_", 1)[1])]


def per_sig_id_formatter(label: Union[str, int, float]) -> str:
    if isinstance(label, int):
        return OldMCMatchingIDsQueryInfo.get_sig_id_to_latex_label()[label]
    assert isinstance(label, str), (label, type(label).__name__)
    return OldMCMatchingIDsQueryInfo.get_sig_id_to_latex_label()[int(label.rsplit("_", 1)[1])]


def decay_mode_id_formatter(label: Union[str, int, float]) -> str:
    return RecoModeInfos.get_reco_id_to_latex_label()[int(label)]


def u4s_mode_id_formatter(label: Union[str, int, float]) -> str:
    return U4SRecoModeInfos.get_reco_id_to_latex_label()[int(label)]


class BaseInfo(NamedTuple):
    key: str
    info_type: str = "mc_info"

    def __str__(self) -> str:
        return str(self._asdict())


class NormInfo(NamedTuple):
    key: str
    info_type: str

    def __str__(self) -> str:
        return str(self._asdict())


@dataclass(frozen=True)
class ConfusionMatrixInfo:
    name: str
    base_info: BaseInfo
    norm_info: NormInfo
    norm_y: bool = True
    x_label: Optional[str] = None
    y_label: Optional[str] = None
    title: Optional[str] = None
    x_format_func: Optional[Callable] = None
    y_format_func: Optional[Callable] = None
    transpose: bool = False
    remove_keys: Optional[List[str]] = None

    def __post_init__(self) -> None:
        if self.norm_y is False and "probe" not in self.base_info.key:
            raise RuntimeError("Inverted normalization only works for probed information!")


default_remove_keys = [
    *rdstar_samples.alternative_gap_samples,
    *rdstar_samples.efficiency_samples,
    *rdstar_samples.mva_samples,
    *rdstar_samples.off_resonance_mc_samples,
    *rdstar_samples.unused_samples,
]  # type: List[str]


confusion_matrix_info_collection = [
    # Gen ID vs Sig ID
    ConfusionMatrixInfo(
        name="gen_id_vs_sig_id",
        base_info=BaseInfo(key="candidates_per_sig_id_probe_per_gen_id"),
        norm_info=NormInfo(key="candidates_per_gen_id", info_type="mc_info"),
        norm_y=True,
        x_label=LatexObsNames.mc_match_sig_id,
        y_label=LatexObsNames.mc_match_gen_id,
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=gen_id_formatter,
        transpose=False,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_gen_id_vs_sig_id",
        base_info=BaseInfo(key="events_per_sig_id_probe_per_gen_id"),
        norm_info=NormInfo(key="events_per_gen_id", info_type="mc_info"),
        norm_y=True,
        x_label=LatexObsNames.mc_match_sig_id,
        y_label=LatexObsNames.mc_match_gen_id,
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=gen_id_formatter,
        transpose=False,
        remove_keys=None,
    ),
    # Data Type vs Sig ID
    ConfusionMatrixInfo(
        name="data_type_all_vs_sig_id",
        base_info=BaseInfo(key="candidates_per_sig_id_probe_per_data_type"),
        norm_info=NormInfo(key="candidates_per_data_type", info_type="basic_info"),
        norm_y=True,
        x_label=LatexObsNames.mc_match_sig_id,
        y_label=LatexObsNames.data_type,
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=None,
        transpose=False,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_data_type_all_vs_sig_id",
        base_info=BaseInfo(key="events_per_sig_id_probe_per_data_type"),
        norm_info=NormInfo(key="events_per_data_type", info_type="basic_info"),
        norm_y=True,
        x_label=LatexObsNames.mc_match_sig_id,
        y_label=LatexObsNames.data_type,
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=None,
        transpose=False,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="data_type_vs_sig_id",
        base_info=BaseInfo(key="candidates_per_sig_id_probe_per_data_type"),
        norm_info=NormInfo(key="candidates_per_data_type", info_type="basic_info"),
        norm_y=True,
        x_label=LatexObsNames.mc_match_sig_id,
        y_label=LatexObsNames.data_type,
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=None,
        transpose=False,
        remove_keys=default_remove_keys,
    ),
    ConfusionMatrixInfo(
        name="unique_data_type_vs_sig_id",
        base_info=BaseInfo(key="events_per_sig_id_probe_per_data_type"),
        norm_info=NormInfo(key="events_per_data_type", info_type="basic_info"),
        norm_y=True,
        x_label=LatexObsNames.mc_match_sig_id,
        y_label=LatexObsNames.data_type,
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=None,
        transpose=False,
        remove_keys=default_remove_keys,
    ),
    ConfusionMatrixInfo(
        name="data_type_all_vs_sig_id_n_all",
        base_info=BaseInfo(key="candidates_per_sig_id_probe_per_data_type"),
        norm_info=NormInfo(key="candidates", info_type="basic_info"),
        norm_y=True,
        x_label=LatexObsNames.mc_match_sig_id,
        y_label=LatexObsNames.data_type,
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=None,
        transpose=False,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_data_type_all_vs_sig_id_n_all",
        base_info=BaseInfo(key="events_per_sig_id_probe_per_data_type"),
        norm_info=NormInfo(key="events", info_type="basic_info"),
        norm_y=False,
        x_label=LatexObsNames.mc_match_sig_id,
        y_label=LatexObsNames.data_type,
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=None,
        transpose=False,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="data_type_vs_sig_id_n_all",
        base_info=BaseInfo(key="candidates_per_sig_id_probe_per_data_type"),
        norm_info=NormInfo(key="candidates", info_type="basic_info"),
        norm_y=True,
        x_label=LatexObsNames.mc_match_sig_id,
        y_label=LatexObsNames.data_type,
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=None,
        transpose=False,
        remove_keys=default_remove_keys,
    ),
    ConfusionMatrixInfo(
        name="unique_data_type_vs_sig_id_n_all",
        base_info=BaseInfo(key="events_per_sig_id_probe_per_data_type"),
        norm_info=NormInfo(key="events", info_type="basic_info"),
        norm_y=False,
        x_label=LatexObsNames.mc_match_sig_id,
        y_label=LatexObsNames.data_type,
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=None,
        transpose=False,
        remove_keys=default_remove_keys,
    ),
    # Gen ID vs Data Type
    ConfusionMatrixInfo(
        name="gen_id_vs_data_type_all",
        base_info=BaseInfo(key="candidates_per_gen_id_per_data_type"),
        norm_info=NormInfo(key="candidates_per_gen_id", info_type="mc_info"),
        norm_y=True,
        x_label=LatexObsNames.mc_match_gen_id,
        y_label=LatexObsNames.data_type,
        title=None,
        x_format_func=gen_id_formatter,
        y_format_func=None,
        transpose=True,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_gen_id_vs_data_type_all",
        base_info=BaseInfo(key="events_per_gen_id_per_data_type"),
        norm_info=NormInfo(key="events_per_gen_id", info_type="mc_info"),
        norm_y=True,
        x_label=LatexObsNames.mc_match_gen_id,
        y_label=LatexObsNames.data_type,
        title=None,
        x_format_func=gen_id_formatter,
        y_format_func=None,
        transpose=True,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="gen_id_vs_data_type",
        base_info=BaseInfo(key="candidates_per_gen_id_per_data_type"),
        norm_info=NormInfo(key="candidates_per_gen_id", info_type="mc_info"),
        norm_y=True,
        x_label=LatexObsNames.mc_match_gen_id,
        y_label=LatexObsNames.data_type,
        title=None,
        x_format_func=gen_id_formatter,
        y_format_func=None,
        transpose=True,
        remove_keys=default_remove_keys,
    ),
    ConfusionMatrixInfo(
        name="unique_gen_id_vs_data_type",
        base_info=BaseInfo(key="events_per_gen_id_per_data_type"),
        norm_info=NormInfo(key="events_per_gen_id", info_type="mc_info"),
        norm_y=True,
        x_label=LatexObsNames.mc_match_gen_id,
        y_label=LatexObsNames.data_type,
        title=None,
        x_format_func=gen_id_formatter,
        y_format_func=None,
        transpose=True,
        remove_keys=default_remove_keys,
    ),
    # Decay Mode ID vs Sig ID
    ConfusionMatrixInfo(
        name="decay_mode_id_vs_sig_id",
        base_info=BaseInfo(key="candidates_per_sig_id_probe_per_decay_id"),
        norm_info=NormInfo(key="candidates_per_decay_mode_id", info_type="basic_info"),
        norm_y=True,
        x_label=LatexObsNames.mc_match_sig_id,
        y_label=LatexObsNames.reco_mode_id,
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=decay_mode_id_formatter,
        transpose=False,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_decay_mode_id_vs_sig_id",
        base_info=BaseInfo(key="events_per_sig_id_probe_per_decay_id"),
        norm_info=NormInfo(key="events_per_decay_mode_id", info_type="basic_info"),
        norm_y=True,
        x_label=LatexObsNames.mc_match_sig_id,
        y_label=LatexObsNames.reco_mode_id,
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=decay_mode_id_formatter,
        transpose=False,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="decay_mode_id_vs_sig_id_n",
        base_info=BaseInfo(key="candidates_per_sig_id_probe_per_decay_id"),
        norm_info=NormInfo(key="candidates_per_sig_id", info_type="mc_info"),
        norm_y=False,
        x_label=LatexObsNames.mc_match_sig_id,
        y_label=LatexObsNames.reco_mode_id,
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=decay_mode_id_formatter,
        transpose=False,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_decay_mode_id_vs_sig_id_n",
        base_info=BaseInfo(key="events_per_sig_id_probe_per_decay_id"),
        norm_info=NormInfo(key="events_per_sig_id", info_type="mc_info"),
        norm_y=False,
        x_label=LatexObsNames.mc_match_sig_id,
        y_label=LatexObsNames.reco_mode_id,
        title=None,
        x_format_func=per_sig_id_formatter,
        y_format_func=decay_mode_id_formatter,
        transpose=False,
        remove_keys=None,
    ),
    # Sig ID vs Decay Mode ID
    ConfusionMatrixInfo(
        name="sig_id_vs_decay_mode_id",
        base_info=BaseInfo(key="candidates_per_sig_id_probe_per_decay_id"),
        norm_info=NormInfo(key="candidates_per_sig_id", info_type="mc_info"),
        norm_y=False,
        x_label=LatexObsNames.reco_mode_id,
        y_label=LatexObsNames.mc_match_sig_id,
        title=None,
        x_format_func=decay_mode_id_formatter,
        y_format_func=per_sig_id_formatter,
        transpose=True,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_sig_id_vs_decay_mode_id",
        base_info=BaseInfo(key="events_per_sig_id_probe_per_decay_id"),
        norm_info=NormInfo(key="events_per_sig_id", info_type="mc_info"),
        norm_y=False,
        x_label=LatexObsNames.reco_mode_id,
        y_label=LatexObsNames.mc_match_sig_id,
        title=None,
        x_format_func=decay_mode_id_formatter,
        y_format_func=per_sig_id_formatter,
        transpose=True,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="sig_id_vs_decay_mode_id_n",
        base_info=BaseInfo(key="candidates_per_sig_id_probe_per_decay_id"),
        norm_info=NormInfo(key="candidates_per_decay_mode_id", info_type="basic_info"),
        norm_y=True,
        x_label=LatexObsNames.reco_mode_id,
        y_label=LatexObsNames.mc_match_sig_id,
        title=None,
        x_format_func=decay_mode_id_formatter,
        y_format_func=per_sig_id_formatter,
        transpose=True,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_sig_id_vs_decay_mode_id_n",
        base_info=BaseInfo(key="events_per_sig_id_probe_per_decay_id"),
        norm_info=NormInfo(key="events_per_decay_mode_id", info_type="basic_info"),
        norm_y=True,
        x_label=LatexObsNames.reco_mode_id,
        y_label=LatexObsNames.mc_match_sig_id,
        title=None,
        x_format_func=decay_mode_id_formatter,
        y_format_func=per_sig_id_formatter,
        transpose=True,
        remove_keys=None,
    ),
    # Gen ID vs Decay Mode ID
    ConfusionMatrixInfo(
        name="gen_id_vs_decay_mode_id",
        base_info=BaseInfo(key="candidates_per_gen_id_probe_per_decay_id"),
        norm_info=NormInfo(key="candidates_per_gen_id", info_type="mc_info"),
        norm_y=False,
        x_label=LatexObsNames.reco_mode_id,
        y_label=LatexObsNames.mc_match_gen_id,
        title=None,
        x_format_func=decay_mode_id_formatter,
        y_format_func=per_gen_id_formatter,
        transpose=True,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_gen_id_vs_decay_mode_id",
        base_info=BaseInfo(key="events_per_gen_id_probe_per_decay_id"),
        norm_info=NormInfo(key="events_per_gen_id", info_type="mc_info"),
        norm_y=False,
        x_label=LatexObsNames.reco_mode_id,
        y_label=LatexObsNames.mc_match_gen_id,
        title=None,
        x_format_func=decay_mode_id_formatter,
        y_format_func=per_gen_id_formatter,
        transpose=True,
        remove_keys=None,
    ),
    # Sig ID vs Tag Decay Mode ID
    ConfusionMatrixInfo(
        name="sig_id_vs_tag_mode_id",
        base_info=BaseInfo(key="candidates_per_sig_id_probe_per_tag_id"),
        norm_info=NormInfo(key="candidates_per_sig_id", info_type="mc_info"),
        norm_y=False,
        x_label=LatexObsNames.reco_b_tag_reco_mode_id,
        y_label=LatexObsNames.mc_match_sig_id,
        title=None,
        x_format_func=None,
        y_format_func=per_sig_id_formatter,
        transpose=True,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_sig_id_vs_tag_mode_id",
        base_info=BaseInfo(key="events_per_sig_id_probe_per_tag_id"),
        norm_info=NormInfo(key="events_per_sig_id", info_type="mc_info"),
        norm_y=False,
        x_label=LatexObsNames.reco_b_tag_reco_mode_id,
        y_label=LatexObsNames.mc_match_sig_id,
        title=None,
        x_format_func=None,
        y_format_func=per_sig_id_formatter,
        transpose=True,
        remove_keys=None,
    ),
    # Sig ID vs U4S Decay Mode ID
    ConfusionMatrixInfo(
        name="sig_id_vs_u4s_mode_id",
        base_info=BaseInfo(key="candidates_per_sig_id_probe_per_u4s_id"),
        norm_info=NormInfo(key="candidates_per_sig_id", info_type="mc_info"),
        norm_y=False,
        x_label=LatexObsNames.reco_y4s_reco_mode_id,
        y_label=LatexObsNames.mc_match_sig_id,
        title=None,
        x_format_func=u4s_mode_id_formatter,
        y_format_func=per_sig_id_formatter,
        transpose=True,
        remove_keys=None,
    ),
    ConfusionMatrixInfo(
        name="unique_sig_id_vs_u4s_mode_id",
        base_info=BaseInfo(key="events_per_sig_id_probe_per_u4s_id"),
        norm_info=NormInfo(key="events_per_sig_id", info_type="mc_info"),
        norm_y=False,
        x_label=LatexObsNames.reco_y4s_reco_mode_id,
        y_label=LatexObsNames.mc_match_sig_id,
        title=None,
        x_format_func=u4s_mode_id_formatter,
        y_format_func=per_sig_id_formatter,
        transpose=True,
        remove_keys=None,
    ),
    # Tag Decay Mode ID
    # U4S Decay Mode ID
]  # type: List[ConfusionMatrixInfo]


def create_confusion_matrix_plot(
    confusion_matrix: pd.DataFrame,
    confusion_matrix_info: ConfusionMatrixInfo,
) -> Tuple[FigureType, AxesType]:
    if confusion_matrix_info.transpose:
        confusion_matrix = confusion_matrix.transpose()

    fig_size = (10.0, 4.5)  # type: Tuple[float, float]
    cb_fraction = 0.15
    if len(confusion_matrix.columns) > 15 or len(confusion_matrix.index) > 10:
        new_x = fig_size[0]
        new_y = fig_size[1]
        if len(confusion_matrix.columns) > 15:
            new_x = (1.0 + (len(confusion_matrix.columns) - 15) * 1.0 / 15) * fig_size[0]
        if len(confusion_matrix.index) > 10:
            new_y = (1.0 + (len(confusion_matrix.index) - 10) * 1.0 / 10) * fig_size[1]
            cb_fraction *= new_x / new_y
        fig_size = (new_x, new_y)
    fig = plt.figure(num=None, figsize=fig_size, dpi=100, facecolor="w", edgecolor="k")

    ax = plt.gca()

    # Alternative to seaborn heatmap:
    # Maybe switch from seaborn heatmap to matplotlib implementation, e.g. from here:
    # https://matplotlib.org/3.1.1/gallery/images_contours_and_fields/image_annotated_heatmap.html
    ax = sns.heatmap(
        data=confusion_matrix,
        cmap=rdstar_color_map,
        mask=confusion_matrix == 0.0,
        annot=True,
        ax=ax,
        xticklabels=[confusion_matrix_info.x_format_func(t) for t in confusion_matrix.columns]
        if confusion_matrix_info.x_format_func is not None
        else "auto",
        yticklabels=[confusion_matrix_info.y_format_func(t) for t in list(confusion_matrix.index)]
        if confusion_matrix_info.y_format_func is not None
        else "auto",
        cbar_kws={"fraction": cb_fraction},
    )

    ax.set_yticklabels(ax.get_yticklabels(), rotation=0, va="center", rotation_mode="anchor")
    ax.set_xticklabels(ax.get_xticklabels(), rotation=-35, ha="left", rotation_mode="anchor")

    if confusion_matrix_info.title is not None:
        ax.set_title(label=confusion_matrix_info.title, loc="center")
    if confusion_matrix_info.x_label is not None:
        ax.set_xlabel(confusion_matrix_info.x_label)
    if confusion_matrix_info.y_label is not None:
        ax.set_ylabel(confusion_matrix_info.y_label)

    ax.tick_params(axis="both", which="major", bottom=True, top=False, left=True, right=False, direction="out")
    ax.tick_params(axis="both", which="minor", bottom=False, top=False, left=False, right=False)

    return fig, ax


def get_confusion_matrix(
    info_container: SelectionEvaluationContainer,
    confusion_matrix_info: ConfusionMatrixInfo,
) -> pd.DataFrame:
    info_entry_for_full_sel = info_container.info_entry_for_full_selection  # type: SelectionEvaluationContainerEntry

    key_identifier = confusion_matrix_info.base_info  # type: BaseInfo
    norm_key_identifier = confusion_matrix_info.norm_info  # type: NormInfo
    assert norm_key_identifier._asdict() != key_identifier._asdict(), (norm_key_identifier, key_identifier)

    if key_identifier.info_type == "mc_info":
        _base_container = info_entry_for_full_sel.get_sub_info_by_key(key="mc_info")
        assert _base_container is not None
        base_container = _base_container  # type: EvalInfoContainerBase
    else:
        raise RuntimeError(f"Unknown base info type {key_identifier.info_type}")

    assert key_identifier.key in base_container.entry_names, (key_identifier, base_container.entry_names)

    if norm_key_identifier.info_type in ("mc_info", "basic_info"):
        _norm_container = info_entry_for_full_sel.get_sub_info_by_key(key=norm_key_identifier.info_type)
        assert _norm_container is not None
        norm_container = _norm_container  # type: EvalInfoContainerBase
    else:
        raise RuntimeError(f"Unknown norm info type {norm_key_identifier.info_type}")

    assert norm_key_identifier.key in norm_container.entry_names, (norm_key_identifier, norm_container.entry_names)

    base_dict = base_container.get_value(key=key_identifier.key)
    assert isinstance(base_dict, dict)
    assert all(isinstance(v, dict) for v in base_dict.values())
    assert all(isinstance(vv, int) for v in base_dict.values() for vv in v.values())

    return get_conf_matrix_from_one_dict(
        base_dict=base_dict,
        norm_info_value=norm_container.get_value(key=norm_key_identifier.key),
        norm_over_y=confusion_matrix_info.norm_y,
        remove_keys=confusion_matrix_info.remove_keys,
    )


def get_conf_matrix_from_one_dict(
    base_dict: Dict[str, Any],
    norm_info_value: Union[int, Dict[str, Any]],
    norm_over_y: bool,
    remove_keys: Optional[List[str]],
) -> pd.DataFrame:
    if remove_keys is None:
        remove_keys = []

    keys = [k for k in base_dict.keys() if k not in remove_keys]  # type: List[str]
    sub_keys = list(set([kk for k in keys for kk in base_dict[k].keys() if kk not in remove_keys]))

    if isinstance(norm_info_value, int):
        result = {
            k: {kk: 1.0 * base_dict[k][kk] / norm_info_value if kk in base_dict[k].keys() else 0.0 for kk in sub_keys}
            for k in keys
        }
    elif isinstance(norm_info_value, dict):
        assert all(isinstance(v, int) for v in norm_info_value.values())
        if norm_over_y:
            assert len(norm_info_value) >= len(sub_keys), (norm_info_value, sub_keys)
            assert all(k in norm_info_value for k in sub_keys)
            assert all(all(base_dict[k][kk] <= norm_info_value[kk] for kk in base_dict[k]) for k in keys)
            result = {
                k: {
                    kk: 1.0 * base_dict[k][kk] / norm_info_value[kk] if kk in base_dict[k].keys() else 0.0
                    for kk in sub_keys
                }
                for k in keys
            }
        else:
            r_keys = [k for k in keys if norm_info_value[k]]
            assert len(norm_info_value) >= len(r_keys), (norm_info_value, r_keys)
            assert all(k in norm_info_value for k in r_keys)
            assert all(all(base_dict[k][kk] <= norm_info_value[k] for kk in base_dict[k]) for k in r_keys)
            result = {
                k: {
                    kk: 1.0 * base_dict[k][kk] / norm_info_value[k] if kk in base_dict[k].keys() else 0.0
                    for kk in sub_keys
                }
                for k in r_keys
            }
    else:
        raise ValueError("Unexpected type retrieved for normalization.")

    return pd.DataFrame(result)


def get_norm_key_for_inverted(key: str) -> int:
    return int(key.rsplit("_", 1)[1])
