"""
Defines methods to extract information about form factor and pID correction weights
which have been added post n-tuple production.

Felix Metzner 2020
"""

import os
import json
import numpy as np
import pandas as pd

from typing import Union, Tuple, List, Dict, Generator, Any

from rdstar.utilities import PathType
from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.rdstar_samples import unused_sample_keys
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.corrections.particle_id_corrections import ParticleIDCorrection


__all__ = [
    "WeightInfoGatherer",
]


class WeightInfoGatherer:
    def __init__(
        self,
        weight_key: str,
        selection_name: str,
        cs_tag: str,
        input_dictionary: Dict[str, PathType],
        output_path: PathType,
    ) -> None:
        assert weight_key in WeightManager.weight_info_keys, (weight_key, WeightManager.weight_info_keys)
        self.weight_key = weight_key

        self._selection_name = selection_name  # type: str
        self._cs_tag = cs_tag  # type: str

        self._input_dictionary = input_dictionary  # type: Dict[str, PathType]
        self.output_path = output_path  # type:PathType

        self.main_sel_tag = SelectionSetups.get_main_selection_tag(
            selection_name=self._selection_name,
            cs_tag=self._cs_tag,
        )  # type: str

    def load_weight_info(self) -> pd.DataFrame:
        dfs = [
            pd.read_hdf(path)
            for key, path in self._input_dictionary.items()
            if key.endswith(self.weight_key) and not any(key.startswith(usk) for usk in unused_sample_keys)
        ]

        df = pd.concat(dfs, axis=0, ignore_index=True, sort=False)
        if "index" in df.columns:
            del df["index"]

        del dfs
        if self.weight_key == WeightManager.pid_weight_info_key:
            if WeightManager.evt_id_col not in df.columns:
                utilities.recreate_event_id(df=df, remove_original_columns=True)
        return df

    def get_selection_conditions(self) -> Tuple[np.ndarray, np.ndarray]:
        df_paths = [
            path
            for key, path in self._input_dictionary.items()
            if key.endswith(self.main_sel_tag)
            and not key.startswith("Data")
            and not any(key.startswith(usk) for usk in unused_sample_keys)
        ]  # type: List[PathType]

        dfs = [pd.read_hdf(path) for path in df_paths]  # type: List[Union[pd.DataFrame, pd.Series, Any]]

        assert all(isinstance(d, pd.DataFrame) for d in dfs), [type(d) for d in dfs]

        df = pd.concat(
            [sdf.loc[:, [WeightManager.evt_id_col, WeightManager.c_id_col]] for sdf in dfs if len(sdf.index) > 0],
            axis=0,
            ignore_index=True,
            sort=False,
        )

        df.reset_index(drop=True, inplace=True)
        assert len(df.index) == len(np.unique(df.index)), (len(df.index), len(np.unique(df.index)))

        if "index" in df.columns:
            del df["index"]

        unique_candidate_id = WeightManager.add_unique_id(df=df, check_if_unique=True)

        return df[WeightManager.evt_id_col].unique(), np.unique(unique_candidate_id)

    def gather_weight_info(self, dry_run: bool = False) -> Generator[Tuple[str, PathType], None, None]:
        selection_conditions = None
        if not dry_run:
            selection_conditions = self.get_selection_conditions()

        if not dry_run:
            df = self.load_weight_info()
        else:
            df = pd.DataFrame()

        if self.weight_key == WeightManager.pid_weight_info_key:
            if not dry_run:
                self.dump_hadron_pid_flag_uncertainty_dictionary(df=df)
            pid_flag_dict_key = WeightManager.get_pid_flag_dict_output_key(selection_name=self._selection_name)
            yield pid_flag_dict_key, self._get_pid_weight_info_file_path()

        key, path = self.dump_weight_info_data_frame(df=df, dry_run=dry_run)
        yield key, path

        if selection_conditions is not None or dry_run:
            event_cond, candidate_cond = None, None
            if not dry_run:
                assert selection_conditions is not None
                event_cond = df[WeightManager.evt_id_col].isin(selection_conditions[0])
                WeightManager.add_unique_id(df=df)
                candidate_cond = df.loc[event_cond, WeightManager.id_col].isin(selection_conditions[1])

            key, path = self.dump_weight_info_data_frame(
                df=df.loc[event_cond].loc[candidate_cond] if not dry_run else df,
                dry_run=dry_run,
                with_selection=True,
            )
            yield key, path

        del df

    def dump_weight_info_data_frame(
        self,
        df: pd.DataFrame,
        dry_run: bool,
        with_selection: bool = False,
    ) -> Tuple[str, PathType]:
        key = WeightManager.get_weight_output_key(weight_key=self.weight_key, with_selection=with_selection)
        path = self._get_weight_info_data_frame_path(with_selection=with_selection)
        tmp_path = self._get_weight_info_data_frame_path(tmp=True, with_selection=with_selection)
        if not dry_run:
            df.to_hdf(
                path_or_buf=tmp_path,
                key=key,
                format="table" if self.weight_key is WeightManager.pid_weight_info_key else "fixed",
            )
            os.rename(src=tmp_path, dst=path)
        return key, path

    def dump_hadron_pid_flag_uncertainty_dictionary(self, df: pd.DataFrame) -> None:
        flag_dict = ParticleIDCorrection.get_hadron_pid_flag_uncert_dict(df)
        with open(self._get_pid_weight_info_file_path(tmp=True), "w") as out_file:
            json.dump(flag_dict, out_file, indent=2)

        os.rename(src=self._get_pid_weight_info_file_path(tmp=True), dst=self._get_pid_weight_info_file_path(tmp=False))

    def _get_pid_weight_info_file_path(self, tmp: bool = False) -> PathType:
        return WeightManager.get_pid_weight_info_dict_file_path(dir_path=self.output_path, tmp=tmp)

    def _get_weight_info_data_frame_path(
        self,
        tmp: bool = False,
        with_selection: bool = False,
    ) -> PathType:
        return WeightManager.get_weight_info_data_frame_path(
            dir_path=self.output_path,
            weight_key=self.weight_key,
            tmp=tmp,
            with_selection=with_selection,
        )
