"""
Manager class providing all the necessary information about the
privately added weights used by the analysis

Felix Metzner 2020
"""

import os
import numpy as np
import pandas as pd

from collections.abc import Set as SetBaseClass
from typing import Optional, Tuple, List, Dict, NamedTuple, Iterator, Collection

from rdstar.utilities import PathType
from rdstar.offline_analysis.mc_scaling_weights import MCSampleScaler
from rdstar.latex_labels.latex_particle_names import LatexParticleNames
from rdstar.mc_reweighting.BFReweighting.bf_reweighting import BFReweighter
from rdstar.mc_reweighting.BFReweighting.d_bf_reweighting import DBFReweighter
from rdstar.mc_reweighting.BFReweighting.b_bf_reweighting import BMesonBFReweighter
from rdstar.offline_analysis.data_handling.data_sample_tag_names import DataSampleTags
from rdstar.offline_analysis.corrections.k_short_corrections import KShortEffCorrector
from rdstar.offline_analysis.corrections.tracking_corrections import TrackingEffCorrector
from rdstar.offline_analysis.corrections.particle_id_corrections import ParticleIDCorrection
from rdstar.offline_analysis.corrections.fei_tag_correction import TagCorrection, InclusiveTagCorrection
from rdstar.offline_analysis.corrections.efficiency_descrepancy_corrector import ReconstructionModeEfficiencyCorrector
from rdstar.offline_analysis.rdstar_sample_weights import scale_weight_column_name, off_to_on_res_scale_weight_column_name

from rdstar.offline_analysis.tasks import utilities


__all__ = [
    "WeightTypeInfo",
    "WeightTypeCollection",
    "WeightManager",
]


_combined_ff_weight_col_name = "__weight_ff_nom__"  # type: str


class WeightTypeInfo(NamedTuple):
    name: str
    column_name: str
    latex_str: str
    is_additional: bool
    is_for_efficiency: bool
    validate: bool
    is_bf_correction_weight: bool = False

    @property
    def is_additional_for_bf(self) -> bool:
        return self.is_additional and self.is_bf_correction_weight


class _WeightTypeCollectionClass(SetBaseClass):
    def __init__(self, input_iterable: Collection) -> None:
        assert all(isinstance(wti, WeightTypeInfo) for wti in input_iterable)

        names = [w.name for w in input_iterable]  # type: List[str]
        assert len(names) == len(set(names)), (len(names), len(set(names)))
        cols = [w.column_name for w in input_iterable]  # type: List[str]
        assert len(cols) == len(set(cols)), (len(cols), len(set(cols)))

        self.__elements = tuple(input_iterable)  # type: Tuple[WeightTypeInfo, ...]

        assert set(self.additional_weight_columns) | set(self.combined_weight_columns) == set(cols)

    def __iter__(self) -> Iterator[WeightTypeInfo]:
        return iter(self.__elements)

    def __contains__(self, value: object) -> bool:
        if not isinstance(value, str):
            raise TypeError(f"Expected argument of type str but got {type(value).__name__}")
        return value in [w.name for w in self.__elements]

    def __len__(self) -> int:
        return len(self.__elements)

    def __getitem__(self, name: str) -> WeightTypeInfo:
        if name not in self:
            raise KeyError(f"Weight type {name} is not available!")
        return [w for w in self.__elements if w.name == name][0]

    @property
    def combined_weight_columns(self) -> Tuple[str, ...]:
        return tuple([w.column_name for w in self.__elements if not w.is_additional])

    @property
    def additional_weight_columns(self) -> Tuple[str, ...]:
        return tuple([w.column_name for w in self.__elements if w.is_additional])

    @property
    def bf_correction_weight_columns(self) -> Tuple[str, ...]:
        return tuple([w.column_name for w in self.__elements if w.is_additional_for_bf])

    @property
    def efficiency_weight_columns(self) -> Tuple[str, ...]:
        return tuple([w.column_name for w in self.__elements if w.is_for_efficiency])

    @property
    def weights_for_validation(self) -> Tuple[WeightTypeInfo, ...]:
        return tuple([w for w in self.__elements if w.validate])


WeightTypeCollection = _WeightTypeCollectionClass(
    [
        WeightTypeInfo(
            name="ff_weight",
            column_name=_combined_ff_weight_col_name,
            latex_str="Form Factor Weights",
            is_additional=False,
            is_for_efficiency=True,
            validate=True,
        ),
        WeightTypeInfo(
            name="pid_weight",
            column_name=ParticleIDCorrection.get_final_pid_weight_col_name(),
            latex_str="Particle ID Weights",
            is_additional=False,
            is_for_efficiency=True,
            validate=True,
        ),
        WeightTypeInfo(
            name="slow_pi_weight",
            column_name=TrackingEffCorrector.slow_pi_corr_col_name,
            latex_str="Slow Pion Correction Weights",
            is_additional=False,
            is_for_efficiency=True,
            validate=True,
        ),
        WeightTypeInfo(
            name="k_short_weight",
            column_name=KShortEffCorrector.k_short_corr_col_name,
            latex_str=LatexParticleNames.k_short_str + r" Weights",
            is_additional=False,
            is_for_efficiency=True,
            validate=True,
        ),
        WeightTypeInfo(
            name="scale_weight",
            column_name=MCSampleScaler.mc_scaling_weight_column_name,
            latex_str="MC Scaling Weights",
            is_additional=False,
            is_for_efficiency=False,
            validate=False,
        ),
        WeightTypeInfo(
            name="b_bf_weight",
            column_name=BMesonBFReweighter.b_bf_weight_col_name,
            latex_str=LatexParticleNames.b_str + " Meson BF Weight",
            is_additional=True,
            is_for_efficiency=False,
            validate=True,
            is_bf_correction_weight=True,
        ),
        WeightTypeInfo(
            name="d_bf_weight",
            column_name=DBFReweighter.d_bf_weight_col,
            latex_str=LatexParticleNames.d_str + " Meson BF Weight",
            is_additional=True,
            is_for_efficiency=True,
            validate=True,
            is_bf_correction_weight=True,
        ),
        WeightTypeInfo(
            name="tag_weight",
            column_name=TagCorrection.tagging_correction_name,
            latex_str="Tagging Calibration Weights",
            is_additional=True,
            is_for_efficiency=False,
            validate=True,
        ),
        # WeightTypeInfo(  # TODO: use instead of simple tag_weight
        #     name="tag_inclusive_weight",
        #     column_name=InclusiveTagCorrection.tagging_correction_name,
        #     latex_str="Tagging Calibration Weights",
        #     is_additional=True,
        #     is_for_efficiency=False,
        #     validate=True,
        # ),
        WeightTypeInfo(
            name="off_res_weight",
            column_name=off_to_on_res_scale_weight_column_name,
            latex_str="Off-Resonance Scaling Weights",
            is_additional=True,  # TODO: Consider not as additional but as combined_weight_column or completely separate?
            is_for_efficiency=False,
            validate=False,
        ),
    ]
)  # type: _WeightTypeCollectionClass


class WeightManager:
    weight_analysis_base_dir_name = "Weight_Analysis"  # type: str

    base_fei_calib_col = TagCorrection.tagging_correction_name  # type: str
    used_fei_calib_col = InclusiveTagCorrection.tagging_correction_name  # type: str

    finalized_weights_appendix = DataSampleTags.weight_finalized_sample_tag  # type: str

    old_final_weight_col = "__old_weight_full__"  # type: str
    old_final_weight_with_bf_corrections_col = "__old_weight_final_wBF_corr__"  # type: str
    old_final_weight_with_additions_col = "__old_weight_final_wAdditions_corr__"  # type: str

    final_weight_col = "__new_weight_full__"  # type: str
    final_weight_with_bf_corrections_col = "__new_weight_final_wBF_corr__"  # type: str
    final_weight_with_additions_col = "__new_weight_final_wAdditions_corr__"  # type: str

    combined_ff_weight_col_name = _combined_ff_weight_col_name  # type: str

    weight_collection = WeightTypeCollection  # type: _WeightTypeCollectionClass

    combined_weight_column_names = weight_collection.combined_weight_columns  # type: Tuple[str, ...]
    additional_weight_column_names = weight_collection.additional_weight_columns  # type: Tuple[str, ...]
    bf_correction_weight_column_names = weight_collection.bf_correction_weight_columns  # type: Tuple[str, ...]
    efficiency_weight_column_names = weight_collection.efficiency_weight_columns  # type: Tuple[str, ...]

    pid_weight_info_key = "pid_weight_info"  # type: str
    ff_d_weight_info_key = "ff_d_weight_info"  # type: str
    ff_ds_weight_info_key = "ff_ds_weight_info"  # type: str
    ff_dss_weight_info_key = "ff_dss_weight_info"  # type: str

    weight_info_keys = [
        pid_weight_info_key,
        ff_d_weight_info_key,
        ff_ds_weight_info_key,
        ff_dss_weight_info_key,
    ]  # type: List[str]

    pid_dict_output_key = "pid_hadron_for_flag_dict"  # type: str
    combined_ff_weights_prefix = "combined"  # type: str

    evt_id_col = utilities.UniqueEventIDColumnInfos.get_unique_event_id_col_name()  # type: str
    c_id_col = "unique_candidate_id"  # type: str
    id_col = "combined_candidate_id"  # type: str

    with_selection_suffix = "_with_selection"  # type: str

    @staticmethod
    def add_final_weight(df: pd.DataFrame) -> None:
        assert WeightManager.final_weight_col not in df.columns, WeightManager.final_weight_col
        weight_col_list = list(WeightManager.combined_weight_column_names)

        if df.empty:
            df[WeightManager.final_weight_col] = 1.0
        else:
            assert all(c in df.columns for c in weight_col_list), [c for c in weight_col_list if c not in df.columns]
            df.loc[:, WeightManager.final_weight_col] = df[weight_col_list].product(axis=1)
            assert pd.notnull(df[WeightManager.final_weight_col]).all()

    @staticmethod
    def add_bf_correction_weights(df: pd.DataFrame) -> None:
        assert WeightManager.final_weight_col in df.columns
        assert WeightManager.final_weight_with_bf_corrections_col not in df.columns
        assert all(c in df.columns for c in WeightManager.bf_correction_weight_column_names), [
            c for c in WeightManager.bf_correction_weight_column_names if c not in df.columns
        ]

        if df.empty:
            df[WeightManager.final_weight_with_bf_corrections_col] = 1.0
        else:
            col_list = list(WeightManager.bf_correction_weight_column_names)
            col_list.append(WeightManager.final_weight_col)
            df.loc[:, WeightManager.final_weight_with_bf_corrections_col] = df[col_list].product(axis=1)

    @staticmethod
    def add_additional_weights(df: pd.DataFrame) -> None:
        assert WeightManager.final_weight_col in df.columns
        assert WeightManager.final_weight_with_additions_col not in df.columns
        assert all(c in df.columns for c in WeightManager.additional_weight_column_names), [
            c for c in WeightManager.additional_weight_column_names if c not in df.columns
        ]

        if df.empty:
            df[WeightManager.final_weight_with_additions_col] = 1.0
        else:
            col_list = list(WeightManager.additional_weight_column_names)
            col_list.append(WeightManager.final_weight_col)
            df.loc[:, WeightManager.final_weight_with_additions_col] = df[col_list].product(axis=1)

    @staticmethod
    def add_old_final_weight(df: pd.DataFrame) -> None:
        WeightManager.add_alternative_final_weight(
            df=df,
            new_weight_col_name=WeightManager.old_final_weight_col,
            alternative_weight_column_mapping={MCSampleScaler.mc_scaling_weight_column_name: scale_weight_column_name},
            additional_columns=(),
        )

    @staticmethod
    def add_old_bf_correction_weights(df: pd.DataFrame) -> None:
        old_w_map = {BMesonBFReweighter.b_bf_weight_col_name: BFReweighter.weight_total_col_name}
        WeightManager.add_alternative_final_weight(
            df=df,
            new_weight_col_name=WeightManager.old_final_weight_with_bf_corrections_col,
            alternative_weight_column_mapping={MCSampleScaler.mc_scaling_weight_column_name: scale_weight_column_name},
            additional_columns=tuple([old_w_map.get(c, c) for c in WeightManager.bf_correction_weight_column_names]),
        )

    @staticmethod
    def add_old_additional_weights(df: pd.DataFrame) -> None:
        old_w_map = {BMesonBFReweighter.b_bf_weight_col_name: BFReweighter.weight_total_col_name}
        WeightManager.add_alternative_final_weight(
            df=df,
            new_weight_col_name=WeightManager.old_final_weight_with_additions_col,
            alternative_weight_column_mapping={MCSampleScaler.mc_scaling_weight_column_name: scale_weight_column_name},
            additional_columns=tuple([old_w_map.get(c, c) for c in WeightManager.additional_weight_column_names]),
        )

    @staticmethod
    def add_alternative_final_weight(
        df: pd.DataFrame,
        new_weight_col_name: str,
        alternative_weight_column_mapping: Dict[str, str],
        additional_columns: Tuple[str, ...],
    ) -> None:
        assert new_weight_col_name not in df.columns, new_weight_col_name

        assert all(c in WeightManager.combined_weight_column_names for c in alternative_weight_column_mapping.keys()), [
            c for c in alternative_weight_column_mapping.keys() if c not in WeightManager.combined_weight_column_names
        ]
        assert all(c in df.columns for c in additional_columns), [c for c in additional_columns if c not in df.columns]

        assert all(c in df.columns for c in alternative_weight_column_mapping.keys()), [
            c for c in alternative_weight_column_mapping.keys() if c not in df.columns
        ]
        assert all(c in df.columns for c in alternative_weight_column_mapping.values()), [
            c for c in alternative_weight_column_mapping.values() if c not in df.columns
        ]

        weight_col_list = [
            alternative_weight_column_mapping.get(c, c) for c in WeightManager.combined_weight_column_names
        ]  # type: List[str]

        assert all(c not in WeightManager.combined_weight_column_names for c in additional_columns), [
            c for c in additional_columns if c in WeightManager.combined_weight_column_names
        ]
        assert all(c not in weight_col_list for c in additional_columns), [
            c for c in additional_columns if c in weight_col_list
        ]

        if df.empty:
            df[new_weight_col_name] = 1.0
            return

        weight_col_list.extend(list(additional_columns))

        df.loc[:, new_weight_col_name] = df[weight_col_list].product(axis=1)
        assert pd.notnull(df[new_weight_col_name]).all()

    @staticmethod
    def add_unique_id(
        df: pd.DataFrame,
        check_if_unique: bool = False,
        mask: Optional[pd.Series] = None,
    ) -> np.ndarray:
        id_values = utilities.add_unique_candidate_id(
            df=df,
            id_column_name=WeightManager.id_col,
            event_id_name=WeightManager.evt_id_col,
            candidate_id_name=WeightManager.c_id_col,
            check_if_unique=check_if_unique,
            mask=mask,
        )
        return id_values

    @staticmethod
    def get_weight_output_key(weight_key: str, with_selection: bool) -> str:
        assert weight_key in WeightManager.weight_info_keys, (weight_key, WeightManager.weight_info_keys)
        selection_suffix = WeightManager.with_selection_suffix if with_selection else ""  # type: str
        if weight_key.startswith("ff"):
            return f"{WeightManager.combined_ff_weights_prefix}_{weight_key}{selection_suffix}"
        else:
            return f"{WeightManager.pid_dict_output_key}{selection_suffix}"

    @staticmethod
    def get_flag_info_dict_path_from_task_input_dict(
        input_dict: Dict[str, PathType],
        selection_name: str,
    ) -> PathType:
        return input_dict[WeightManager.get_pid_flag_dict_output_key(selection_name=selection_name)]

    @staticmethod
    def get_pid_weight_info_dict_file_path(
        dir_path: PathType,
        tmp: bool = False,
    ) -> PathType:
        if tmp:
            return os.path.join(dir_path, f"temp_{WeightManager.pid_dict_output_key}.json")
        else:
            return os.path.join(dir_path, f"{WeightManager.pid_dict_output_key}.json")

    @staticmethod
    def get_pid_flag_dict_output_key(selection_name: str) -> str:
        return f"{WeightManager.pid_dict_output_key}_{selection_name}"

    @staticmethod
    def get_weight_info_data_frame_path(
        dir_path: PathType,
        weight_key: str,
        tmp: bool = False,
        with_selection: bool = False,
    ) -> PathType:
        file_name = WeightManager.get_weight_output_key(weight_key=weight_key, with_selection=with_selection)
        if tmp:
            return os.path.join(dir_path, f"temp_{file_name}.h5")
        else:
            return os.path.join(dir_path, f"{file_name}.h5")

    @staticmethod
    def remove_off_to_on_resonance_scaling_from_weight_with_adds(df: pd.DataFrame) -> None:
        off_res_scale_weight_info = WeightManager.weight_collection["off_res_weight"]  # type: WeightTypeInfo
        assert off_res_scale_weight_info.column_name in df.columns, off_res_scale_weight_info.column_name
        assert WeightManager.final_weight_with_additions_col in df.columns, WeightManager.final_weight_with_additions_col

        old_weights_with_additions = df[WeightManager.final_weight_with_additions_col].values  # type: np.ndarray
        off_to_on_resonance_scaling_weights = df[off_res_scale_weight_info.column_name].values  # type: np.ndarray
        new_weights_with_additions = old_weights_with_additions / off_to_on_resonance_scaling_weights  # type: np.ndarray
        df[WeightManager.final_weight_with_additions_col] = new_weights_with_additions

    @staticmethod
    def is_active_weight_col_name(col_name: str) -> bool:
        return any(col_name == wi.column_name for wi in WeightManager.weight_collection)

    @staticmethod
    def check_is_active_weight_col_name(col_name: str) -> None:
        assert WeightManager.is_active_weight_col_name(col_name=col_name), [
            col_name,
            [wi.column_name for wi in WeightManager.weight_collection],
        ]

    @staticmethod
    def add_quick_fei_recalibration(
        df: pd.DataFrame,
        weight_column_name: str,
        fei_calib_weight_col_name: str,
    ) -> None:
        assert weight_column_name in df.columns, weight_column_name

        _fei_cal_cols = [
            InclusiveTagCorrection.tagging_correction_name,
            InclusiveTagCorrection.tagging_correction_allB_name,
        ]  # type: List[str]
        assert fei_calib_weight_col_name in _fei_cal_cols, (fei_calib_weight_col_name, _fei_cal_cols)

        WeightManager.check_is_active_weight_col_name(col_name=WeightManager.base_fei_calib_col)
        assert WeightManager.base_fei_calib_col in df.columns, WeightManager.base_fei_calib_col

        assert all(c in df.columns for c in InclusiveTagCorrection.created_columns), [
            c for c in InclusiveTagCorrection.created_columns if c not in df.columns
        ]

        weight_col_reset_info = {
            WeightManager.final_weight_col: False,
            WeightManager.final_weight_with_bf_corrections_col: False,
            WeightManager.final_weight_with_additions_col: True,
        }  # type: Dict[str, bool]

        if weight_column_name not in weight_col_reset_info:
            raise RuntimeError(f"Don't know what to do for weight column name '{weight_column_name}'!")

        df.rename(mapper={c: f"__old__{c}" for c in InclusiveTagCorrection.created_columns}, axis=1, inplace=True)

        # Inclusive Tagging Correction:
        fei_inclusive_tag_corrector = InclusiveTagCorrection(
            tag_calibration_name=InclusiveTagCorrection.default_tag_calibration_name,
        )
        fei_inclusive_tag_corrector.apply_tagging_correction(sample=df)

        assert all(c in df.columns for c in InclusiveTagCorrection.created_columns), [
            c for c in InclusiveTagCorrection.created_columns if c not in df.columns
        ]

        for original_weight_col_name, reset_weight in weight_col_reset_info.items():
            df[f"{original_weight_col_name}without_new_fei_corr_factors__"] = df[original_weight_col_name].values
            old_fei_cal_values = df[WeightManager.base_fei_calib_col].values  # type: np.ndarray
            if reset_weight:
                new_weight_values = (
                    df[original_weight_col_name].values / old_fei_cal_values * df[fei_calib_weight_col_name].values
                )  # type: np.ndarray
                df[original_weight_col_name] = new_weight_values
            else:
                df[original_weight_col_name] *= df[fei_calib_weight_col_name].values

    @staticmethod
    def switch_fei_calibration(
        df: pd.DataFrame,
        new_fei_calib_col: str,
        main_weight_col_name: str,
    ) -> None:

        fei_base_col_name = WeightManager.base_fei_calib_col  # type: str
        WeightManager.check_is_active_weight_col_name(col_name=fei_base_col_name)
        assert fei_base_col_name in df.columns, fei_base_col_name

        if new_fei_calib_col == fei_base_col_name:
            return

        weight_reset_cols = (
            WeightManager.final_weight_col,
            WeightManager.final_weight_with_bf_corrections_col,
            WeightManager.final_weight_with_additions_col,
        )  # type: Tuple[str, ...]

        if main_weight_col_name not in weight_reset_cols:
            raise RuntimeError(f"Don't know what to do for weight column name '{main_weight_col_name}'!")

        for weight_col in weight_reset_cols:
            WeightManager.switch_fei_calibration_for(
                df=df,
                weight_column_name=weight_col,
                new_fei_calib_col=new_fei_calib_col,
            )

    @staticmethod
    def switch_fei_calibration_for(
        df: pd.DataFrame,
        weight_column_name: str,
        new_fei_calib_col: str,
    ) -> None:
        assert weight_column_name in df.columns, weight_column_name

        assert new_fei_calib_col in df.columns, new_fei_calib_col
        assert (
            new_fei_calib_col in TagCorrection.created_columns
            or new_fei_calib_col in InclusiveTagCorrection.created_columns
        ), (new_fei_calib_col, TagCorrection.created_columns, InclusiveTagCorrection.created_columns)

        fei_base_col_name = WeightManager.base_fei_calib_col  # type: str
        WeightManager.check_is_active_weight_col_name(col_name=fei_base_col_name)
        assert fei_base_col_name in df.columns, fei_base_col_name

        assert new_fei_calib_col != fei_base_col_name, (new_fei_calib_col, fei_base_col_name)

        weight_col_reset_info = {
            WeightManager.final_weight_col: False,
            WeightManager.final_weight_with_bf_corrections_col: False,
            WeightManager.final_weight_with_additions_col: True,
        }  # type: Dict[str, bool]

        if weight_column_name not in weight_col_reset_info:
            raise RuntimeError(f"Don't know what to do for weight column name '{weight_column_name}'!")

        df[f"{weight_column_name}without_new_fei_corr_factors__"] = df[weight_column_name].values

        if weight_col_reset_info[weight_column_name]:
            old_fei_cal_values = df[fei_base_col_name].values  # type: np.ndarray
            df[weight_column_name] = df[weight_column_name].values / old_fei_cal_values * df[new_fei_calib_col].values
        else:
            df[weight_column_name] *= df[new_fei_calib_col].values

    @staticmethod
    def add_reco_mode_dependent_efficiency_correction(
        df: pd.DataFrame,
        weight_column_name: str,
        reco_efficiency_correction_container_path: PathType,
    ) -> None:
        reco_eff_corrector = ReconstructionModeEfficiencyCorrector.initialize_with_container_path(
            reco_efficiency_correction_info_container_path=reco_efficiency_correction_container_path,
            relative_rescaling_only=True,
        )

        weight_col_names = (
            WeightManager.final_weight_col,
            WeightManager.final_weight_with_bf_corrections_col,
            WeightManager.final_weight_with_additions_col,
        )  # type: Tuple[str, ...]
        if weight_column_name not in weight_col_names:
            raise RuntimeError(f"Don't know what to do for weight column name '{weight_column_name}'!")

        reco_eff_corrector.add_weight_to_existing_weight_columns(
            df=df,
            weight_columns=weight_col_names,
        )
