"""
This module defines the final state particles considered for the analysis.

Felix Metzner 2018
"""

from typing import List, Union, Optional

from analysistools.processing.final_state_particles_utils import (
    FinalStateParticle,
    FSRCorrectionSettings,
    FinalStateParticleCollection,
)

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.cut_information_objects.cut_info import CutInfo
from rdstar.latex_labels.latex_observable_names import LatexObsNames


__all__ = [
    "generate_fsp",
    "fsp_collection",
    "fsp_cut_sets",
    "electron_fsr_correction_settings",
    "electron_cuts",
    "muon_cuts",
    "tau_cuts",
    "charged_pion_cuts",
    "neutral_pion_cuts",
    "photon_cuts",
    "charged_kaon_cuts",
    "neutral_kaon_cuts",
    "neutrino_cuts",
    "neutrino_cuts",
    "neutrino_cuts",
]


def generate_fsp(
    label: str,
    particle_list: str,
    origin: Union[str, List[str]],
    cut_infos: List[CutInfo],
    electron_fsr_correction: Optional[FSRCorrectionSettings] = None,
) -> FinalStateParticle:
    return FinalStateParticle(
        label=label,
        particlelist=particle_list,
        origin=origin,
        cuts=[cut_info.cut_str for cut_info in cut_infos],
        electron_fsr_correction=electron_fsr_correction,
    )


def get_cut_infos(fsp: FinalStateParticle) -> List[CutInfo]:
    assert fsp.label in fsp_cut_sets.keys()
    return fsp_cut_sets[fsp.label]


# define some often used cuts
dr_cut = CutInfo(
    cut_str="dr < 2.",
    cut_label=LatexObsNames.impact_delta_r[:-1] + r" < 2.0\," + LatexUnits.cm[1:],
    cut_name="dr_cut",
    cut_var="dr",
    cut_var_latex_label=LatexObsNames.impact_delta_r,
    cut_type="<",
    cut_limits=2.0,
    cut_online=True,
    latex_unit=LatexUnits.cm,
)
dz_cut = CutInfo(
    cut_str="abs(dz) < 4.",
    cut_label=LatexObsNames.impact_abs_delta_z[:-1] + r" < 4.0\," + LatexUnits.cm[1:],
    cut_name="dz_cut",
    cut_var="abs(dz)",
    cut_var_latex_label=LatexObsNames.impact_abs_delta_z,
    cut_type="<",
    cut_limits=4.0,
    cut_abs=True,
    cut_online=True,
    latex_unit=LatexUnits.cm,
)
pt_cut = CutInfo(
    cut_str="useCMSFrame(pt) > 0.1",
    cut_label=LatexObsNames.momentum_pt[:-1] + r" > 100\," + LatexUnits.mev[1:],
    cut_name="pt_cut",
    cut_var="useCMSFrame(pt)",
    cut_var_latex_label=LatexObsNames.momentum_pt,
    cut_type=">",
    cut_limits=0.1,
    cut_online=True,
    latex_unit=LatexUnits.gev,
)
fsp_sig_mbf_cut = CutInfo(
    cut_str="abs(SigMBF) < 3.",
    cut_label=LatexObsNames.mass_abs_sig_mbf[:-1] + r" < 3.0$",
    cut_name="sig_mbf_cut",
    cut_var="abs(SigMBF)",
    cut_var_latex_label=LatexObsNames.mass_abs_sig_mbf,
    cut_type="<",
    cut_limits=3.0,
    cut_abs=True,
    cut_online=True,
    latex_unit=LatexUnits.no_unit,
)

# Definition of CutInfo Lists
electron_cuts = [
    dr_cut,
    dz_cut,
    pt_cut,
    CutInfo(
        cut_str="eIDBelle > 0.1",
        cut_label=LatexObsNames.pid_electron[:-1] + r" > 0.1$",
        cut_name="e_id_cut",
        cut_var="eIDBelle",
        cut_var_latex_label=LatexObsNames.pid_electron,
        cut_type=">",
        cut_limits=0.1,
        cut_online=True,
        latex_unit=LatexUnits.no_unit,
    ),
]

muon_cuts = [
    dr_cut,
    dz_cut,
    pt_cut,
    CutInfo(
        cut_str="muIDBelle > 0.1",
        cut_label=LatexObsNames.pid_muon[:-1] + r" > 0.1$",
        cut_name="mu_id_cut",
        cut_var="muIDBelle",
        cut_var_latex_label=LatexObsNames.pid_muon,
        cut_type=">",
        cut_limits=0.1,
        cut_online=True,
        latex_unit=LatexUnits.no_unit,
    ),
]

tau_cuts = [
    dr_cut,
    dz_cut,
]

charged_pion_cuts = [
    dr_cut,
    dz_cut,
    pt_cut,
    CutInfo(
        cut_str="atcPIDBelle(2,3) > 0.1",
        cut_label=LatexObsNames.pid_atc23[:-1] + r" > 0.1$",
        cut_name="charged_pion_pid_cut",
        cut_var="atcPIDBelle(2,3)",
        cut_var_latex_label=LatexObsNames.pid_atc23,
        cut_type=">",
        cut_limits=0.1,
        cut_online=True,
        latex_unit=LatexUnits.no_unit,
    ),
]

neutral_pion_cuts = [
    CutInfo(
        cut_str="0.104 < InvM < 0.165",
        cut_label=LatexObsNames.mass_inv_m[:-1] + r"\;\in\;(104, 165)\," + LatexUnits.mev[1:],
        cut_name="pion_invM_cut",
        cut_var="InvM",
        cut_var_latex_label=LatexObsNames.mass_inv_m,
        cut_type=("<", "<"),
        cut_limits=(0.104, 0.165),
        cut_online=True,
        latex_unit=LatexUnits.gev,
    ),
    fsp_sig_mbf_cut,
    CutInfo(
        cut_str="E > 0.1",
        cut_label=LatexObsNames.energy_e[:-1] + r" > 100\," + LatexUnits.mev[1:],
        cut_name="E_cut",
        cut_var="E",
        cut_var_latex_label=LatexObsNames.energy_e,
        cut_type=">",
        cut_limits=0.1,
        cut_online=True,
        latex_unit=LatexUnits.gev,
    ),
    CutInfo(
        cut_str="daughter(0, goodBelleGamma) == 1",
        cut_label=LatexObsNames.pi0_good_gamma_0[:-1] + r" == 1$",
        cut_name="gamma_0_cut",
        cut_var="daughter(0, goodBelleGamma)",
        cut_var_latex_label=LatexObsNames.pi0_good_gamma_0,
        cut_type="==",
        cut_limits=1,
        cut_online=True,
        latex_unit=LatexUnits.no_unit,
    ),
    CutInfo(
        cut_str="daughter(1, goodBelleGamma) == 1",
        cut_label=LatexObsNames.pi0_good_gamma_1[:-1] + r" == 1$",
        cut_name="gamma_1_cut",
        cut_var="daughter(1, goodBelleGamma)",
        cut_var_latex_label=LatexObsNames.pi0_good_gamma_1,
        cut_type="==",
        cut_limits=1,
        cut_online=True,
        latex_unit=LatexUnits.no_unit,
    ),
]

photon_cuts = [
    CutInfo(
        cut_str="goodBelleGamma == 1",
        cut_label=LatexObsNames.cluster_good_belle_gamma[:-1] + r" == 1$",
        cut_name="gamma_cut",
        cut_var="goodBelleGamma",
        cut_var_latex_label=LatexObsNames.cluster_good_belle_gamma,
        cut_type="==",
        cut_limits=1,
        cut_online=True,
        latex_unit=LatexUnits.no_unit,
    ),
]

charged_kaon_cuts = [
    dr_cut,
    dz_cut,
    pt_cut,
    CutInfo(
        cut_str="atcPIDBelle(3,2) > 0.1",
        cut_label=LatexObsNames.pid_atc32[:-1] + r" > 0.1$",
        cut_name="charged_kaon_pid_cut",
        cut_var="atcPIDBelle(3,2)",
        cut_var_latex_label=LatexObsNames.pid_atc32,
        cut_type=">",
        cut_limits=0.1,
        cut_online=True,
        latex_unit=LatexUnits.no_unit,
    ),
]

neutral_kaon_cuts = [
    CutInfo(
        cut_str="extraInfo(ksnbStandard) == 1",
        cut_label=LatexObsNames.pid_ksnb_standard[:-1] + r" == 1$",
        cut_name="ksnb_standard_cut",
        cut_var="extraInfo(ksnbStandard)",
        cut_var_latex_label=LatexObsNames.pid_ksnb_standard,
        cut_type="==",
        cut_limits=1,
        cut_online=True,
        latex_unit=LatexUnits.no_unit,
    ),
    CutInfo(
        cut_str="abs(dM) < 0.1",
        cut_label=LatexObsNames.mass_abs_delta_m[:-1] + r" < 100\," + LatexUnits.mev[1:],
        cut_name="neutral_kaon_dM_cut",
        cut_var="abs(dM)",
        cut_var_latex_label=LatexObsNames.mass_abs_delta_m,
        cut_type="<",
        cut_limits=0.1,
        cut_abs=True,
        cut_online=True,
        latex_unit=LatexUnits.gev,
    ),
    fsp_sig_mbf_cut,
]

neutrino_cuts = [
    dr_cut,
    dz_cut,
]

# FSRCorrectionSettings
electron_fsr_correction_settings = FSRCorrectionSettings(
    fsr_gamma_list_name="gamma:fsp",
    angle_threshold=2.0,
    energy_threshold=0.4,
    theta_threshold=1.0,
    momentum_factor=0.4,
    method="centeredCone",
    take_all_gammas=True,
    write_out=False,
)

# Particle Definition
electron = generate_fsp(
    label="e",
    particle_list="e-:fsp",
    origin="reco",
    cut_infos=electron_cuts,
    electron_fsr_correction=electron_fsr_correction_settings,
)

muon = generate_fsp(
    label="mu",
    particle_list="mu-:fsp",
    origin="reco",
    cut_infos=muon_cuts,
)

tau = generate_fsp(
    label="tau",
    particle_list="tau-:mc",
    origin="mc",
    cut_infos=tau_cuts,
)

charged_pion = generate_fsp(
    label="pi+",
    particle_list="pi+:fsp",
    origin="reco",
    cut_infos=charged_pion_cuts,
)

neutral_pion = generate_fsp(
    label="pi0",
    particle_list="pi0:fsp",
    origin=["pi0:mdst"],
    cut_infos=neutral_pion_cuts,
)

photon = generate_fsp(
    label="gamma",
    particle_list="gamma:fsp",
    origin=["gamma:mdst", "gamma:v0mdst"],
    cut_infos=photon_cuts,
)

charged_kaon = generate_fsp(
    label="K+",
    particle_list="K+:fsp",
    origin="reco",
    cut_infos=charged_kaon_cuts,
)

neutral_kaon = generate_fsp(
    label="K_S0",
    particle_list="K_S0:fsp",
    origin=["K_S0:mdst"],
    cut_infos=neutral_kaon_cuts,
)

e_neutrino = generate_fsp(
    label="nu_e",
    particle_list="nu_e:mc",
    origin="mc",
    cut_infos=neutrino_cuts,
)

mu_neutrino = generate_fsp(
    label="nu_mu",
    particle_list="nu_mu:mc",
    origin="mc",
    cut_infos=neutrino_cuts,
)

tau_neutrino = generate_fsp(
    label="nu_tau",
    particle_list="nu_tau:mc",
    origin="mc",
    cut_infos=neutrino_cuts,
)

# Add Particles to FSPCollection
_fsp_collection = FinalStateParticleCollection()
_fsp_collection.add_fsps(
    electron=electron,
    muon=muon,
    # tau=tau,
    charged_pion=charged_pion,
    neutral_pion=neutral_pion,
    photon=photon,
    charged_kaon=charged_kaon,
    neutral_kaon=neutral_kaon,
    e_neutrino=e_neutrino,
    mu_neutrino=mu_neutrino,
    tau_neutrino=tau_neutrino,
)
fsp_collection = _fsp_collection

fsp_cut_sets = {
    electron.label: electron_cuts,
    muon.label: muon_cuts,
    tau.label: tau_cuts,
    charged_pion.label: charged_pion_cuts,
    neutral_pion.label: neutral_pion_cuts,
    photon.label: photon_cuts,
    charged_kaon.label: charged_kaon_cuts,
    neutral_kaon.label: neutral_kaon_cuts,
    e_neutrino.label: neutrino_cuts,
    mu_neutrino.label: neutrino_cuts,
    tau_neutrino.label: neutrino_cuts,
}

if __name__ == "__main__":
    print(fsp_collection)
