"""
Cuts used for skims or selections are defined here.

For each skim a dictionary is defined, which contains:

description: A short description of the skim.
skim_name: String to be used for output file name.
target_lists: List of names of the particle lists on which the
              skim is performed.
cuts: List of tuples of (variable, cut) which will be applied
      in the respective order!
cands_in_lists: Cut criterion which will be used to reject every
                event based on the number of events in the provided
                target particle lists target_lists. E.g. with
                target_lists = ["B+:sig"] and cands_in_lists = "<1")
                every event which does not have at least one candidate
                in the B+:sig particle list will be rejected.

It is important to note, that the final Y(4S) particle is always
reconstructed in such a way, that the first daughter (index 0) is
the tag B meson, and the second (index 1) is the signal B meson.

Felix Metzner 2018
"""

from analysistools.processing.skim_cuts_utils import print_skim_cuts


__all__ = [
    "pre_skim_cuts",
    "post_signal_skim_cuts",
    "post_tag_skim_cuts",
    "post_reco_skim_cuts",
    "list_of_skims",
]


# Event level skim to reject crowded events as early as possible.
# An empty string as target list means that the cuts are applied
# on the whole event => target_lists = [""].
# Events/Candidates fulfilling the here defined criteria will be
# rejected!
pre_skim_cuts = {
    "description": "Event Level Skim Cuts.",
    "skim_name": "pre_event_skim",
    "target_lists": [""],
    "cuts": [
        ("nCleanedTracks(dr < 2 and abs(dz) < 4)", ">17"),
        ("countInList(gamma:fsp)", ">=18"),
    ],
    "cands_in_lists": None,
}

# Post signal-side B reconstruction skim.
# Events/Candidates fulfilling the here defined criteria will be
# rejected!
post_signal_skim_cuts = {
    "description": "Post Signal-Side B Reconstruction Skim Cuts.",
    "skim_name": "post_signal_skim",
    "target_lists": ["B+:sig", "B0:sig"],
    "cuts": [],
    "cands_in_lists": "<1",
}

# Post tag-side B reconstruction skim.
# Events/Candidates fulfilling the here defined criteria will be
# rejected!
post_tag_skim_cuts = {
    "description": "Post Tag-Side B Reconstruction Skim Cuts.",
    "skim_name": "post_tag_skim",
    "target_lists": ["B+:hadronic_tag", "B0:hadronic_tag"],
    "cuts": [],
    "cands_in_lists": "<1",
}

# Post Upsilon(4S) reconstruction skim.
# This skim is applied to stop processing every event for which no
# Upsilon(4S) candidate was obtained, and thus also no other information
# such as the generator level B meson info is required to be written
# to n-tuple files.
# Events/Candidates fulfilling the here defined criteria will be
# rejected!
post_reco_skim_cuts = {
    "description": "Post Upsilon(4S) Reconstruction Skim Cuts.",
    "skim_name": "post_reco_skim_cuts",
    "target_lists": ["Upsilon(4S):all"],
    "cuts": [],
    "cands_in_lists": "<1",
}

list_of_skims = [
    pre_skim_cuts,
    post_signal_skim_cuts,
    post_tag_skim_cuts,
    post_reco_skim_cuts,
]

if __name__ == "__main__":
    print_skim_cuts(list_of_skims)
